#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_ABSTRACTCHEMICALELEMENT_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_ABSTRACTCHEMICALELEMENT_H

#include <string>
#include <stdexcept>
#include <utility>


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			class AbstractChemicalElement
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operator

			public:
				AbstractChemicalElement();
				virtual ~AbstractChemicalElement() = default;

				AbstractChemicalElement(const AbstractChemicalElement&) = default;
				AbstractChemicalElement(AbstractChemicalElement&&) noexcept = default;
				AbstractChemicalElement& operator=(const AbstractChemicalElement&) = default;
				AbstractChemicalElement& operator=(AbstractChemicalElement&&) noexcept = default;

			// Constructors, destructor, and operator
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static size_type getElementNumber(const std::string& elementSymbol);
				static std::string getElementSymbol(const size_type elementNumber);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				bool isAnion() const;
				bool isCation() const;
				double bondingAffinity() const;
				size_type coordinationNumber() const;
				size_type maxCommonBridgingAtomsSharing() const;
				size_type possibleMaxCoordinationNumber() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void resetElementNumber(const size_type elementNumber);
				void resetElementSymbol(const std::string& elementSymbol);

				double getMinimumIonicRadius() const;
				double getMaximumIonicRadius() const;
				double getMetallicRadius() const;
				size_type getElementNumber() const;
				std::string getElementSymbol() const;
				std::pair<size_type, double> getBondingMotivation() const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool isValid() const noexcept;

				double getAnionicRadius(const std::string& elementSymbol) const;
				double getMinCationicRadius(const std::string& elementSymbol) const;
				double getMaxCationicRadius(const std::string& elementSymbol) const;
				double getMetallicRadius(const bool isAnion) const;

				double getClosedPackedCationicRadius(const size_type possibleMaxCoordinationNumber) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private translators for element symbols

				bool isAnion(const std::string& elementSymbol) const;
				double bondingAffinity(const std::string& elementSymbol) const;
				size_type coordinationNumber(const std::string& elementSymbol) const;
				size_type maxCommonBridgingAtomsSharing(const std::string& elementSymbol) const;
				size_type possibleMaxCoordinationNumber(const std::string& elementSymbol) const;

				void validateElementSymbol(const std::string& elementSymbol) const;

			// Private translators for element symbols
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				bool m_isAnion;
				double m_bondingAffinity;
				size_type m_coordinationNumber;
				size_type m_maxCommonBridgingAtomsSharing;
				size_type m_possibleMaxCoordinationNumber;

				mutable std::string m_elementSymbol{};
				mutable double m_minimumIonicRadius{ 0.0 };
				mutable double m_maximumIonicRadius{ 0.0 };
				mutable double m_metallicRadius{ 0.0 };
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline bool SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::isAnion() const
{
	if (isValid())
		return m_isAnion;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"isAnion\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline bool SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::isCation() const
{
	return !(isAnion());
}

inline double SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::bondingAffinity() const
{
	if (isValid())
		return m_bondingAffinity;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"bondingAffinity\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::size_type SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::coordinationNumber() const
{
	if (isValid())
		return m_coordinationNumber;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"coordinationNumber\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::size_type SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::maxCommonBridgingAtomsSharing() const
{
	if (isValid())
		return m_maxCommonBridgingAtomsSharing;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"maxCommonBridgingAtomsSharing\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::size_type SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::possibleMaxCoordinationNumber() const
{
	if (isValid())
		return m_possibleMaxCoordinationNumber;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"possibleMaxCoordinationNumber\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::initialize() noexcept
{
	m_isAnion = false;
	m_bondingAffinity = 0.0;
	m_coordinationNumber = 0;
	m_maxCommonBridgingAtomsSharing = 0;
	m_possibleMaxCoordinationNumber = 0;

	m_elementSymbol.clear();
}

inline double SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getMinimumIonicRadius() const
{
	if (isValid())
		return m_minimumIonicRadius;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"getMinimumIonicRadius\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline double SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getMaximumIonicRadius() const
{
	if (isValid())
		return m_maximumIonicRadius;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"getMaximumIonicRadius\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline double SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getMetallicRadius() const
{
	if (isValid())
		return m_metallicRadius;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"getMetallicRadius\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::size_type SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementNumber() const
{
	size_type elementNumber = getElementNumber(m_elementSymbol);
	{
		if (isValid() && (elementNumber != 0))
			return elementNumber;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member funtion of \"getElementNumber\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
			errorMessage += "Message : This is invalid.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}
	}
}

inline std::string SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol() const
{
	if (isValid())
		return m_elementSymbol;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"getElementSymbol\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline std::pair<SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::size_type, double> SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getBondingMotivation() const
{
	if (isValid())
		return std::make_pair(m_coordinationNumber, m_bondingAffinity);

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member funtion of \"getBondingMotivation\" in the class of SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement.\n";
		errorMessage += "Message : This is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::isValid() const noexcept
{
	if (m_coordinationNumber == 0 || m_maxCommonBridgingAtomsSharing == 0 || m_possibleMaxCoordinationNumber == 0)
		return false;
	else
		return true;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_ABSTRACTCHEMICALELEMENT_H
