#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_ATOMICNUCLEUS_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_ATOMICNUCLEUS_H

#include <stdexcept>
#include <string>

#include "LinearAlgebra.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class AtomicNucleus
			{
				using size_type = std::size_t;
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructors, and operators

			public:
				explicit AtomicNucleus(const size_type atomicNumber);
				AtomicNucleus(const size_type atomicNumber, const NumericVector& cartesianCoordinate);
				virtual ~AtomicNucleus() = default;

				AtomicNucleus(const AtomicNucleus&) = default;
				AtomicNucleus(AtomicNucleus&&) noexcept = default;
				AtomicNucleus& operator=(const AtomicNucleus&) = default;
				AtomicNucleus& operator=(AtomicNucleus&&) noexcept = default;

			// Constructors, destructors, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type atomicNumber() const noexcept;
				const NumericVector& cartesianCoordinate() const noexcept;
				const NumericVector& appliedForce() const noexcept;

				void resetCartesianCoordinate(const NumericVector& cartesianCoordinate = NumericVector{});
				void resetCartesianCoordinate(NumericVector&& cartesianCoordinate);
				void resetAppliedForce(const NumericVector& appliedForce = NumericVector{});
				void resetAppliedForce(NumericVector&& appliedForce);

				void receiveForce(const double scalingMultiplier, const NumericVector& force) noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				virtual void initialize();
				virtual void move(const NumericVector& displacement);
				virtual void move(const double maxDisplacement) noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				size_type m_atomicNumber;
				NumericVector m_cartesianCoordinate;
				NumericVector m_appliedForce;


			private:
				AtomicNucleus() = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::size_t SolidStatePhysics::Crystallization::Components::AtomicNucleus::atomicNumber() const noexcept
{
	return m_atomicNumber;
}

inline const NumericalAnalysis::LinearAlgebra::NumericVector& SolidStatePhysics::Crystallization::Components::AtomicNucleus::cartesianCoordinate() const noexcept
{
	return m_cartesianCoordinate;
}

inline const NumericalAnalysis::LinearAlgebra::NumericVector& SolidStatePhysics::Crystallization::Components::AtomicNucleus::appliedForce() const noexcept
{
	return m_appliedForce;
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::resetCartesianCoordinate(const NumericVector& cartesianCoordinate)
{
	if(m_cartesianCoordinate.size() == cartesianCoordinate.size())
		m_cartesianCoordinate = cartesianCoordinate;

	else
	{
		if (cartesianCoordinate.empty())
			m_cartesianCoordinate = 0.0;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member funtion of \"resetCartesianCoordinate\" in the class of SolidStatePhysics::Crystallization::Components::AtomicNucleus.\n";
			errorMessage += "Message : The dimension of \"cartesianCoordinate\" is wrong.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::resetCartesianCoordinate(NumericVector&& cartesianCoordinate)
{
	if (m_cartesianCoordinate.size() == cartesianCoordinate.size())
		m_cartesianCoordinate = std::move(cartesianCoordinate);

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member funtion of \"resetCartesianCoordinate\" in the class of SolidStatePhysics::Crystallization::Components::AtomicNucleus.\n";
		errorMessage += "Message : The dimension of \"cartesianCoordinate\" is wrong.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::resetAppliedForce(const NumericVector& appliedForce)
{
	if (m_appliedForce.size() == appliedForce.size())
		m_appliedForce = appliedForce;

	else
	{
		if (appliedForce.empty())
			m_appliedForce = 0.0;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member funtion of \"resetAppliedForce\" in the class of SolidStatePhysics::Crystallization::Components::AtomicNucleus.\n";
			errorMessage += "Message : The dimension of \"appliedForce\" is wrong.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::resetAppliedForce(NumericVector&& appliedForce)
{
	if (m_appliedForce.size() == appliedForce.size())
		m_appliedForce = std::move(appliedForce);

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member funtion of \"resetAppliedForce\" in the class of SolidStatePhysics::Crystallization::Components::AtomicNucleus.\n";
		errorMessage += "Message : The dimension of \"appliedForce\" is wrong.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::receiveForce(const double scalingMultiplier, const NumericVector& force) noexcept
{
	m_appliedForce.add(scalingMultiplier, force);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::initialize()
{
	m_appliedForce = 0.0;
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::move(const NumericVector& displacement)
{
	if(m_cartesianCoordinate.size() == displacement.size())
		m_cartesianCoordinate += displacement;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member funtion of \"move\" in the class of SolidStatePhysics::Crystallization::Components::AtomicNucleus.\n";
		errorMessage += "Message : The dimension of \"displacement\" is wrong.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Components::AtomicNucleus::move(const double maxDisplacement) noexcept
{
	double displacement = m_appliedForce.norm();
	{
		if (maxDisplacement < displacement)
			m_cartesianCoordinate.add((maxDisplacement / displacement), m_appliedForce);
		else
			m_cartesianCoordinate.add(1.0, m_appliedForce);
	}

	m_appliedForce = 0.0;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_ATOMICNUCLEUS_H
