#ifndef SOLIDSTATEPHYSICS_MATERIALS_ATOMICROSTER_H
#define SOLIDSTATEPHYSICS_MATERIALS_ATOMICROSTER_H

#include <stdexcept>
#include <string>
#include <map>


namespace SolidStatePhysics
{
	namespace Materials
	{
		class AtomicRoster
		{
		public:
			using size_type = std::size_t;
			using iterator = std::map<size_type, size_type>::iterator;
			using const_iterator = std::map<size_type, size_type>::const_iterator;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			AtomicRoster() = default;
			virtual ~AtomicRoster() = default;

			AtomicRoster(const AtomicRoster&) = default;
			AtomicRoster(AtomicRoster&&) noexcept = default;
			AtomicRoster& operator=(const AtomicRoster&) = default;
			AtomicRoster& operator=(AtomicRoster&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			iterator begin() noexcept;
			const_iterator begin() const noexcept;
			const_iterator cbegin() const noexcept;
			iterator end() noexcept;
			const_iterator end() const noexcept;
			const_iterator cend() const noexcept;

			bool isEmpty() const noexcept;
			bool isCoordinated(const size_type atomicNumber) const;
			size_type countAtoms(const size_type atomicNumber) const noexcept;
			size_type countTotalAtoms() const noexcept;
			size_type countElementSpecies() const noexcept;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Customize methods

			void addAtoms(const size_type atomicNumber, const size_type atomCount = 1);
			void overEmplaceAtoms(const size_type atomicNumber, const size_type atomCount);
			bool tryEmplaceAtoms(const size_type atomicNumber, const size_type atomCount);
			void eraseAtoms(const size_type atomicNumber);
			void clear() noexcept;

		// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Protected methods

		protected:
			const std::map<size_type, size_type>& dictionary() const noexcept;

		// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::map<size_type, size_type> m_dictionary;


// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends

			friend bool operator==(const AtomicRoster&, const AtomicRoster&);
			friend bool operator!=(const AtomicRoster&, const AtomicRoster&);

		// Friends
// **********************************************************************************************************************************************************************************************************************************************************************************************
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline SolidStatePhysics::Materials::AtomicRoster::iterator SolidStatePhysics::Materials::AtomicRoster::begin() noexcept
{
	return m_dictionary.begin();
}

inline SolidStatePhysics::Materials::AtomicRoster::const_iterator SolidStatePhysics::Materials::AtomicRoster::begin() const noexcept
{
	return m_dictionary.begin();
}

inline SolidStatePhysics::Materials::AtomicRoster::const_iterator SolidStatePhysics::Materials::AtomicRoster::cbegin() const noexcept
{
	return m_dictionary.cbegin();
}

inline SolidStatePhysics::Materials::AtomicRoster::iterator SolidStatePhysics::Materials::AtomicRoster::end() noexcept
{
	return m_dictionary.end();
}

inline SolidStatePhysics::Materials::AtomicRoster::const_iterator SolidStatePhysics::Materials::AtomicRoster::end() const noexcept
{
	return m_dictionary.end();
}

inline SolidStatePhysics::Materials::AtomicRoster::const_iterator SolidStatePhysics::Materials::AtomicRoster::cend() const noexcept
{
	return m_dictionary.cend();
}

inline bool SolidStatePhysics::Materials::AtomicRoster::isEmpty() const noexcept
{
	return m_dictionary.empty();
}

inline bool SolidStatePhysics::Materials::AtomicRoster::isCoordinated(const size_type atomicNumber) const
{
	return !(m_dictionary.find(atomicNumber) == m_dictionary.end());
}

inline SolidStatePhysics::Materials::AtomicRoster::size_type SolidStatePhysics::Materials::AtomicRoster::countAtoms(const size_type atomicNumber) const noexcept
{
	const_iterator iter = m_dictionary.find(atomicNumber);

	if (iter == m_dictionary.end())
		return 0;
	else
		return iter->second;
}

inline SolidStatePhysics::Materials::AtomicRoster::size_type SolidStatePhysics::Materials::AtomicRoster::countTotalAtoms() const noexcept
{
	size_type totalCoordinatedAtomCount = 0;

	for (const auto& NumAndCount : m_dictionary)
		totalCoordinatedAtomCount += NumAndCount.second;

	return totalCoordinatedAtomCount;
}

inline SolidStatePhysics::Materials::AtomicRoster::size_type SolidStatePhysics::Materials::AtomicRoster::countElementSpecies() const noexcept
{
	return m_dictionary.size();
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize functions

inline void SolidStatePhysics::Materials::AtomicRoster::addAtoms(const size_type atomicNumber, const size_type atomCount)
{
	iterator iter = m_dictionary.find(atomicNumber);

	if (iter == m_dictionary.end())
		m_dictionary.emplace(atomicNumber, atomCount);
	else
		iter->second += atomCount;
}

inline void SolidStatePhysics::Materials::AtomicRoster::overEmplaceAtoms(const size_type atomicNumber, const size_type atomCount)
{
	const_iterator iter = m_dictionary.find(atomicNumber);
	{
		if (iter != m_dictionary.end())
			m_dictionary.erase(iter);
	}

	m_dictionary.emplace(atomicNumber, atomCount);
}

inline bool SolidStatePhysics::Materials::AtomicRoster::tryEmplaceAtoms(const size_type atomicNumber, const size_type atomCount)
{
	return m_dictionary.try_emplace(atomicNumber, atomCount).second;
}

inline void SolidStatePhysics::Materials::AtomicRoster::eraseAtoms(const size_type atomicNumber)
{
	m_dictionary.erase(atomicNumber);
}

inline void SolidStatePhysics::Materials::AtomicRoster::clear() noexcept
{
	m_dictionary.clear();
}

// Customize functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline const std::map<SolidStatePhysics::Materials::AtomicRoster::size_type, SolidStatePhysics::Materials::AtomicRoster::size_type>& SolidStatePhysics::Materials::AtomicRoster::dictionary() const noexcept
{
	return m_dictionary;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

namespace SolidStatePhysics
{
	namespace Materials
	{
		inline bool operator==(const AtomicRoster& cca, const AtomicRoster& ccb)
		{
			return (cca.m_dictionary == ccb.m_dictionary);
		}

		inline bool operator!=(const AtomicRoster& cca, const AtomicRoster& ccb)
		{
			return (cca.m_dictionary != ccb.m_dictionary);
		}
	}
}

// Friend functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_ATOMICROSTER_H
