#ifndef SOLIDSTATEPHYSICS_MATERIALS_CHEMICALCOMPOSITION_H
#define SOLIDSTATEPHYSICS_MATERIALS_CHEMICALCOMPOSITION_H

#include <map>
#include <string>

#include "PeriodicTable.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		class ChemicalConstituents;

		class ChemicalComposition final
		{
		public:
			using size_type = std::size_t;
			using iterator = std::map<size_type, size_type>::iterator;
			using const_iterator = std::map<size_type, size_type>::const_iterator;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			~ChemicalComposition() = default;

			ChemicalComposition(const ChemicalComposition&) = default;
			ChemicalComposition(ChemicalComposition&&) noexcept = default;
			ChemicalComposition& operator=(const ChemicalComposition&) = default;
			ChemicalComposition& operator=(ChemicalComposition&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access functions

			bool isEmpty() const noexcept;
			bool isCoordinated(const size_type atomicNumber) const;
			size_type countComposition(const size_type atomicNumber) const noexcept;
			size_type countElementSpecies() const noexcept;

			iterator begin() noexcept;
			const_iterator begin() const noexcept;
			const_iterator cbegin() const noexcept;
			iterator end() noexcept;
			const_iterator end() const noexcept;
			const_iterator cend() const noexcept;

		// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			std::string callName() const noexcept;
			std::string getChemicalSystemName() const;
			void clear() noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends and private constructors

			friend class ChemicalConstituents;
			friend bool operator==(const ChemicalComposition&, const ChemicalComposition&);
			friend bool operator!=(const ChemicalComposition&, const ChemicalComposition&);

		private:
			ChemicalComposition() = default;
			ChemicalComposition(std::map<size_type, size_type>&& chemicalCompositionDictionary);

		// Friends and private constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::map<size_type, size_type> m_chemicalCompositionDictionary;
		};
	}
}


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access functions

inline bool SolidStatePhysics::Materials::ChemicalComposition::isEmpty() const noexcept
{
	return m_chemicalCompositionDictionary.empty();
}

inline bool SolidStatePhysics::Materials::ChemicalComposition::isCoordinated(const size_type atomicNumber) const
{
	return !(m_chemicalCompositionDictionary.find(atomicNumber) == m_chemicalCompositionDictionary.end());
}

inline std::size_t SolidStatePhysics::Materials::ChemicalComposition::countComposition(const size_type atomicNumber) const noexcept
{
	const_iterator iter = m_chemicalCompositionDictionary.find(atomicNumber);

	if (iter == m_chemicalCompositionDictionary.end())
	{
		return 0;
	}

	else
	{
		return iter->second;
	}
}

inline std::size_t SolidStatePhysics::Materials::ChemicalComposition::countElementSpecies() const noexcept
{
	return m_chemicalCompositionDictionary.size();
}

inline SolidStatePhysics::Materials::ChemicalComposition::iterator SolidStatePhysics::Materials::ChemicalComposition::begin() noexcept
{
	return m_chemicalCompositionDictionary.begin();
}

inline SolidStatePhysics::Materials::ChemicalComposition::const_iterator SolidStatePhysics::Materials::ChemicalComposition::begin() const noexcept
{
	return m_chemicalCompositionDictionary.begin();
}

inline SolidStatePhysics::Materials::ChemicalComposition::const_iterator SolidStatePhysics::Materials::ChemicalComposition::cbegin() const noexcept
{
	return m_chemicalCompositionDictionary.cbegin();
}

inline SolidStatePhysics::Materials::ChemicalComposition::iterator SolidStatePhysics::Materials::ChemicalComposition::end() noexcept
{
	return m_chemicalCompositionDictionary.end();
}

inline SolidStatePhysics::Materials::ChemicalComposition::const_iterator SolidStatePhysics::Materials::ChemicalComposition::end() const noexcept
{
	return m_chemicalCompositionDictionary.end();
}

inline SolidStatePhysics::Materials::ChemicalComposition::const_iterator SolidStatePhysics::Materials::ChemicalComposition::cend() const noexcept
{
	return m_chemicalCompositionDictionary.cend();
}

// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline std::string SolidStatePhysics::Materials::ChemicalComposition::callName() const noexcept
{
	std::string materialName;
	{
		for (const auto& numAndCount : m_chemicalCompositionDictionary)
		{
			materialName += SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(numAndCount.first);
			materialName += "_";
			materialName += std::to_string(numAndCount.second);
			materialName += "_";
		}

		materialName.pop_back();
	}

	return materialName;
}

inline std::string SolidStatePhysics::Materials::ChemicalComposition::getChemicalSystemName() const
{
	std::string name;
	{
		for (const auto& numAndCount : m_chemicalCompositionDictionary)
		{
			name += SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(numAndCount.first);
			name += "-";
		}

		name.pop_back();
	}

	return name;
}


inline void SolidStatePhysics::Materials::ChemicalComposition::clear() noexcept
{
	m_chemicalCompositionDictionary.clear();
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private constructors

inline SolidStatePhysics::Materials::ChemicalComposition::ChemicalComposition(std::map<size_type, size_type>&& chemicalCompositionDictionary)
	: m_chemicalCompositionDictionary{ std::move(chemicalCompositionDictionary) }
{
}

// Private constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Friend functions

namespace SolidStatePhysics
{
	namespace Materials
	{
		inline bool operator==(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_chemicalCompositionDictionary == ccb.m_chemicalCompositionDictionary);
		}

		inline bool operator!=(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_chemicalCompositionDictionary != ccb.m_chemicalCompositionDictionary);
		}
	}
}

// Friend functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !SOLIDSTATEPHYSICS_MATERIALS_CHEMICALCOMPOSITION_H
