#ifndef SOLIDSTATEPHYSICS_MATERIALS_IO_CIFFILEOWNER_H
#define SOLIDSTATEPHYSICS_MATERIALS_IO_CIFFILEOWNER_H

#include <stdexcept>
#include <string>
#include <filesystem>

#include "FileOwner.h"

#include "LinearAlgebra.h"

#include "CrystalStructure.h"
#include "UnitConversion.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		namespace IO
		{
			class CifFileOwner :public ControlCenter::System::IO::FileOwner
			{
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
				using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CifFileOwner() = default;
				explicit CifFileOwner(const std::filesystem::path& cifFilePath);
				explicit CifFileOwner(std::filesystem::path&& cifFilePath);

				CifFileOwner(CifFileOwner&&) noexcept = default;
				CifFileOwner& operator=(CifFileOwner&&) noexcept = default;

				virtual ~CifFileOwner() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public functions to create a new cif file

				template <typename A, typename AA>
				void overCreateFile(const CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				bool tryCreateFile(const CrystalStructure<A, AA>&) const;

			// Public functions to create a new cif file
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				static std::string correctExtension() noexcept;
				static bool isCorrectExtension(const std::filesystem::path filePath) noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string writeHeaderMessage(const std::string& materialName) const;
				std::string writeCrystalInfo(const ChemicalConstituents&, const double cellVolume) const;

				template <typename A, typename AA>
				std::string writePrimitiveLatticeVectors(const CrystalStructure<A, AA>&) const;
				
				template <typename A, typename AA>
				std::string writeCoordinatedAtomicArrangement(const CrystalStructure<A, AA>&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				CifFileOwner(const CifFileOwner&) = delete;
				CifFileOwner& operator=(const CifFileOwner&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public functions to create a new cif file

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::CifFileOwner::overCreateFile(const CrystalStructure<A, AA>& crystalStructure) const
{
	ChemicalConstituents chemicalConstituents = crystalStructure.getChemicalConstituents();

	constexpr double lengthScalingMultiplier = SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>{}.scalingMultiplier();
	constexpr double volumeMultiplier = (lengthScalingMultiplier * lengthScalingMultiplier * lengthScalingMultiplier);

	std::string outputString;
	outputString += writeHeaderMessage(chemicalConstituents.name());
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeCrystalInfo(chemicalConstituents, (volumeMultiplier * crystalStructure.crystalLattice().getPrimitiveCellVolume()));
	outputString += writeCoordinatedAtomicArrangement(crystalStructure);

	FileOwner::overCreateFile(outputString);
}

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::IO::CifFileOwner::tryCreateFile(const CrystalStructure<A, AA>& crystalStructure) const
{
	ChemicalConstituents chemicalConstituents = crystalStructure.getChemicalConstituents();

	constexpr double lengthScalingMultiplier = SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>{}.scalingMultiplier();
	constexpr double volumeMultiplier = (lengthScalingMultiplier * lengthScalingMultiplier * lengthScalingMultiplier);

	std::string outputString;
	outputString += writeHeaderMessage(chemicalConstituents.name());
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeCrystalInfo(chemicalConstituents, (volumeMultiplier * crystalStructure.crystalLattice().getPrimitiveCellVolume()));
	outputString += writeCoordinatedAtomicArrangement(crystalStructure);

	return FileOwner::tryCreateFile(outputString);
}

// Public functions to create a new cif file
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline std::string SolidStatePhysics::Materials::IO::CifFileOwner::correctExtension() noexcept
{
	return std::string{ ".cif" };
}

inline bool SolidStatePhysics::Materials::IO::CifFileOwner::isCorrectExtension(const std::filesystem::path filePath) noexcept
{
	if (filePath.extension().string() == correctExtension())
		return true;
	else
		return false;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::CifFileOwner::writePrimitiveLatticeVectors(const CrystalStructure<A, AA>& crystalStructure) const
{
	CrystalLatticeParameters latticeParameters = crystalStructure.crystalLattice().getCrystalLatticeParameters();

	std::stringstream outputStream;
	{
		using length_cast = SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>;

		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);

		outputStream << "_cell_length_a        " << length_cast{}(latticeParameters.length.a) << "\n";
		outputStream << "_cell_length_b        " << length_cast{}(latticeParameters.length.b) << "\n";
		outputStream << "_cell_length_c        " << length_cast{}(latticeParameters.length.c) << "\n";
		outputStream << "_cell_angle_alpha     " << latticeParameters.angle.alpha << "\n";
		outputStream << "_cell_angle_beta      " << latticeParameters.angle.beta << "\n";
		outputStream << "_cell_angle_gamma     " << latticeParameters.angle.gamma << "\n";
	}

	return outputStream.str();
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::CifFileOwner::writeCoordinatedAtomicArrangement(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputText;
	{
		outputText += " _atom_site_type_symbol\n";
		outputText += " _atom_site_label\n";
		outputText += " _atom_site_symmetry_multiplicity\n";
		outputText += " _atom_site_fract_x\n";
		outputText += " _atom_site_fract_y\n";
		outputText += " _atom_site_fract_z\n";
		outputText += " _atom_site_occupancy\n";
	}

	std::stringstream outputStream;
	{
		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);

		std::map<std::size_t, std::size_t> constituents;

		for (const auto& coordinatedAtom : crystalStructure.atomicArrangement())
		{
			std::map<std::size_t, std::size_t>::iterator iter = constituents.find(coordinatedAtom.atomicNumber());
			{
				if (iter == constituents.end())
					constituents.emplace(coordinatedAtom.atomicNumber(), 1);
				else
					iter->second += 1;
			}

			NumericVector fractionalCoordinate = crystalStructure.crystalLattice().toFractionalCoordinate(coordinatedAtom.cartesianCoordinate());
			outputStream << "  " << SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(coordinatedAtom.atomicNumber());
			outputStream << "  " << SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(coordinatedAtom.atomicNumber()) << constituents[coordinatedAtom.atomicNumber()] << "  1  ";
			outputStream << fractionalCoordinate[0] << "  " << fractionalCoordinate[1] << "  " << fractionalCoordinate[2];
			outputStream << "  1\n";
		}

		outputText += outputStream.str();
	}

	return outputText;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !SOLIDSTATEPHYSICS_MATERIALS_IO_CIFFILEOWNER_H
