#ifndef SOLIDSTATEPHYSICS_MATERIALS_CRYSTALSTRUCTURE_H
#define SOLIDSTATEPHYSICS_MATERIALS_CRYSTALSTRUCTURE_H

#include <stdexcept>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <random>

#include "LinearAlgebra.h"

#include "ChemicalConstituents.h"
#include "CrystalLattice.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		/*
		Crystalline Solid
		*/

		template <typename A, typename AA = std::vector<A>>
		class CrystalStructure
		{
		public:
			using size_type = std::size_t;
			using Atom = A;
			using AtomicArrangement = AA;

			using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
			using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructors, and operators

		public:
			CrystalStructure() = default;
			CrystalStructure(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement = AtomicArrangement{}, const std::string& name = std::string{});
			CrystalStructure(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name = std::string{});

			template <typename B, typename BA>
			explicit CrystalStructure(const CrystalStructure<B, BA>&);

			virtual ~CrystalStructure() = default;

			CrystalStructure(const CrystalStructure&) = default;
			CrystalStructure(CrystalStructure&&) noexcept = default;
			CrystalStructure& operator=(const CrystalStructure&) = default;
			CrystalStructure& operator=(CrystalStructure&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const CrystalLattice& crystalLattice() const noexcept;
			CrystalLattice& crystalLattice() noexcept;

			const AtomicArrangement& atomicArrangement() const noexcept;
			AtomicArrangement& atomicArrangement() noexcept;

			std::string name() const;
			void rename(const std::string& name = std::string{}) noexcept;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			virtual void renewCrystalLatticeRandomly(const double minLength, const double maxLength);
			virtual void renewAtomicArrangementRandomly() noexcept;
			virtual void renewAtomicArrangementRandomly(const ChemicalConstituents&) noexcept;
			virtual void distortCrystalLatticeRandomly(const double maxDistortion);
			virtual void distributeAtomsRandomly(const double maxDisplacement) noexcept;

			virtual void normalize() noexcept;
			virtual void normalizeCrystalLattice() noexcept;
			virtual void normalizeFractionalCoordinates() noexcept;

			void erase(const size_type trashAtomicNumber);
			virtual void sort();
			virtual bool isValid() const noexcept;
			virtual void clear() noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Crystal Information

			bool isCoordinated(const size_type atomicNumber) const noexcept;
			size_type countAtoms(const size_type atomicNumber) const noexcept;
			size_type countTotalAtoms() const noexcept;
			size_type countElementSpecies() const noexcept; // The number of the kinds of chemical elements

			ChemicalConstituents getChemicalConstituents() const noexcept;

		// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::string m_name;
			CrystalLattice m_crystalLattice;
			AtomicArrangement m_atomicArrangement;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

template <typename A, typename AA>
inline SolidStatePhysics::Materials::CrystalStructure<A, AA>::CrystalStructure(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement, const std::string& name)
	: m_name{ name }
	, m_crystalLattice{ crystalLattice }
	, m_atomicArrangement{ atomicArrangement }
{
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::CrystalStructure<A, AA>::CrystalStructure(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name)
	: m_name{ name }
	, m_crystalLattice{ crystalLattice }
	, m_atomicArrangement{ std::move(atomicArrangement) }
{
}

template <typename A, typename AA>
template <typename B, typename BA>
inline SolidStatePhysics::Materials::CrystalStructure<A, AA>::CrystalStructure(const CrystalStructure<B, BA>& crystalStructure)
	: m_name{}
	, m_crystalLattice{ crystalStructure.crystalLattice() }
	, m_atomicArrangement{}
{
	if (crystalStructure.isValid())
	{
		for (const auto& inputAtom : crystalStructure.atomicArrangement())
			m_atomicArrangement.push_back(Atom{ inputAtom.atomicNumber(), inputAtom.cartesianCoordinate() });
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The constructor of the class of SolidStatePhysics::Materials::CrystalStructure.\n";
		errorMessage += "Message : \"crystalStructure\" is invalid.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Crystal Information

template <typename A, typename AA>
inline const SolidStatePhysics::Materials::CrystalLattice& SolidStatePhysics::Materials::CrystalStructure<A, AA>::crystalLattice() const noexcept
{
	return m_crystalLattice;
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::CrystalLattice& SolidStatePhysics::Materials::CrystalStructure<A, AA>::crystalLattice() noexcept
{
	return m_crystalLattice;
}

template <typename A, typename AA>
inline const typename SolidStatePhysics::Materials::CrystalStructure<A, AA>::AtomicArrangement& SolidStatePhysics::Materials::CrystalStructure<A, AA>::atomicArrangement() const noexcept
{
	return m_atomicArrangement;
}

template <typename A, typename AA>
inline typename SolidStatePhysics::Materials::CrystalStructure<A, AA>::AtomicArrangement& SolidStatePhysics::Materials::CrystalStructure<A, AA>::atomicArrangement() noexcept
{
	return m_atomicArrangement;
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::CrystalStructure<A, AA>::name() const
{
	if (m_name.empty())
	{
		if (m_atomicArrangement.empty())
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"name\" in the class of SolidStatePhysics::Materials::CrystalStructure.\n";
			errorMessage += "Message : \"m_atomicArrangement\" is empty.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}

		else
			return getChemicalConstituents().name();
	}

	else
		return m_name;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::rename(const std::string& name) noexcept
{
	if (name.empty())
		m_name.clear();
	else
		m_name = name;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

#include "UnitConversion.h"

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::renewCrystalLatticeRandomly(const double minLength, const double maxLength)
{
	std::random_device seedGenerator;
	std::default_random_engine engine(seedGenerator());
	std::uniform_real_distribution<> distributor(0.0, 1.0);

	constexpr double angleConstant = (1.0 / 6.0);
	NumericMatrix perimitiveLatticeVectors(3, 3);
	{
		perimitiveLatticeVectors(0, 0) = minLength + (maxLength - minLength) * distributor(engine);
	}
	{
		double length = minLength + (maxLength - minLength) * distributor(engine);
		double angle = ((0.5 - angleConstant) * NumericalAnalysis::MathematicalConstants::getPi()) + (2.0 * angleConstant * NumericalAnalysis::MathematicalConstants::getPi() * distributor(engine));
		perimitiveLatticeVectors(0, 1) = length * std::cos(angle);
		perimitiveLatticeVectors(1, 1) = length * std::sin(angle);
	}
	{
		double length = minLength + (maxLength - minLength) * distributor(engine);
		double angleA = ((0.5 - angleConstant) * NumericalAnalysis::MathematicalConstants::getPi()) + (2.0 * angleConstant * NumericalAnalysis::MathematicalConstants::getPi() * distributor(engine));
		double angleB = (2.0 * angleConstant * NumericalAnalysis::MathematicalConstants::getPi() * distributor(engine)) - (angleConstant * NumericalAnalysis::MathematicalConstants::getPi());
		perimitiveLatticeVectors(0, 2) = length * std::sin(angleB) * std::cos(angleA);
		perimitiveLatticeVectors(1, 2) = length * std::sin(angleB) * std::sin(angleA);
		perimitiveLatticeVectors(2, 2) = length * std::cos(angleB);
	}

	crystalLattice().resetPrimitiveLatticeVectors(std::move(perimitiveLatticeVectors));
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::renewAtomicArrangementRandomly()  noexcept
{
	std::random_device seedGenerator;
	std::default_random_engine engine(seedGenerator());
	std::uniform_real_distribution<> uniformDistributor(0.0, 1.0);

	for (auto& atom : m_atomicArrangement)
	{
		NumericVector fractionalCoordinate(3);
		fractionalCoordinate[0] = uniformDistributor(engine);
		fractionalCoordinate[1] = uniformDistributor(engine);
		fractionalCoordinate[2] = uniformDistributor(engine);

		atom.resetCartesianCoordinate(m_crystalLattice.primitiveLatticeVectors() * fractionalCoordinate);
	}
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::renewAtomicArrangementRandomly(const ChemicalConstituents& chemicalConstituents)  noexcept
{
	m_atomicArrangement.clear();
	{
		for (const auto& numAndCount : chemicalConstituents)
		{
			for (size_type rep = 0; rep < numAndCount.second; ++rep)
				m_atomicArrangement.push_back(Atom{ numAndCount.first });
		}
	}

	renewAtomicArrangementRandomly();
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::distortCrystalLatticeRandomly(const double maxDistortion)
{
	std::random_device seedGenerator;
	std::default_random_engine engine(seedGenerator());
	std::uniform_real_distribution<> distributor(0.0, 1.0);

	NumericMatrix transformationTensor(0.0, 3, 3);
	{
		double distortionNorm = 0.0;
		{
			double distortionNormSquare = 0.0;

			for (auto& num : transformationTensor)
			{
				num = distributor(engine);
				distortionNormSquare += num * num;
			}

			distortionNorm = std::sqrt(distortionNormSquare);
		}

		const double scalingMultiplier = ((maxDistortion * distributor(engine)) / distortionNorm);
		transformationTensor *= scalingMultiplier;
	}

	crystalLattice().transform(transformationTensor);
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::distributeAtomsRandomly(const double maxDisplacement) noexcept
{
	std::random_device seedGenerator;
	std::default_random_engine engine(seedGenerator());
	std::uniform_real_distribution<> distributor(0.0, 1.0);

	for (auto& intelligentAtom : m_atomicArrangement)
	{
		const double displacementNorm = maxDisplacement * distributor(engine);
		const double anglePhi = 2.0 * NumericalAnalysis::MathematicalConstants::getPi() * distributor(engine);
		const double angleTheta = NumericalAnalysis::MathematicalConstants::getPi() * distributor(engine);

		NumericalAnalysis::LinearAlgebra::NumericVector displacement(3);
		{
			displacement[0] = displacementNorm * std::sin(angleTheta) * std::cos(anglePhi);
			displacement[1] = displacementNorm * std::sin(angleTheta) * std::sin(anglePhi);
			displacement[2] = displacementNorm * std::cos(angleTheta);
		}

		intelligentAtom.move(displacement);
	}
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::normalize() noexcept
{
	normalizeCrystalLattice();
	normalizeFractionalCoordinates();
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::normalizeCrystalLattice() noexcept
{
	m_crystalLattice.normalize();
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::normalizeFractionalCoordinates() noexcept
{
	for (auto& atom : m_atomicArrangement)
		atom.normalizeFractionalCoordinate(m_crystalLattice);
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::erase(const size_type trashAtomicNumber)
{
	for (auto iter = m_atomicArrangement.begin(); iter != m_atomicArrangement.end();)
	{
		if (iter->atomicNumber() == trashAtomicNumber)
			iter = m_atomicArrangement.erase(iter);
		else
			++iter;
	}
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::sort()
{
	std::sort(m_atomicArrangement.begin(), m_atomicArrangement.end(), [](const Atom& aa, const Atom& ab) { return (aa.atomicNumber() < ab.atomicNumber()); });
}

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::CrystalStructure<A, AA>::isValid() const noexcept
{
	return !(m_atomicArrangement.empty());
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::CrystalStructure<A, AA>::clear() noexcept
{
	m_name.clear();
	m_crystalLattice.clear();
	m_atomicArrangement.clear();
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Crystal Information

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::CrystalStructure<A, AA>::isCoordinated(const size_type atomicNumber) const noexcept
{
	for (const auto& atom : m_atomicArrangement)
	{
		if (atom.atomicNumber() == atomicNumber)
			return true;
	}

	return false;
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::CrystalStructure<A, AA>::countAtoms(const size_type atomicNumber) const noexcept
{
	size_type atomCounting = 0;

	for (const auto& atom : m_atomicArrangement)
	{
		if (atom.atomicNumber() == atomicNumber)
			++atomCounting;
	}

	return atomCounting;
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::CrystalStructure<A, AA>::countTotalAtoms() const noexcept
{
	return m_atomicArrangement.size();
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::CrystalStructure<A, AA>::countElementSpecies() const noexcept
{
	std::map<size_type, size_type> constituents;

	for (const auto& atom : m_atomicArrangement)
	{
		std::map<size_type, size_type>::iterator iter = constituents.find(atom.atomicNumber());

		if (iter == constituents.end())
			constituents.emplace(atom.atomicNumber(), 1);
	}

	return constituents.size();
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::ChemicalConstituents SolidStatePhysics::Materials::CrystalStructure<A, AA>::getChemicalConstituents() const noexcept
{
	ChemicalConstituents chemicalConstituents;

	for (const auto& atom : m_atomicArrangement)
	{
		if (!(chemicalConstituents.tryEmplaceAtoms(atom.atomicNumber(), 1)))
			chemicalConstituents.addAtoms(atom.atomicNumber());
	}

	return chemicalConstituents;
}

// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_CRYSTALSTRUCTURE_H
