#ifndef SOLIDSTATEPHYSICS_MATERIALS_IO_DATFILEOWNER_H
#define SOLIDSTATEPHYSICS_MATERIALS_IO_DATFILEOWNER_H

#include <string>
#include <filesystem>
#include <stdexcept>
#include <sstream>
#include <regex>

#include "FileOwner.h"

#include "LinearAlgebra.h"
#include "UnitConversion.h"

#include "CrystalStructure.h"
#include "MolecularStructure.h"


namespace SolidStatePhysics
{
	namespace Materials
	{
		namespace IO
		{
			class DatFileOwner :public ControlCenter::System::IO::FileOwner
			{
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;
				using NumericMatrix = NumericalAnalysis::LinearAlgebra::NumericMatrix;

				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				DatFileOwner() = default;
				explicit DatFileOwner(const std::filesystem::path& datFilePath);
				explicit DatFileOwner(std::filesystem::path&& datFilePath);

				DatFileOwner(DatFileOwner&&) noexcept = default;
				DatFileOwner& operator=(DatFileOwner&&) noexcept = default;

				virtual ~DatFileOwner() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				static std::string correctExtension() noexcept;
				static bool isCorrectExtension(const std::filesystem::path& filePath) noexcept;

				template <typename A, typename AA>
				void overCreateFile(const CrystalStructure<A, AA>&) const;
				
				template <typename A, typename AA>
				bool tryCreateFile(const CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				void overCreateFile(const MolecularStructure<A, AA>&) const;

				template <typename A, typename AA>
				bool tryCreateFile(const MolecularStructure<A, AA>&) const;

				template <typename A, typename AA>
				void readDatFile(CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				void readDatFile(MolecularStructure<A, AA>&) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods to create a new dat file

			private:
				std::string writeHeaderMessage(const std::string& materialName) const;

				template <typename A, typename AA>
				std::string writePrimitiveLatticeVectors(const CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				std::string writeCrystallineAtomicArrangement(const CrystalStructure<A, AA>&) const;

				template <typename A, typename AA>
				std::string writeMolecularAtomicArrangement(const MolecularStructure<A, AA>&) const;

			// Private methods to create a new dat file
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods to read a dat file

				std::string readDatFileString() const;

				std::string readPrimitiveLatticeVectorsUnit(const std::string& inputString) const;
				std::string readAtomicCoordinatesUnit(const std::string& inputString) const;
				CrystalLattice readPrimitiveLatticeVectors(const std::string& inputString) const;

				template <typename A, typename AA>
				AA readAtomicArrangement(const std::string& inputString, const NumericMatrix& primitiveTranslationVectors) const;

			// Private methods to read a dat file
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				DatFileOwner(const DatFileOwner&) = delete;
				DatFileOwner& operator=(const DatFileOwner&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline std::string SolidStatePhysics::Materials::IO::DatFileOwner::correctExtension() noexcept
{
	return std::string{ ".dat" };
}

inline bool SolidStatePhysics::Materials::IO::DatFileOwner::isCorrectExtension(const std::filesystem::path& filePath) noexcept
{
	if (filePath.extension().string() == correctExtension())
		return true;
	else
		return false;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::DatFileOwner::overCreateFile(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.getChemicalConstituents().name());
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeCrystallineAtomicArrangement(crystalStructure);

	FileOwner::overCreateFile(outputString);
}

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::IO::DatFileOwner::tryCreateFile(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.getChemicalConstituents().name());
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeCrystallineAtomicArrangement(crystalStructure);

	return FileOwner::tryCreateFile(outputString);
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::DatFileOwner::overCreateFile(const MolecularStructure<A, AA>& molecularStructure) const
{
	std::string outputString;
	outputString += writeHeaderMessage(molecularStructure.chemicalConstituents().callName());
	outputString += writeMolecularAtomicArrangement(molecularStructure);

	FileOwner::overCreateFile(outputString);
}

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::IO::DatFileOwner::tryCreateFile(const MolecularStructure<A, AA>& molecularStructure) const
{
	std::string outputString;
	outputString += writeHeaderMessage(molecularStructure.chemicalConstituents().callName());
	outputString += writeMolecularAtomicArrangement(molecularStructure);

	return FileOwner::tryCreateFile(outputString);
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::DatFileOwner::readDatFile(CrystalStructure<A, AA>& crystalStructure) const
 {
	std::string inputString = readDatFileString();

	CrystalLattice crystalLattice = readPrimitiveLatticeVectors(inputString);
	crystalStructure = CrystalStructure<A, AA>{ crystalLattice, readAtomicArrangement<A, AA>(inputString, crystalLattice.primitiveLatticeVectors()) };
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::IO::DatFileOwner::readDatFile(MolecularStructure<A, AA>& molecularStructure) const
{
	std::string inputString = readDatFileString();

	typename MolecularStructure<A, AA>::AtomicArrangement atomicArrangement;
	molecularStructure = MolecularStructure<A, AA>{ std::move(atomicArrangement) };
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods to create a new dat file

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::DatFileOwner::writePrimitiveLatticeVectors(const CrystalStructure<A, AA>& crystalStructure) const
{
	NumericMatrix primitiveLatticeVectors = crystalStructure.crystalLattice().primitiveLatticeVectors();
	primitiveLatticeVectors *= SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>{}.scalingMultiplier();

	std::string outputText;
	outputText += "---------------------------------------------------------------------------------\n";
	outputText += "#\n";
	outputText += "# Primitive translation vectors\n";
	outputText += "#\n\n";
	outputText += "Atoms.UnitVectors.Unit   Ang\n";
	outputText += "<Atoms.UnitVectors\n";
	{
		std::stringstream outputStream;
		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);
		outputStream << primitiveLatticeVectors(0, 0) << "  " << primitiveLatticeVectors(1, 0) << "  " << primitiveLatticeVectors(2, 0) << "\n";
		outputStream << primitiveLatticeVectors(0, 1) << "  " << primitiveLatticeVectors(1, 1) << "  " << primitiveLatticeVectors(2, 1) << "\n";
		outputStream << primitiveLatticeVectors(0, 2) << "  " << primitiveLatticeVectors(1, 2) << "  " << primitiveLatticeVectors(2, 2) << "\n";
		outputStream << "Atoms.UnitVectors>\n\n";
		outputText += outputStream.str();
	}

	outputText += "---------------------------------------------------------------------------------\n";
	return outputText;
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::DatFileOwner::writeCrystallineAtomicArrangement(const CrystalStructure<A, AA>& crystalStructure) const
{
	std::string outputText;
	outputText += "---------------------------------------------------------------------------------\n";
	outputText += "#\n";
	outputText += "# Atomic arrangement\n";
	outputText += "#\n\n";
	outputText += "Atoms.SpeciesAndCoordinates.Unit   FRAC\n";
	outputText += "Atoms.Number  ";
	outputText += std::to_string(crystalStructure.countTotalAtoms());
	outputText += "\n<Atoms.SpeciesAndCoordinates\n";
	{
		std::stringstream outputStream;
		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);
		{
			unsigned int num = 1;

			for (const auto& coordinatedAtom : crystalStructure.atomicArrangement())
			{
				NumericVector fractionalCoordinate = crystalStructure.crystalLattice().toFractionalCoordinate(coordinatedAtom.cartesianCoordinate());
				outputStream << num++ << "  " << SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(coordinatedAtom.atomicNumber()) << "  ";
				outputStream << fractionalCoordinate[0] << "  " << fractionalCoordinate[1] << "  " << fractionalCoordinate[2];
				outputStream << "\n";
			}
		}
		outputText += outputStream.str();
	}

	outputText += "Atoms.SpeciesAndCoordinates>\n\n";
	outputText += "---------------------------------------------------------------------------------\n";
	return outputText;
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::IO::DatFileOwner::writeMolecularAtomicArrangement(const MolecularStructure<A, AA>& molecularStructure) const
{
	std::string outputText;
	outputText += "---------------------------------------------------------------------------------\n";
	outputText += "#\n";
	outputText += "# Atomic arrangement\n";
	outputText += "#\n\n";
	outputText += "Atoms.SpeciesAndCoordinates.Unit   Ang\n";
	outputText += "Atoms.Number  ";
	outputText += std::to_string(molecularStructure.countTotalAtoms());
	outputText += "\n<Atoms.SpeciesAndCoordinates\n";
	{
		std::stringstream outputStream;
		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);
		{
			unsigned int num = 1;

			for (const auto& coordinatedAtom : molecularStructure.atomicArrangement())
			{
				NumericVector cartesianCoordinate = coordinatedAtom.cartesianCoordinate();
				cartesianCoordinate *= SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::Angstrom, SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>{}.scalingMultiplier();

				outputStream << num++ << "  " << SolidStatePhysics::ChemicalElements::PeriodicTable::replyElementSymbol(coordinatedAtom.atomicNumber()) << "  ";
				outputStream << cartesianCoordinate[0] << "  " << cartesianCoordinate[1] << "  " << cartesianCoordinate[2];
				outputStream << "\n";
			}
		}
		outputText += outputStream.str();
	}

	outputText += "Atoms.SpeciesAndCoordinates>\n\n";
	outputText += "---------------------------------------------------------------------------------\n";
	return outputText;
}

// Private methods to create a new dat file
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods to read a dat file

template <typename A, typename AA>
inline AA SolidStatePhysics::Materials::IO::DatFileOwner::readAtomicArrangement(const std::string& inputString, const NumericMatrix& primitiveTranslationVectors) const
{
	using AtomicArrangement = AA;

	std::string coordinateUnit = readAtomicCoordinatesUnit(inputString);
	{
		if (coordinateUnit == "FRAC")
		{
			if (primitiveTranslationVectors.columnSize() != 3 || primitiveTranslationVectors.rowSize() != 3)
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Runtime error\n";
				errorMessage += "Location: The member function of \"readAtomicArrangement\" in the class of SolidStatePhysics::Materials::IO::DatFileOwner.\n";
				errorMessage += "Message : The sizes of column and/or row are not three.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::runtime_error(errorMessage);
			}
		}
	}

	AtomicArrangement atomicArrangement;
	{
		std::stringstream inputStream{ inputString };

		for (std::string stringLine; std::getline(inputStream, stringLine);)
		{
			if (stringLine == "<Atoms.SpeciesAndCoordinates")
				break;
		}

		bool existFinalFlag = false;
		{
			for (std::string stringLine; std::getline(inputStream, stringLine);)
			{
				std::regex atomPat("[\\s]*[\\d]+[\\s]+([\\w]+)[\\s]+([e\\d\\.\\-]+)[\\s]+([e\\d\\.\\-]+)[\\s]+([e\\d\\.\\-]+)[\\s]*[[:alnum:]:/\\-\\.\\_\\s]*");
				std::smatch atomSm;

				if (std::regex_match(stringLine, atomSm, atomPat))
				{
					typename CrystalStructure<A, AA>::Atom atom(SolidStatePhysics::ChemicalElements::PeriodicTable::replyAtomicNumber(atomSm.str(1)));

					NumericVector coordinate(3);
					{
						std::stringstream coordinateStreamA{ atomSm.str(2) };
						std::stringstream coordinateStreamB{ atomSm.str(3) };
						std::stringstream coordinateStreamC{ atomSm.str(4) };
						coordinateStreamA >> coordinate[0];
						coordinateStreamB >> coordinate[1];
						coordinateStreamC >> coordinate[2];
					}

					if (coordinateUnit == "FRAC")
						atom.resetCartesianCoordinate(NumericalAnalysis::LinearAlgebra::multiply(1.0, primitiveTranslationVectors, coordinate));

					else
					{
						if (coordinateUnit == "Ang")
							atom.resetCartesianCoordinate(SolidStatePhysics::UnitSystem::length_cast<SolidStatePhysics::UnitSystem::LengthUnit::AtomicUnit>{}.scalingMultiplier()* coordinate);
						else
							atom.resetCartesianCoordinate(coordinate);
					}

					atomicArrangement.push_back(std::move(atom));
				}

				else
				{
					if (stringLine == "Atoms.SpeciesAndCoordinates>")
					{
						existFinalFlag = true;
						break;
					}
				}
			}

			if (!existFinalFlag || atomicArrangement.empty())
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Runtime error\n";
				errorMessage += "Location: The member function of \"readAtomicArrangement\" in the class of SolidStatePhysics::Materials::IO::DatFileOwner.\n";
				errorMessage += "Message : Couldn't read the atomic arrangement from the dat file of ";
				{
					errorMessage += filePath().generic_string();
					errorMessage += ".\n\n";
				}
				errorMessage += "------------------------------------------------------\n\n";
				throw std::runtime_error(errorMessage);
			}
		}
	}

	return atomicArrangement;
}

// Private methods to read a dat file
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_IO_DATFILEOWNER_H
