#ifndef CONTROLCENTER_SYSTEM_IO_DIRECTORYMANIPULATOR_H
#define CONTROLCENTER_SYSTEM_IO_DIRECTORYMANIPULATOR_H

#include <stdexcept>
#include <string>
#include <filesystem>


namespace ControlCenter
{
	namespace System
	{
		namespace IO
		{
			class DirectoryManipulator
			{
				/*
				This never create directory if the parent directory does not exist.
				The "over~" methods forces to create, copy, or move directories and files after deleting all the old objects.
				The "try~" methods never delete old objects. If old objects exists, the methods return "false" without operations.
				Note that all the method neglects unforseen eitities, for example, if the old objects are unforeseen entity, the "try~" methods throw exceptions.
				"entity" is the objects such as directory, file, and so on.
				*/

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructor and destructor

			private:
				DirectoryManipulator() = default;
				virtual ~DirectoryManipulator() = default;

			// Constructor and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

			public:
				static void overCreateDirectory(const std::filesystem::path& directoryPath);
				static void overInitializeDirectory(const std::filesystem::path& directoryPath);
				static bool tryCreateDirectory(const std::filesystem::path& directoryPath);
				static bool tryInitializeDirectory(const std::filesystem::path& directoryPath);

				static void overDuplicateAllFiles(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath);
				static bool tryDuplicateAllFiles(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath);

				static void overDuplicateDirectory(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath);
				static bool tryDuplicateDirectory(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath);

				static void overMoveDirectory(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);
				static bool tryMoveDirectory(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);

				static void overTransmitAllFiles(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);
				static bool tryTransmitAllFiles(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);
				static void overTransmitAllFilesRecursively(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);
				static void transmitAllFilesRecursively(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath);

				static void overDuplicateFile(const std::filesystem::path& originalFilePath, const std::filesystem::path& replicaFilePath);
				static bool tryDuplicateFile(const std::filesystem::path& originalFilePath, const std::filesystem::path& replicaFilePath);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				static void validateDirectoryPath(const std::filesystem::path& directoryPath);
				static void validateParentDirectory(const std::filesystem::path& directoryPath);

				static void createDirectory(const std::filesystem::path& directoryPath);
				static void overTransmitFile(const std::filesystem::path& oldFilePath, const std::filesystem::path& newFilePath);
				static bool tryTransmitFile(const std::filesystem::path& oldFilePath, const std::filesystem::path& newFilePath);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				DirectoryManipulator(const DirectoryManipulator&) = delete;
				DirectoryManipulator(DirectoryManipulator&&) noexcept = delete;
				DirectoryManipulator& operator=(const DirectoryManipulator&) = delete;
				DirectoryManipulator& operator=(DirectoryManipulator&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(const std::filesystem::path& directoryPath)
{
	if (std::filesystem::exists(directoryPath))
	{
		std::filesystem::remove_all(directoryPath);
		createDirectory(directoryPath);
	}

	else
		createDirectory(directoryPath);
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryCreateDirectory(const std::filesystem::path& directoryPath)
{
	if (std::filesystem::exists(directoryPath))
	{
		if (std::filesystem::is_directory(directoryPath))
			return false;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryCreateDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += directoryPath.generic_string();
				errorMessage += " is a path to a non-directory object.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}

	else
	{
		createDirectory(directoryPath);
		return true;
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::overInitializeDirectory(const std::filesystem::path& directoryPath)
{
	overCreateDirectory(directoryPath);
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryInitializeDirectory(const std::filesystem::path& directoryPath)
{
	return tryCreateDirectory(directoryPath);
}

inline void ControlCenter::System::IO::DirectoryManipulator::overDuplicateDirectory(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath)
{
	validateDirectoryPath(originalDirectoryPath);

	if (std::filesystem::exists(replicaDirectoryPath))
	{
		std::filesystem::remove_all(replicaDirectoryPath);
		std::filesystem::copy(originalDirectoryPath, replicaDirectoryPath, std::filesystem::copy_options::recursive);
	}

	else
	{
		validateParentDirectory(replicaDirectoryPath);
		std::filesystem::copy(originalDirectoryPath, replicaDirectoryPath, std::filesystem::copy_options::recursive);
	}
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryDuplicateDirectory(const std::filesystem::path& originalDirectoryPath, const std::filesystem::path& replicaDirectoryPath)
{
	validateDirectoryPath(originalDirectoryPath);

	if (std::filesystem::is_directory(replicaDirectoryPath))
		return false;

	else
	{
		if (std::filesystem::exists(replicaDirectoryPath))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryDuplicateDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += replicaDirectoryPath.generic_string();
				errorMessage += " is not a path to a directory.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}

		else
		{
			validateParentDirectory(replicaDirectoryPath);
			std::filesystem::copy(originalDirectoryPath, replicaDirectoryPath, std::filesystem::copy_options::recursive);
			return true;
		}
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::overMoveDirectory(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath)
{
	validateDirectoryPath(oldDirectoryPath);

	if (std::filesystem::exists(newDirectoryPath))
	{
		std::filesystem::remove_all(newDirectoryPath);
		validateParentDirectory(newDirectoryPath);
		std::filesystem::rename(oldDirectoryPath, newDirectoryPath);
	}

	else
	{
		validateParentDirectory(newDirectoryPath);
		std::filesystem::rename(oldDirectoryPath, newDirectoryPath);
	}
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryMoveDirectory(const std::filesystem::path& oldDirectoryPath, const std::filesystem::path& newDirectoryPath)
{
	validateDirectoryPath(oldDirectoryPath);

	if (std::filesystem::is_directory(newDirectoryPath))
		return false;

	else
	{
		if (std::filesystem::exists(newDirectoryPath))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryMoveDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += newDirectoryPath.generic_string();
				errorMessage += " is not a path to a directory.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}

		else
		{
			validateParentDirectory(newDirectoryPath);
			std::filesystem::rename(oldDirectoryPath, newDirectoryPath);
			return true;
		}
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::overDuplicateFile(const std::filesystem::path& originalFilePath, const std::filesystem::path& replicaFilePath)
{
	{
		if (std::filesystem::exists(replicaFilePath))
			std::filesystem::remove_all(replicaFilePath);
	}

	if (!(std::filesystem::copy_file(originalFilePath, replicaFilePath)))
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Runtime error\n";
		errorMessage += "Location: The member function of \"overDuplicateFile\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
		errorMessage += "Message : Could not copy the file of ";
		{
			errorMessage += originalFilePath.generic_string();
			errorMessage += " to the file of ";
			errorMessage += replicaFilePath.generic_string();
			errorMessage += ".\n\n";
		}
		errorMessage += "------------------------------------------------------\n\n";
		throw std::runtime_error(errorMessage);
	}
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryDuplicateFile(const std::filesystem::path& originalFilePath, const std::filesystem::path& replicaFilePath)
{
	if (std::filesystem::exists(replicaFilePath))
	{
		if (std::filesystem::is_regular_file(replicaFilePath))
			return false;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryDuplicateFile\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += replicaFilePath.generic_string();
				errorMessage += " is not a path to a regular file.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}

	else
	{
		if (std::filesystem::copy_file(originalFilePath, replicaFilePath))
			return true;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryDuplicateFile\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : Could not copy the file of ";
			{
				errorMessage += originalFilePath.generic_string();
				errorMessage += " to the file of ";
				errorMessage += replicaFilePath.generic_string();
				errorMessage += ".\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void ControlCenter::System::IO::DirectoryManipulator::validateDirectoryPath(const std::filesystem::path& directoryPath)
{
	if (!(std::filesystem::is_directory(directoryPath)))
	{
		if (std::filesystem::exists(directoryPath))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"validateDirectoryPath\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += directoryPath.generic_string();
				errorMessage += " is not a path to a directory.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"validateDirectoryPath\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : The directory of ";
			{
				errorMessage += directoryPath.generic_string();
				errorMessage += " does not exist.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::validateParentDirectory(const std::filesystem::path& directoryPath)
{
	if (!(std::filesystem::is_directory(directoryPath.parent_path())))
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Runtime error\n";
		errorMessage += "Location: The member function of \"validateParentDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
		errorMessage += "Message : The parent directory of ";
		{
			errorMessage += directoryPath.parent_path().generic_string();
			errorMessage += " does not exist.\n\n";
		}
		errorMessage += "------------------------------------------------------\n\n";
		throw std::runtime_error(errorMessage);
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::createDirectory(const std::filesystem::path& directoryPath)
{
	if (std::filesystem::is_directory(directoryPath.parent_path()))
	{
		try
		{
			std::filesystem::create_directory(directoryPath);
		}

		catch (const std::exception& e)
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Exception\n";
			errorMessage += "Location: The member function of \"createDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += e.what();
				errorMessage += "\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Runtime error\n";
		errorMessage += "Location: The member function of \"createDirectory\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
		errorMessage += "Message : The parent directory, which is ";
		{
			errorMessage += directoryPath.parent_path().generic_string();
			errorMessage += ", does not exist.\n\n";
		}
		errorMessage += "------------------------------------------------------\n\n";
		throw std::runtime_error(errorMessage);
	}
}

inline void ControlCenter::System::IO::DirectoryManipulator::overTransmitFile(const std::filesystem::path& oldFilePath, const std::filesystem::path& newFilePath)
{
	if (std::filesystem::exists(newFilePath))
	{
		std::filesystem::remove_all(newFilePath);
		std::filesystem::rename(oldFilePath, newFilePath);
	}

	else
		std::filesystem::rename(oldFilePath, newFilePath);
}

inline bool ControlCenter::System::IO::DirectoryManipulator::tryTransmitFile(const std::filesystem::path& oldFilePath, const std::filesystem::path& newFilePath)
{
	if (std::filesystem::exists(newFilePath))
	{
		if (std::filesystem::is_regular_file(newFilePath))
			return false;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"tryTransmitFile\" in the class of ControlCenter::System::IO::DirectoryManipulator.\n";
			errorMessage += "Message : ";
			{
				errorMessage += newFilePath.generic_string();
				errorMessage += " is not a path to a regular file.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}

	else
	{
		std::filesystem::rename(oldFilePath, newFilePath);
		return true;
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CONTROLCENTER_SYSTEM_IO_DIRECTORYMANIPULATOR_H
