#ifndef CONTROLCENTER_SYSTEM_IO_FILEOWNER_H
#define CONTROLCENTER_SYSTEM_IO_FILEOWNER_H

#include <stdexcept>
#include <string>
#include <filesystem>


namespace ControlCenter
{
	namespace System
	{
		namespace IO
		{
			class FileOwner
			{
				/*
				This never operate files if the parent directory does not exist.
				The "over~" methods forces to create, copy, or move directories and files after deleting all the old objects.
				The "try~" methods never delete old objects. If old objects exists, the methods return "false" without operations.
				Note that all the method neglects unforseen eitities, for example, if the old objects are unforeseen entity, the "try~" methods throw exceptions.
				"entity" is the objects such as directory, file, and so on.
				*/

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				FileOwner() = default;
				explicit FileOwner(const std::filesystem::path& filePath);
				explicit FileOwner(std::filesystem::path&& filePath);

				FileOwner(FileOwner&&) noexcept = default;
				FileOwner& operator=(FileOwner&&) noexcept = default;

				virtual ~FileOwner() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				const std::filesystem::path& filePath() const noexcept;

				virtual bool isValid() const;
				virtual void releaseFile() noexcept;

				virtual void write(const std::string& message) const;
				virtual std::string read() const;
				virtual std::string readLine() const;
				virtual void clearLine() const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods on file system

				bool isOverCreatable() const; // Has the parent directory already existed?
				bool isTryCreatable() const; // Has the parent directory already existed? Isn't the m_filePath a path to an existing object?

				void overCreateEmptyFile() const;
				bool tryCreateEmptyFile() const;
				bool remove() const;

				void overCopy(const std::filesystem::path& replicaFilePath) const;
				bool tryCopy(const std::filesystem::path& replicaFilePath) const;
				void overRename(const std::filesystem::path& newFilePath);
				bool tryRename(const std::filesystem::path& newFilePath);

			// Public methods on file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected methods

			protected:
				virtual void overCreateFile(const std::string& message = std::string{}) const;
				virtual bool tryCreateFile(const std::string& message = std::string{}) const;

			// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void validateOwningFile() const;
				bool isValidFilePath(const std::filesystem::path& filePath) const;
				bool isValidExtension(const std::filesystem::path& filePath) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::filesystem::path m_filePath;

			private:
				FileOwner(const FileOwner&) = delete;
				FileOwner& operator=(const FileOwner&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline const std::filesystem::path& ControlCenter::System::IO::FileOwner::filePath() const noexcept
{
	return m_filePath;
}

inline bool ControlCenter::System::IO::FileOwner::isValid() const
{
	if (std::filesystem::is_regular_file(m_filePath))
		return true;

	else
	{
		if (std::filesystem::exists(m_filePath))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"isRegularFile\" in the class of ControlCenter::System::IO::FileOwner.\n";
			errorMessage += "Message : The own path of ";
			{
				errorMessage += m_filePath.generic_string();
				errorMessage += " is not a path to a regular file.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}

		else
			return false;
	}
}

inline void ControlCenter::System::IO::FileOwner::releaseFile() noexcept
{
	m_filePath.clear();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods on file system

inline bool ControlCenter::System::IO::FileOwner::isOverCreatable() const
{
	return std::filesystem::is_directory(m_filePath.parent_path());
}

inline bool ControlCenter::System::IO::FileOwner::isTryCreatable() const
{
	if (std::filesystem::is_directory(m_filePath.parent_path()))
	{
		if (std::filesystem::is_regular_file(m_filePath))
			return false;

		else
		{
			if (std::filesystem::exists(m_filePath))
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Runtime error\n";
				errorMessage += "Location: The member function of \"isTryCreatable\" in the class of ControlCenter::System::IO::FileOwner.\n";
				errorMessage += "Message : The own path of ";
				{
					errorMessage += m_filePath.generic_string();
					errorMessage += " is not a path to a regular file.\n\n";
				}
				errorMessage += "------------------------------------------------------\n\n";
				throw std::runtime_error(errorMessage);
			}

			else
				return true;
		}
	}

	else
		return false;
}

inline void ControlCenter::System::IO::FileOwner::overCreateEmptyFile() const
{
	overCreateFile();
}

inline bool ControlCenter::System::IO::FileOwner::tryCreateEmptyFile() const
{
	return tryCreateFile();
}

inline bool ControlCenter::System::IO::FileOwner::remove() const
{
	if (std::filesystem::exists(m_filePath))
		return std::filesystem::remove_all(m_filePath);
	else
		return false;
}

// Public methods on file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void ControlCenter::System::IO::FileOwner::validateOwningFile() const
{
	if (!(std::filesystem::is_regular_file(m_filePath)))
	{
		if (std::filesystem::exists(m_filePath))
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"validateOwningFile\" in the class of ControlCenter::System::IO::FileOwner.\n";
			errorMessage += "Message : The own file path of ";
			{
				errorMessage += m_filePath.generic_string();
				errorMessage += " is not a path to a regular file.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Runtime error\n";
			errorMessage += "Location: The member function of \"validateOwningFile\" in the class of ControlCenter::System::IO::FileOwner.\n";
			errorMessage += "Message : The own file of ";
			{
				errorMessage += m_filePath.generic_string();
				errorMessage += " does not exist.\n\n";
			}
			errorMessage += "------------------------------------------------------\n\n";
			throw std::runtime_error(errorMessage);
		}
	}
}

inline bool ControlCenter::System::IO::FileOwner::isValidFilePath(const std::filesystem::path& filePath) const
{
	if (filePath.has_filename() && isValidExtension(filePath))
	{
		if (std::filesystem::is_directory(filePath.parent_path()))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool ControlCenter::System::IO::FileOwner::isValidExtension(const std::filesystem::path& filePath) const
{
	if (m_filePath.extension() == filePath.extension())
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CONTROLCENTER_SYSTEM_IO_FILEOWNER_H
