#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_FINALSIMULATIONSREPORTER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_FINALSIMULATIONSREPORTER_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <vector>
#include <map>
#include <algorithm>


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Simulators
		{
			class FinalSimulationsReporter
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				FinalSimulationsReporter();
				explicit FinalSimulationsReporter(const std::filesystem::path& workDirectoryPath, const size_type necessaryGeometricalOptimizing);

				FinalSimulationsReporter(const FinalSimulationsReporter&) = default;
				FinalSimulationsReporter(FinalSimulationsReporter&&) noexcept = default;
				FinalSimulationsReporter& operator=(const FinalSimulationsReporter&) = default;
				FinalSimulationsReporter& operator=(FinalSimulationsReporter&&) noexcept = default;

				virtual ~FinalSimulationsReporter() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void clear() noexcept;
				void createDistributionFiles() const;
				void createFinalReportFile(const std::string reportString) const;

				void registerElapsedTime(const size_type elapsedTime);
				void registerLocalOptimizationHistory(const size_type localOptimizationHistory) noexcept;
				void registerRemodellingHistory(const size_type remodellingHistory) noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void createElapsedTimesReportFile() const;
				void createLocalOptimizationHistoriesReportFile() const;
				void createRemodellingHistoriesReportFile() const;

				void createElapsedTimesPlotFile() const;
				void createLocalOptimizationHistoriesPlotFile() const;
				void createRemodellingHistoriesPlotFile() const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private utilities

				size_type getMaxElapsedTime() const noexcept;
				size_type getMaxSampleCountInElapsedTimes() const noexcept;
				size_type getElapsedTimesIndex(const size_type elapsedTime) noexcept;

				size_type getMaxLocalOptimizing() const noexcept;
				size_type getMaxSampleCountInLocalOptimizationHistories() const noexcept;

				size_type getMaxRemodelling() const noexcept;
				size_type getMaxSampleCountInRemodellingHistories() const noexcept;

				std::filesystem::path getElapsedTimesReportFilePath() const noexcept;
				std::filesystem::path getLocalOptimizationHistoriesReportFilePath() const noexcept;
				std::filesystem::path getRemodellingHistoriesReportFilePath() const noexcept;
				std::filesystem::path getElapsedTimesPlotFilePath() const noexcept;
				std::filesystem::path getLocalOptimizationHistoriesPlotFilePath() const noexcept;
				std::filesystem::path getRemodellingHistoriesPlotFilePath() const noexcept;

			// Private utilities
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::filesystem::path m_workDirectoryPath;

				size_type m_elapsedTimesDistance;
				std::vector<size_type> m_elapsedTimes;
				std::map<size_type, size_type> m_localOptimizationHistories;
				std::map<size_type, size_type> m_remodellingHistories;

				static size_type m_elapsedTimesResolution;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::createDistributionFiles() const
{
	createElapsedTimesReportFile();
	createLocalOptimizationHistoriesReportFile();
	createRemodellingHistoriesReportFile();
	createElapsedTimesPlotFile();
	createLocalOptimizationHistoriesPlotFile();
	createRemodellingHistoriesPlotFile();
}

inline void SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::registerElapsedTime(const size_type elapsedTime)
{
	if (elapsedTime != 0)
	{
		const size_type index = getElapsedTimesIndex(elapsedTime);

		if (index < m_elapsedTimes.size())
			m_elapsedTimes.at(index) += 1;

		else
		{
			std::vector<size_type>newElapsedTimes;
			newElapsedTimes.resize((index + 1), 0);
			newElapsedTimes.at(index) += 1;

			std::copy(m_elapsedTimes.begin(), m_elapsedTimes.end(), newElapsedTimes.begin());
			m_elapsedTimes.swap(newElapsedTimes);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"registerElapsedTime\" in the class of SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter.\n";
		errorMessage += "Message : \"elapsedTime\" is zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::registerLocalOptimizationHistory(const size_type localOptimizationHistory) noexcept
{
	auto iter = m_localOptimizationHistories.find(localOptimizationHistory);

	if (iter == m_localOptimizationHistories.end())
		m_localOptimizationHistories.emplace(localOptimizationHistory, 1);
	else
		iter->second += 1;
}

inline void SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::registerRemodellingHistory(const size_type remodellingHistory) noexcept
{
	auto iter = m_remodellingHistories.find(remodellingHistory);

	if (iter == m_remodellingHistories.end())
		m_remodellingHistories.emplace(remodellingHistory, 1);
	else
		iter->second += 1;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utilities

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxElapsedTime() const noexcept
{
	return (m_elapsedTimesDistance * m_elapsedTimes.size());
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxSampleCountInElapsedTimes() const noexcept
{
	auto iter = std::max_element(m_elapsedTimes.begin(), m_elapsedTimes.end());
	return *iter;
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getElapsedTimesIndex(const size_type elapsedTime) noexcept
{
	size_type index = (elapsedTime / m_elapsedTimesDistance);
	{
		if (0 == (elapsedTime % m_elapsedTimesDistance))
			--index;
	}

	return index;
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxLocalOptimizing() const noexcept
{
	auto iter = std::prev(m_localOptimizationHistories.end());
	return iter->first;
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxSampleCountInLocalOptimizationHistories() const noexcept
{
	size_type maxSampleCount{ 0 };
	{
		for (const auto numAndCount : m_localOptimizationHistories)
		{
			if (maxSampleCount < numAndCount.second)
				maxSampleCount = numAndCount.second;
		}
	}

	return maxSampleCount;
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxRemodelling() const noexcept
{
	auto iter = std::prev(m_remodellingHistories.end());
	return iter->first;
}

inline SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::size_type SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getMaxSampleCountInRemodellingHistories() const noexcept
{
	size_type maxSampleCount{ 0 };
	{
		for (const auto numAndCount : m_remodellingHistories)
		{
			if (maxSampleCount < numAndCount.second)
				maxSampleCount = numAndCount.second;
		}
	}

	return maxSampleCount;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getElapsedTimesReportFilePath() const noexcept
{
	std::filesystem::path elapsedTimesReportFilePath{ m_workDirectoryPath };
	elapsedTimesReportFilePath /= "elapsedTimesDistribution.txt";
	return elapsedTimesReportFilePath;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getLocalOptimizationHistoriesReportFilePath() const noexcept
{
	std::filesystem::path localOptimizationHistoriesReportFilePath{ m_workDirectoryPath };
	localOptimizationHistoriesReportFilePath /= "localOptimizationHistoriesDistribution.txt";
	return localOptimizationHistoriesReportFilePath;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getRemodellingHistoriesReportFilePath() const noexcept
{
	std::filesystem::path remodellingHistoriesReportFilePath{ m_workDirectoryPath };
	remodellingHistoriesReportFilePath /= "remodellingHistoriesDistribution.txt";
	return remodellingHistoriesReportFilePath;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getElapsedTimesPlotFilePath() const noexcept
{
	std::filesystem::path elapsedTimesPlotFilePath{ m_workDirectoryPath };
	elapsedTimesPlotFilePath /= "elapsedTimesDistribution.gpl";
	return elapsedTimesPlotFilePath;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getLocalOptimizationHistoriesPlotFilePath() const noexcept
{
	std::filesystem::path localOptimizationHistoriesPlotFilePath{ m_workDirectoryPath };
	localOptimizationHistoriesPlotFilePath /= "localOptimizationHistoriesDistribution.gpl";
	return localOptimizationHistoriesPlotFilePath;
}

inline std::filesystem::path SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter::getRemodellingHistoriesPlotFilePath() const noexcept
{
	std::filesystem::path remodellingHistoriesPlotFilePath{ m_workDirectoryPath };
	remodellingHistoriesPlotFilePath /= "remodellingHistoriesDistribution.gpl";
	return remodellingHistoriesPlotFilePath;
}

// Private utilities
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_FINALSIMULATIONSREPORTER_H
