#ifndef CONTROLCENTER_ADMINISTERING_GENERATEPERIODICSOLIDS_H
#define CONTROLCENTER_ADMINISTERING_GENERATEPERIODICSOLIDS_H

#include <string>
#include <filesystem>
#include <mutex>
#include <map>

#include "ChemicalConstituents.h"
#include "StructuralDictionary.h"

#include "CrystallizationPlan.h"
#include "MonitoringPlan.h"
#include "FinalSimulationsReporter.h"


namespace ControlCenter
{
	namespace Administering
	{
		class GeneratePeriodicSolids
		{
			using size_type = std::size_t;
			using ChemicalConstituents = SolidStatePhysics::Materials::ChemicalConstituents;
			using CrystallizationPlan = SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan;
			using MonitoringPlan = SolidStatePhysics::Crystallization::Simulators::MonitoringPlan;
			using StructuralDictionary = SolidStatePhysics::Crystallization::Analysis::StructuralDictionary;
			using FinalSimulationsReporter = SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor and operators

		public:
			GeneratePeriodicSolids();
			explicit GeneratePeriodicSolids(const size_type threadIndex);
			GeneratePeriodicSolids(const size_type threadIndex, const CrystallizationPlan&, const MonitoringPlan&);
			GeneratePeriodicSolids(const size_type threadIndex, CrystallizationPlan&&, MonitoringPlan&&);

			virtual ~GeneratePeriodicSolids();

			GeneratePeriodicSolids(const GeneratePeriodicSolids&) = default;
			GeneratePeriodicSolids(GeneratePeriodicSolids&&) noexcept = default;
			GeneratePeriodicSolids& operator=(const GeneratePeriodicSolids&) = default;
			GeneratePeriodicSolids& operator=(GeneratePeriodicSolids&&) noexcept = default;

		// Constructors, destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Static methods

			static void resetSampleIndex() noexcept;
			static void resetFinalSimulationsReporter(FinalSimulationsReporter&& finalSimulationsReporter);

			static void reportResults(const StructuralDictionary& structuralDictionary, const std::string& elapsedTime);

		// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const CrystallizationPlan& crystallizationPlan() const noexcept;
			const MonitoringPlan& monitoringPlan() const noexcept;

			void resetCrystallizationPlan(const CrystallizationPlan&);
			void resetCrystallizationPlan(CrystallizationPlan&&);
			void resetMonitoringPlan(const MonitoringPlan&);
			void resetMonitoringPlan(MonitoringPlan&&);

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			void operator()(const std::filesystem::path& chemicalConstituentsDirectoryPath, const ChemicalConstituents&, const size_type feedstocks) noexcept;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			void operateCrystallizer(const std::filesystem::path& sampleDirectoryPath);

			void registerFinalSimulationReport(const std::string& structuralFingerprint, const size_type elapsedTime, const size_type localOptimizationCount, const size_type remodellingCount);

			void transmitSampleToStructuralDirectory(const std::filesystem::path& sampleDirectoryPath, const size_type structuralIndex, const std::map<std::string, size_type>& atomicEnvironments);
			void transmitSampleToFailureDirectory(const std::filesystem::path& sampleDirectoryPath);
			void transmitSampleToExceptionDirectory(const std::filesystem::path& sampleDirectoryPath);

			std::string getDirectoryName(std::filesystem::path directoryPath) const noexcept;
			std::string getSampleName() const noexcept;
			bool shouldMakeSample() const noexcept;
			bool tryUseFeedstock() noexcept;

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			size_type m_threadIndex;
			CrystallizationPlan m_crystallizationPlan;
			MonitoringPlan m_monitoringPlan;

			mutable size_type m_feedstocks{};
			mutable ChemicalConstituents m_chemicalConstituents{};
			mutable StructuralDictionary m_structuralDictionary{};

			static size_type m_sampleIndex;
			static FinalSimulationsReporter m_finalSimulationsReporter;

			static std::mutex m_generateMutex;
			static std::mutex m_reportMutex;
			static std::mutex m_directoryMutex;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetSampleIndex() noexcept
{
	m_sampleIndex = 0;
}

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetFinalSimulationsReporter(FinalSimulationsReporter&& finalSimulationsReporter)
{
	m_finalSimulationsReporter = std::move(finalSimulationsReporter);
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan& ControlCenter::Administering::GeneratePeriodicSolids::crystallizationPlan() const noexcept
{
	return m_crystallizationPlan;
}

inline const SolidStatePhysics::Crystallization::Simulators::MonitoringPlan& ControlCenter::Administering::GeneratePeriodicSolids::monitoringPlan() const noexcept
{
	return m_monitoringPlan;
}

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetCrystallizationPlan(const CrystallizationPlan& crystallizationPlan)
{
	m_crystallizationPlan = crystallizationPlan;
}

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetCrystallizationPlan(CrystallizationPlan&& crystallizationPlan)
{
	m_crystallizationPlan = std::move(crystallizationPlan);
}

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetMonitoringPlan(const MonitoringPlan& monitoringPlan)
{
	m_monitoringPlan = monitoringPlan;
}

inline void ControlCenter::Administering::GeneratePeriodicSolids::resetMonitoringPlan(MonitoringPlan&& monitoringPlan)
{
	m_monitoringPlan = std::move(monitoringPlan);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string ControlCenter::Administering::GeneratePeriodicSolids::getDirectoryName(std::filesystem::path directoryPath) const noexcept
{
	std::string name;
	{
		std::string directoryPathName = directoryPath.generic_string();

		if (directoryPathName.back() == '/')
		{
			directoryPathName.pop_back();
			directoryPath = std::filesystem::path{ directoryPathName };
		}

		name = directoryPath.filename().string();
	}

	return name;
}

inline std::string ControlCenter::Administering::GeneratePeriodicSolids::getSampleName() const noexcept
{
	std::string sampleName{ "Sample_" };
	sampleName += std::to_string(m_sampleIndex);
	return sampleName;
}

inline bool ControlCenter::Administering::GeneratePeriodicSolids::shouldMakeSample() const noexcept
{
	if (m_sampleIndex < m_feedstocks)
		return true;
	else
		return false;
}

inline bool ControlCenter::Administering::GeneratePeriodicSolids::tryUseFeedstock() noexcept
{
	if (m_sampleIndex < m_feedstocks)
	{
		++m_sampleIndex;
		return true;
	}

	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CONTROLCENTER_ADMINISTERING_GENERATEPERIODICSOLIDS_H
