#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_GEOMETRICALOPTIMIZER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_GEOMETRICALOPTIMIZER_H

#include <stdexcept>
#include <string>

#include "MdFileOwner.h"
#include "ChemicalBondingRecorder.h"
#include "MonitoringPlan.h"
#include "IntelligentLiquid.h"

#include "CrystallizationVariables.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Simulators
		{
			class GeometricalOptimizer
			{
				using size_type = std::size_t;
				using IntelligentLiquid = SolidStatePhysics::Crystallization::Components::IntelligentLiquid;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			public:
				GeometricalOptimizer() = default;
				GeometricalOptimizer(const size_type maxGeometricalOptimizing, const CrystallizationVariables&, const MonitoringPlan&);
				virtual ~GeometricalOptimizer() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize(const IntelligentLiquid&, const std::filesystem::path& workDirectoryPath) const;
				void finalize() const noexcept;

				void anneal(IntelligentLiquid&, size_type& initialElapsedTime, const double maxDisplacement = -1.0, size_type necessaryAnnealing = 0) const;
				void optimize(IntelligentLiquid&, size_type& initialElapsedTime, size_type& initialLocalOptimizationCount, size_type& remodellingCount) const;
				bool canOptimize() const noexcept;

				void resetMaxGeometricalOptimizing(const size_type maxGeometricalOptimizing) noexcept;
				void resetCrystallizationVariables(const CrystallizationVariables&) noexcept;
				void resetCrystallizationVariables(CrystallizationVariables&&) noexcept;
				void resetMonitoringPlan(const MonitoringPlan&) noexcept;
				void resetMonitoringPlan(MonitoringPlan&&) noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void applyGeometricalOptimization(IntelligentLiquid&) const;

				bool initializeGeometricalConstraintersIfNecessary(IntelligentLiquid&) const;
				void manageTime() const;
				bool manageGeometricalConstrainters(IntelligentLiquid&) const;
				bool manageLinkedPolyhedra(IntelligentLiquid&) const;
				bool manageCrystalLattice(IntelligentLiquid&) const;
				bool manageThermalTransformation(IntelligentLiquid&) const;
				bool manageMdRecorder(const IntelligentLiquid&) const;
				bool manageChemicalBondingRecorder(const IntelligentLiquid&) const;

				void linkMdFile(const IntelligentLiquid&) const;
				void linkChemicalBondingRecorder() const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				size_type m_maxGeometricalOptimizing;
				CrystallizationVariables m_crystallizationVariables;
				MonitoringPlan m_monitoringPlan;

				mutable bool m_shouldKeepRunning{ false };
				mutable size_type m_time{ 0 };
				mutable size_type m_elapsedTime{ 0 };
				mutable size_type m_localOptimizationCount{ 0 };
				mutable size_type m_remodellingCount{ 0 };
				mutable size_type m_geometricalConstraintsTimer{ 0 };
				mutable size_type m_linkedPolyhedraTimer{ 0 };
				mutable size_type m_crystalLatticeTimer{ 0 };
				mutable size_type m_thermalTransformationTimer{ 0 };
				mutable size_type m_mdRecordTimer{ 0 };
				mutable size_type m_chemicalBondingRecordTimer{ 0 };
				mutable std::filesystem::path m_workDirectoryPath{};
				mutable ChemicalBondingRecorder m_chemicalBondingRecorder{};
				mutable SolidStatePhysics::Materials::IO::MdFileOwner m_mdFileOwner{};


			private:
				GeometricalOptimizer(const GeometricalOptimizer&) = delete;
				GeometricalOptimizer(GeometricalOptimizer&&) noexcept = delete;
				GeometricalOptimizer& operator=(const GeometricalOptimizer&) = delete;
				GeometricalOptimizer& operator=(GeometricalOptimizer&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::canOptimize() const noexcept
{
	if (m_time < m_crystallizationVariables.necessaryGeometricalOptimizing())
		return true;
	else
		return false;
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::resetMaxGeometricalOptimizing(const size_type maxGeometricalOptimizing) noexcept
{
	m_maxGeometricalOptimizing = maxGeometricalOptimizing;
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::resetCrystallizationVariables(const CrystallizationVariables& crystallizationVariables) noexcept
{
	m_crystallizationVariables = crystallizationVariables;
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::resetCrystallizationVariables(CrystallizationVariables&& crystallizationVariables) noexcept
{
	m_crystallizationVariables = std::move(crystallizationVariables);
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::resetMonitoringPlan(const MonitoringPlan& monitoringPlan) noexcept
{
	m_monitoringPlan = monitoringPlan;
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::resetMonitoringPlan(MonitoringPlan&& monitoringPlan) noexcept
{
	m_monitoringPlan = std::move(monitoringPlan);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::applyGeometricalOptimization(IntelligentLiquid& intelligentLiquid) const
{
	intelligentLiquid.update();
	intelligentLiquid.applyInteratomicForces();
	intelligentLiquid.optimizeAtomicArrangement();
	intelligentLiquid.transformCrystalLattice();
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::initializeGeometricalConstraintersIfNecessary(IntelligentLiquid& intelligentLiquid) const
{
	if (m_crystallizationVariables.needRenewGeometricalConstraints())
	{
		intelligentLiquid.renewGeometricalConstrainters();
		return true;
	}

	else
		return false;
}

inline void SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageTime() const
{
	++m_time;
	++m_elapsedTime;

	if (m_crystallizationVariables.necessaryGeometricalOptimizing() <= m_time)
		m_shouldKeepRunning = false;
	if(m_maxGeometricalOptimizing <= m_elapsedTime)
		m_shouldKeepRunning = false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageGeometricalConstrainters(IntelligentLiquid& intelligentLiquid) const
{
	if (m_crystallizationVariables.needRenewGeometricalConstraints())
	{
		++m_geometricalConstraintsTimer;

		if (m_geometricalConstraintsTimer == m_crystallizationVariables.geometricalConstraintersLifetime())
		{
			intelligentLiquid.renewGeometricalConstrainters();
			m_geometricalConstraintsTimer = 0;
			m_linkedPolyhedraTimer = 0;
			return true;
		}

		else
			return false;
	}

	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageLinkedPolyhedra(IntelligentLiquid& intelligentLiquid) const
{
	if (m_crystallizationVariables.needRenewLinkedPolyhedra())
	{
		++m_linkedPolyhedraTimer;

		if (m_linkedPolyhedraTimer == m_crystallizationVariables.linkedPolyhedraLifetime())
		{
			intelligentLiquid.renewLinkedPolyhedra();
			m_linkedPolyhedraTimer = 0;
			return true;
		}

		else
			return false;
	}

	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageCrystalLattice(IntelligentLiquid& intelligentLiquid) const
{
	if (m_crystallizationVariables.needNormalizeCrystalLattice())
	{
		++m_crystalLatticeTimer;

		if (m_crystalLatticeTimer == m_crystallizationVariables.crystalLatticeLifetime())
		{
			intelligentLiquid.normalize();
			m_crystalLatticeTimer = 0;

			if (intelligentLiquid.crystalLattice().hasNormalCrystalLattice())
			{
				intelligentLiquid.renewGeometricalConstrainters();
				return true;
			}

			else
			{
				m_shouldKeepRunning = false;
				return false;
			}
		}

		else
			return false;
	}

	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageThermalTransformation(IntelligentLiquid& intelligentLiquid) const
{
	if (m_crystallizationVariables.needThermalTransformation())
	{
		++m_thermalTransformationTimer;

		if (m_thermalTransformationTimer == m_crystallizationVariables.thermalTransformationFrequency())
		{
			intelligentLiquid.distortCrystalLatticeRandomly(intelligentLiquid.dynamicsController().maxLatticeTransformation());
			intelligentLiquid.distributeAtomsRandomly(intelligentLiquid.dynamicsController().maxDisplacement());
			m_thermalTransformationTimer = 0;
			return true;
		}

		else
			return false;
	}

	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageMdRecorder(const IntelligentLiquid& intelligentLiquid) const
{
	if (m_monitoringPlan.needMdRecord())
	{
		++m_mdRecordTimer;

		if (m_mdRecordTimer == m_monitoringPlan.mdRecordFrequency())
		{
			m_mdFileOwner.drawFrame(intelligentLiquid, m_elapsedTime, intelligentLiquid.getPackingFraction());
			m_mdRecordTimer = 0;
			return true;
		}

		else
			return false;
	}

	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Simulators::GeometricalOptimizer::manageChemicalBondingRecorder(const IntelligentLiquid& intelligentLiquid) const
{
	if (m_monitoringPlan.needChemicalBondingRecord())
	{
		++m_chemicalBondingRecordTimer;

		if (m_chemicalBondingRecordTimer == m_monitoringPlan.chemicalBondingRecordFrequency())
		{
			m_chemicalBondingRecorder.record(m_monitoringPlan.chemicalBondingRecordFrequency(), intelligentLiquid.getChemicalBondingRate(), intelligentLiquid.getRelaxedChemicalBondsRate(), intelligentLiquid.getPackingFraction());
			m_chemicalBondingRecordTimer = 0;
			return true;
		}

		else
			return false;
	}

	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_SIMULATORS_GEOMETRICALOPTIMIZER_H
