#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTELLIGENTLIQUID_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTELLIGENTLIQUID_H

#include <string>
#include <stdexcept>
#include <map>

#include "DynamicsController.h"
#include "PeriodicLiquid.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class IntelligentLiquid :public PeriodicLiquid
			{
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;
				using CrystalStructure = SolidStatePhysics::Materials::CrystalStructure<IntelligentAtom>;

				using AtomicAdaptor = SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<InteractingAtom>;
				using DynamicsController = SolidStatePhysics::Crystallization::Calculators::DynamicsController;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructor and destructor

			public:
				IntelligentLiquid();
				IntelligentLiquid(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement = AtomicArrangement{}, const std::string& name = std::string{});
				IntelligentLiquid(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name = std::string{});

				template <typename A, typename AA>
				explicit IntelligentLiquid(const SolidStatePhysics::Materials::CrystalStructure<A, AA>&);

				virtual ~IntelligentLiquid() = default;

			// Constructor and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type idealChemicalBonding() const noexcept;

				const DynamicsController& dynamicsController() const noexcept;
				DynamicsController& dynamicsController() noexcept;

				void resetMaxBondingLengthErrorRate(const double maxBondingLengthErrorRate);
				void resetMinPackingFraction(const double minPackingFraction);
				void renewIdealChemicalBonding() noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept override;
				void update() override;
				void clear() noexcept override;

				bool hasIdealChemicalBonding() const noexcept;
				double getChemicalBondingRate() const noexcept;
				double getRelaxedChemicalBondsRate() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void optimizeAtomicArrangement();
				void transformCrystalLattice();

				void renewGeometricalConstrainters();
				void renewLinkedPolyhedra();

				void remodel();

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				size_type countAtomicEnvironments() const noexcept;
				std::map<std::string, size_type> getAtomicEnvironments() const noexcept;

				bool hasNonbondingAtoms() const noexcept;
				bool isProbablyCrystallized() const noexcept;
				bool isDefinitelyCrystallized() noexcept;

				double getPackingFraction() const noexcept;
				bool hasEnoughPackingFraction() const noexcept;
				bool isDenseCrystalStructure(const double minPackingFraction = -1.0) const noexcept;

				std::string writeFingerprint() const noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void createBasicConstrainters();
				void createChemicalBonds();
				void createPolyhedralConstrainters();
				void createSecondNeighborConstrainters();

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_maxBondingLengthErrorRate;
				double m_minPackingFraction;
				size_type m_idealChemicalBonding;
				DynamicsController m_dynamicsController;


			private:
				IntelligentLiquid(const IntelligentLiquid&) = delete;
				IntelligentLiquid(IntelligentLiquid&&) noexcept = delete;
				IntelligentLiquid& operator=(const IntelligentLiquid&) = delete;
				IntelligentLiquid& operator=(IntelligentLiquid&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

template <typename A, typename AA>
inline SolidStatePhysics::Crystallization::Components::IntelligentLiquid::IntelligentLiquid(const SolidStatePhysics::Materials::CrystalStructure<A, AA>& crystalStructure)
	: PeriodicLiquid{ crystalStructure }
	, m_maxBondingLengthErrorRate{ 0.0 }
	, m_minPackingFraction{ 0.0 }
	, m_idealChemicalBonding{ 0 }
	, m_dynamicsController{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline SolidStatePhysics::Crystallization::Components::IntelligentLiquid::size_type SolidStatePhysics::Crystallization::Components::IntelligentLiquid::idealChemicalBonding() const noexcept
{
	return m_idealChemicalBonding;
}

inline const SolidStatePhysics::Crystallization::Calculators::DynamicsController& SolidStatePhysics::Crystallization::Components::IntelligentLiquid::dynamicsController() const noexcept
{
	return m_dynamicsController;
}

inline SolidStatePhysics::Crystallization::Calculators::DynamicsController& SolidStatePhysics::Crystallization::Components::IntelligentLiquid::dynamicsController() noexcept
{
	return m_dynamicsController;
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::resetMaxBondingLengthErrorRate(const double maxBondingLengthErrorRate)
{
	if (0.0 < maxBondingLengthErrorRate)
		m_maxBondingLengthErrorRate = maxBondingLengthErrorRate;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetMaxBondingLengthErrorRate\" in the class of SolidStatePhysics::Crystallization::Components::IntelligentLiquid.\n";
		errorMessage += "Message : \"maxBondingLengthErrorRate\" is not more than 0.0.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::resetMinPackingFraction(const double minPackingFraction)
{
	if (0.0 < minPackingFraction)
	{
		if (minPackingFraction < 1.0)
			m_minPackingFraction = minPackingFraction;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetMinPackingFraction\" in the class of SolidStatePhysics::Crystallization::Components::IntelligentLiquid.\n";
			errorMessage += "Message : \"minPackingFraction\" is not less than 1.0.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetMinPackingFraction\" in the class of SolidStatePhysics::Crystallization::Components::IntelligentLiquid.\n";
		errorMessage += "Message : \"minPackingFraction\" is not more than 0.0.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::renewIdealChemicalBonding() noexcept
{
	m_idealChemicalBonding = 0;
	{
		for (auto& intelligentAtom : atomicArrangement())
		{
			if (intelligentAtom.isCation())
				m_idealChemicalBonding += intelligentAtom.bondingMotivation().first;
		}
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::initialize() noexcept
{
	PeriodicLiquid::initialize();

	m_maxBondingLengthErrorRate = 0.0;
	m_minPackingFraction = 0.0;
	m_idealChemicalBonding = 0;
	m_dynamicsController.initialize();
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::update()
{
	PeriodicLiquid::update();
	m_dynamicsController.update();
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::clear() noexcept
{
	PeriodicLiquid::clear();

	m_maxBondingLengthErrorRate = 0.0;
	m_minPackingFraction = 0.0;
	m_idealChemicalBonding = 0;
	m_dynamicsController.clear();
}

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::hasIdealChemicalBonding() const noexcept
{
	if (chemicalBonding() == m_idealChemicalBonding)
		return true;
	else
		return false;
}

inline double SolidStatePhysics::Crystallization::Components::IntelligentLiquid::getChemicalBondingRate() const noexcept
{
	return (static_cast<double>(chemicalBonding()) / static_cast<double>(m_idealChemicalBonding));
}

inline double SolidStatePhysics::Crystallization::Components::IntelligentLiquid::getRelaxedChemicalBondsRate() const noexcept
{
	size_type chemicalBonding = 0;
	size_type relaxedChemicalBonding = 0;
	{
		for (const auto& chemicalBondWeakPtr : m_chemicalBondsPtrs)
		{
			if (std::shared_ptr<ChemicalBond> chemicalBondPtr = chemicalBondWeakPtr.lock())
			{
				++chemicalBonding;

				if (chemicalBondPtr->isFeasible(m_maxBondingLengthErrorRate))
					++relaxedChemicalBonding;
			}
		}
	}

	if (chemicalBonding == 0)
		return 0;
	else
		return (static_cast<double>(relaxedChemicalBonding) / static_cast<double>(chemicalBonding));
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::renewGeometricalConstrainters()
{
	normalizeFractionalCoordinates();

	createBasicConstrainters();
	createChemicalBonds();
	createPolyhedralConstrainters();
	createSecondNeighborConstrainters();

	cleanGeometricalConstrainters();
	renew();
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::renewLinkedPolyhedra()
{
	convertIntoBasicConstraints();
	cleanGeometricalConstrainters();

	createChemicalBonds();
	createPolyhedralConstrainters();
	createSecondNeighborConstrainters();

	cleanGeometricalConstrainters();
	renew();
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::remodel()
{
	clearChemicalBonds();
	clearIonicConstrainters();
	clearMetallicConstrainters();
	clearPolyhedralConstrainters();
	clearSecondNeighborConstrainters();

	renewCrystalLatticeRandomly(1.0, 2.0);
	renewAtomicArrangementRandomly();
	renewGeometricalConstrainters();
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::hasNonbondingAtoms() const noexcept
{
	for (const auto& intelligentAtom : atomicArrangement())
	{
		if (intelligentAtom.getCoordinationNumber() == 0)
			return true;
	}

	return false;
}

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::isProbablyCrystallized() const noexcept
{
	if (chemicalBonding() == m_idealChemicalBonding)
		return true;
	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::isDefinitelyCrystallized() noexcept
{
	if (hasInfeasibleConstrainters(m_maxBondingLengthErrorRate))
		return false;
	
	else
	{
		if (chemicalBonding() == m_idealChemicalBonding)
			return true;
		else
			return false;
	}
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::optimizeAtomicArrangement()
{
	for (auto& intelligentAtom : atomicArrangement())
		intelligentAtom.move(m_dynamicsController.maxDisplacement());
}

inline void SolidStatePhysics::Crystallization::Components::IntelligentLiquid::transformCrystalLattice()
{
	applyPressure();
	{
		double transformationNorm = 0.0;
		{
			double transformationNormSquare = 0.0;
			{
				for (const auto num : crystalLattice().transformationTensor())
					transformationNormSquare += num * num;
			}

			transformationNorm = std::sqrt(transformationNormSquare);
		}

		if (m_dynamicsController.maxLatticeTransformation() < transformationNorm)
			crystalLattice().scaleTransformationTensor(m_dynamicsController.maxLatticeTransformation() / transformationNorm);
	}

	crystalLattice().transform();
}

inline double SolidStatePhysics::Crystallization::Components::IntelligentLiquid::getPackingFraction() const noexcept
{
	double packingFraction = 0.0;
	{
		for (const auto& intelligentAtom : atomicArrangement())
			packingFraction += intelligentAtom.maximumIonicRadius() * intelligentAtom.maximumIonicRadius() * intelligentAtom.maximumIonicRadius();

		constexpr double sphericalVolumeCoef = (4.0 * NumericalAnalysis::MathematicalConstants::getPi() / 3.0);
		packingFraction *= sphericalVolumeCoef;
		packingFraction /= crystalLattice().getPrimitiveCellVolume();
	}

	return packingFraction;
}

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::hasEnoughPackingFraction() const noexcept
{
	if (m_minPackingFraction <= getPackingFraction())
		return true;
	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Components::IntelligentLiquid::isDenseCrystalStructure(const double minPackingFraction) const noexcept
{
	double minimumPackingFraction = m_minPackingFraction;
	{
		if (0.0 < minPackingFraction)
			minimumPackingFraction = minPackingFraction;
	}

	if (minimumPackingFraction <= getPackingFraction())
		return !(crystalLattice().hasTooLongLatticeVector(10.0));
	else
		return false;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_INTELLIGENTLIQUID_H
