#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINMAXCONSTRAINTER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINMAXCONSTRAINTER_H

#include <memory>

#include "InteratomicConnector.h"
#include "IConstrainter.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			class MinMaxConstrainter :public InteratomicConnector, public IConstrainter
			{
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected constructors and public destructor

			protected:
				MinMaxConstrainter(InteractingAtom& ownerAtom, AtomicAdaptor&& attachedAtom);
				explicit MinMaxConstrainter(const std::shared_ptr<MinMaxConstrainter>& twinPtr);

			public:
				virtual ~MinMaxConstrainter() = default;

			// Protected constructors and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static double getDefaultAttractiveForceConstant() noexcept;
				static double getDefaultRepulsiveForceConstant() noexcept;
				static void resetAttractiveForceConstant(const double attractiveForceConstant);
				static void resetRepulsiveForceConstant(const double repulsiveForceConstant);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() override;
				double minimumDistance() const noexcept;
				double maximumDistance() const noexcept;

				double getPenaltyEnergy() const override;
				void applyForce() override;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected methods

			protected:
				void renewMinMaxDistance(const double minimumDistance, const double maximumDistance);

			// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void applyForce(const double forceScalingMultiplier);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_minimumDistance;
				double m_maximumDistance;

				static double m_attractiveForceConstant;
				static double m_repulsiveForceConstant;


			private:
				MinMaxConstrainter() = delete;
				MinMaxConstrainter(const MinMaxConstrainter&) = delete;
				MinMaxConstrainter(MinMaxConstrainter&&) = delete;
				MinMaxConstrainter& operator=(const MinMaxConstrainter&) = delete;
				MinMaxConstrainter& operator=(MinMaxConstrainter&&) = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline double SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::getDefaultAttractiveForceConstant() noexcept
{
	return 30.0;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::getDefaultRepulsiveForceConstant() noexcept
{
	return (-100.0);
}

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::resetAttractiveForceConstant(const double attractiveForceConstant)
{
	if (0.0 < attractiveForceConstant)
		m_attractiveForceConstant = attractiveForceConstant;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetAttractiveForceConstant\" in the class of SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter.\n";
		errorMessage += "Message : \"attractiveForceConstant\" is not more than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::resetRepulsiveForceConstant(const double repulsiveForceConstant)
{
	if (repulsiveForceConstant < 0.0)
		m_repulsiveForceConstant = repulsiveForceConstant;

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetRepulsiveForceConstant\" in the class of SolidStatePhysics::Crystallization::Calculators::EqualConstrainter.\n";
		errorMessage += "Message : \"repulsiveForceConstant\" is not less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::initialize()
{
	m_minimumDistance = 0.0;
	m_maximumDistance = 0.0;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::minimumDistance() const noexcept
{
	return m_minimumDistance;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::maximumDistance() const noexcept
{
	return m_maximumDistance;
}

inline double SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::getPenaltyEnergy() const
{
	if (interatomicDistance() < m_minimumDistance)
		return (m_repulsiveForceConstant * (m_minimumDistance - interatomicDistance()));

	else
	{
		if (m_maximumDistance < interatomicDistance())
			return (m_attractiveForceConstant * (interatomicDistance() - m_maximumDistance));
		else
			return 0.0;
	}
}

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::applyForce()
{
	if (interatomicDistance() < m_minimumDistance)
	{
		double forceScalingMultiplier = m_repulsiveForceConstant / interatomicDistance();
		applyForce(forceScalingMultiplier);
	}

	else
	{
		if (m_maximumDistance < interatomicDistance())
		{
			double forceScalingMultiplier = m_attractiveForceConstant / interatomicDistance();
			applyForce(forceScalingMultiplier);
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::renewMinMaxDistance(const double minimumDistance, const double maximumDistance)
{
	if (minimumDistance < 0.0 || maximumDistance < 0.0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"renewMinMaxDistance\" in the class of SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter.\n";
		errorMessage += "Message : \"minimumDistance\" and/or \"maximumDistance\" is less than zero.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		if (maximumDistance < minimumDistance)
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"renewMinMaxDistance\" in the class of SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter.\n";
			errorMessage += "Message : \"maximumDistance\" is less than \"minimumDistance\".\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}

		else
		{
			m_minimumDistance = minimumDistance;
			m_maximumDistance = maximumDistance;

			std::shared_ptr<MinMaxConstrainter> twinMinMaxConstrainterPtr = std::dynamic_pointer_cast<MinMaxConstrainter>(twinPtr());
			twinMinMaxConstrainterPtr->m_minimumDistance = m_minimumDistance;
			twinMinMaxConstrainterPtr->m_maximumDistance = m_maximumDistance;
		}
	}
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void SolidStatePhysics::Crystallization::Calculators::MinMaxConstrainter::applyForce(const double forceScalingMultiplier)
{
	ownerAtom().receiveForce(forceScalingMultiplier, displacement());
	attachedAtom().entity().receiveForce((-1.0 * forceScalingMultiplier), displacement());

	NumericalAnalysis::LinearAlgebra::NumericMatrix additiveTransformationTensor(0.0, 3, 3);
	{
		NumericVector fractionalDisplacement = crystalLattice().toFractionalCoordinate(displacement());
		additiveTransformationTensor[0] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[0]);
		additiveTransformationTensor[1] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[1]);
		additiveTransformationTensor[2] -= displacement() * (forceScalingMultiplier * fractionalDisplacement[2]);

		crystalLattice().addTransformationTensor(1.0, additiveTransformationTensor);
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_MINMAXCONSTRAINTER_H
