#ifndef SOLIDSTATEPHYSICS_MATERIALS_MOLECULARSTRUCTURE_H
#define SOLIDSTATEPHYSICS_MATERIALS_MOLECULARSTRUCTURE_H


namespace SolidStatePhysics
{
	namespace Materials
	{
		/*
		Molecule
		*/

		template <typename A, typename AA = std::vector<A>>
		class MolecularStructure
		{
		public:
			using size_type = std::size_t;
			using Atom = A;
			using AtomicArrangement = AA;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors and destructor

		public:
			MolecularStructure() = default;
			explicit MolecularStructure(const std::string& name);
			MolecularStructure(const AtomicArrangement& atomicArrangement, const std::string& name = std::string{});
			MolecularStructure(AtomicArrangement&& atomicArrangement, const std::string& name = std::string{});

			virtual ~MolecularStructure() = default;

			MolecularStructure(const MolecularStructure&) = default;
			MolecularStructure(MolecularStructure&&) noexcept = default;
			MolecularStructure& operator=(const MolecularStructure&) = default;
			MolecularStructure& operator=(MolecularStructure&&) noexcept = default;

		// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access functions

			const AtomicArrangement& atomicArrangement() const noexcept;
			virtual void resetAtomicArrangement(const AtomicArrangement& atomicArrangement);
			virtual void resetAtomicArrangement(AtomicArrangement&& atomicArrangement);

			std::string name() const;
			void rename(const std::string& name = std::string{}) noexcept;

		// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			void erase(const size_type trashAtomicNumber);
			virtual void sort();
			virtual void clear() noexcept;

			virtual bool isValid() const noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Crystal Information

			bool isCoordinated(const size_type atomicNumber) const noexcept;
			size_type countAtoms(const size_type atomicNumber) const noexcept;
			size_type countTotalAtoms() const noexcept;
			size_type countElementSpecies() const noexcept;

			ChemicalConstituents chemicalConstituents() const noexcept;

		// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::string m_name;
			AtomicArrangement m_atomicArrangement;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

template <typename A, typename AA>
inline SolidStatePhysics::Materials::MolecularStructure<A, AA>::MolecularStructure(const std::string& name)
	: m_name{ name }
	, m_atomicArrangement{}
{
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::MolecularStructure<A, AA>::MolecularStructure(const AtomicArrangement& atomicArrangement, const std::string& name)
	: m_name{ name }
	, m_atomicArrangement{ atomicArrangement }
{
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::MolecularStructure<A, AA>::MolecularStructure::MolecularStructure(AtomicArrangement&& atomicArrangement, const std::string& name)
	: m_name{ name }
	, m_atomicArrangement{ std::move(atomicArrangement) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <typename A, typename AA>
inline const typename SolidStatePhysics::Materials::MolecularStructure<A, AA>::AtomicArrangement& SolidStatePhysics::Materials::MolecularStructure<A, AA>::atomicArrangement() const noexcept
{
	return m_atomicArrangement;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::resetAtomicArrangement(const AtomicArrangement& atomicArrangement)
{
	if (atomicArrangement.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetAtomicArrangement\" in the class of SolidStatePhysics::Materials::MolecularStructure.\n";
		errorMessage += "Message : \"atomicArrangement\" is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
		m_atomicArrangement = atomicArrangement;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::resetAtomicArrangement(AtomicArrangement&& atomicArrangement)
{
	if (atomicArrangement.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"resetAtomicArrangement\" in the class of SolidStatePhysics::Materials::MolecularStructure.\n";
		errorMessage += "Message : \"atomicArrangement\" is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
		m_atomicArrangement = std::move(atomicArrangement);
}

template <typename A, typename AA>
inline std::string SolidStatePhysics::Materials::MolecularStructure<A, AA>::name() const
{
	if (m_name.empty())
	{
		if (m_atomicArrangement.empty())
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"name\" in the class of SolidStatePhysics::Materials::MolecularStructure.\n";
			errorMessage += "Message : \"m_atomicArrangement\" is empty.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}

		else
			return chemicalConstituents().callName();
	}

	else
		return m_name;
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::rename(const std::string& name) noexcept
{
	if (name.empty())
		m_name.clear();
	else
		m_name = name;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::erase(const size_type trashAtomicNumber)
{
	for (auto iter = m_atomicArrangement.begin(); iter != m_atomicArrangement.end();)
	{
		if (iter->atomicNumber() == trashAtomicNumber)
			iter = m_atomicArrangement.erase(iter);
		else
			++iter;
	}
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::sort()
{
	std::sort(m_atomicArrangement.begin(), m_atomicArrangement.end(), [](const Atom& aa, const Atom& ab) { return (aa.atomicNumber() < ab.atomicNumber()); });
}

template <typename A, typename AA>
inline void SolidStatePhysics::Materials::MolecularStructure<A, AA>::clear() noexcept
{
	m_name.clear();
	m_atomicArrangement.clear();
}

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::MolecularStructure<A, AA>::isValid() const noexcept
{
	return !(m_atomicArrangement.empty());
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Crystal Information

template <typename A, typename AA>
inline bool SolidStatePhysics::Materials::MolecularStructure<A, AA>::isCoordinated(const size_type atomicNumber) const noexcept
{
	for (const auto& coordinatedAtom : m_atomicArrangement)
	{
		if (coordinatedAtom.atomicNumber() == atomicNumber)
			return true;
	}

	return false;
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::MolecularStructure<A, AA>::countAtoms(const size_type atomicNumber) const noexcept
{
	size_type atomCounting = 0;

	for (const auto& coordinatedAtom : m_atomicArrangement)
	{
		if (coordinatedAtom.atomicNumber() == atomicNumber)
			++atomCounting;
	}

	return atomCounting;
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::MolecularStructure<A, AA>::countTotalAtoms() const noexcept
{
	return m_atomicArrangement.size();
}

template <typename A, typename AA>
inline std::size_t SolidStatePhysics::Materials::MolecularStructure<A, AA>::countElementSpecies() const noexcept
{
	std::map<size_type, size_type> constituents;

	for (const auto& coordinatedAtom : m_atomicArrangement)
	{
		std::map<size_type, size_type>::iterator iter = constituents.find(coordinatedAtom.atomicNumber());

		if (iter == constituents.end())
			constituents.emplace(coordinatedAtom.atomicNumber(), 1);
	}

	return constituents.size();
}

template <typename A, typename AA>
inline SolidStatePhysics::Materials::ChemicalConstituents SolidStatePhysics::Materials::MolecularStructure<A, AA>::chemicalConstituents() const noexcept
{
	ChemicalConstituents chemicalConstituents;
	{
		for (const auto& atom : m_atomicArrangement)
		{
			if (!(chemicalConstituents.tryEmplaceAtoms(atom.atomicNumber(), 1)))
				chemicalConstituents.addAtoms(atom.atomicNumber());
		}
	}

	return chemicalConstituents;
}

// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_MATERIALS_MOLECULARSTRUCTURE_H
