﻿#ifndef NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTOR_H
#define NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTOR_H

#include <valarray>
#include <string>
#include <stdexcept>

#include <mkl_cblas.h>
#include <mkl_trans.h>
#include <mkl_vml.h>

#include "NumericMatrix.h"


namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		class NumericVectorSlice;

		class NumericVector final
		{
		public:
			using size_type = std::size_t;
			using value_type = double;
			using pointer = value_type*;
			using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors and destructor

		public:
			NumericVector() = default;
			NumericVector(std::initializer_list<value_type>);
			explicit NumericVector(const size_type);
			NumericVector(const value_type, const size_type);

			NumericVector(const NumericVector&);
			NumericVector(const NumericVectorSlice&);
			NumericVector(NumericVector&&) noexcept = default;

			~NumericVector() {}

		// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Arithmetic functions

			NumericVector& add(const value_type scalingMultiplier, const NumericVector&) noexcept;
			NumericVector& add(const value_type scalingMultiplier, const NumericVectorSlice&) noexcept;
			NumericVector& add(const value_type scalingMultiplier, const NumericVector&, const value_type selfScalingMultiplier) noexcept;
			NumericVector& add(const value_type scalingMultiplier, const NumericVectorSlice&, const value_type selfScalingMultiplier) noexcept;

			template<typename matrix_transform_tag = numeric_matrix_identity_tag>
			NumericVector& add(const value_type scalingMultiplier, const NumericMatrix&, const NumericVector&, const value_type selfScalingMultiplier = 1.0) noexcept;

			template<typename matrix_transform_tag = numeric_matrix_identity_tag>
			NumericVector& transform(const NumericMatrix&, const value_type scalingMultiplier = 1.0);

		// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			NumericVector& operator=(const value_type scalar) noexcept;
			NumericVector& operator=(const NumericVector&) noexcept;
			NumericVector& operator=(const NumericVectorSlice&) noexcept;
			NumericVector& operator=(NumericVector&&) noexcept = default;

			NumericVector& operator+=(const value_type scalar) noexcept;
			NumericVector& operator+=(const NumericVector&) noexcept;
			NumericVector& operator+=(const NumericVectorSlice&) noexcept;
			NumericVector& operator-=(const value_type scalar) noexcept;
			NumericVector& operator-=(const NumericVector&) noexcept;
			NumericVector& operator-=(const NumericVectorSlice&) noexcept;
			NumericVector& operator*=(const value_type scalingMultiplier) noexcept;
			NumericVector& operator*=(const NumericVector&) noexcept;
			NumericVector& operator*=(const NumericVectorSlice&) noexcept;
			NumericVector& operator/=(const value_type scalingDivider) noexcept;
			NumericVector& operator/=(const NumericVector&) noexcept;
			NumericVector& operator/=(const NumericVectorSlice&) noexcept;

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access functions

			value_type operator[](const size_type index) const noexcept;
			value_type& operator[](const size_type index) noexcept;
			value_type at(const size_type index) const;
			value_type& at(const size_type index);

			pointer begin() noexcept;
			const_pointer begin() const noexcept;
			const_pointer cbegin() const noexcept;
			pointer end() noexcept;
			const_pointer end() const noexcept;
			const_pointer cend() const noexcept;

		// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Slice functions

			NumericVector operator[](const std::slice&) const noexcept;
			NumericVectorSlice operator[](const std::slice&) noexcept;

		// Slice functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			size_type size() const noexcept;
			void resize(const size_type, const value_type value = value_type{}) noexcept;
			void clear() noexcept;
			bool empty() const noexcept;
			void swap(NumericVector&) noexcept;

			double norm() const noexcept;
			double normSquare() const noexcept;

			value_type max();
			value_type min();

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends and private methods

			friend class NumericMatrix;

		// Friends and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************


		private:
			std::valarray<value_type> m_numericVector;
		};
	}
}


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic functions

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::add(const value_type scalingMultiplier, const NumericVector& numericVector) noexcept
{
	cblas_daxpy(m_numericVector.size(), scalingMultiplier, numericVector.begin(), 1, std::begin(m_numericVector), 1);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::add(const value_type scalingMultiplier, const NumericVector& numericVector, const value_type selfScalingMultiplier) noexcept
{
	cblas_daxpby(m_numericVector.size(), scalingMultiplier, numericVector.begin(), 1, selfScalingMultiplier, std::begin(m_numericVector), 1);
	return *this;
}

template<typename matrix_transform_tag>
inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::add(const value_type scalingMultiplier, const NumericMatrix& numericMatrix, const NumericVector& numericVector, const value_type selfScalingMultiplier) noexcept
{
	constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_tag>();
	cblas_dgemv(CblasColMajor, transposeType, numericMatrix.columnSize(), numericMatrix.rowSize(), scalingMultiplier, numericMatrix.begin(), numericMatrix.columnSize(), numericVector.begin(), 1, selfScalingMultiplier, std::begin(m_numericVector), 1);
	return *this;
}

template<typename matrix_transform_tag>
inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::transform(const NumericMatrix& numericMatrix, const value_type scalingMultiplier)
{
	constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_tag>();
	std::valarray<value_type> result(m_numericVector.size());
	cblas_dgemv(CblasColMajor, transposeType, numericMatrix.columnSize(), numericMatrix.rowSize(), scalingMultiplier, numericMatrix.begin(), numericMatrix.columnSize(), std::begin(m_numericVector), 1, 0.0, std::begin(result), 1);
	m_numericVector.swap(result);
	return *this;
}

// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator=(const value_type scalar) noexcept
{
	m_numericVector = scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator=(const NumericVector& numericVector) noexcept
{
	m_numericVector.resize(numericVector.size());
	cblas_dcopy(m_numericVector.size(), numericVector.begin(), 1, std::begin(m_numericVector), 1);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator+=(const value_type scalar) noexcept
{
	m_numericVector += scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator+=(const NumericVector& numericVector) noexcept
{
	return add(1.0, numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator-=(const value_type scalar) noexcept
{
	m_numericVector -= scalar;
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator-=(const NumericVector& numericVector) noexcept
{
	return add(-1.0, numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator*=(const value_type scalingMultiplier) noexcept
{
	cblas_dscal(m_numericVector.size(), scalingMultiplier, std::begin(m_numericVector), 1);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator*=(const NumericVector& numericVector) noexcept
{
	std::valarray<value_type> temp = std::move(m_numericVector);
	m_numericVector.resize(temp.size());
	vmdMul(m_numericVector.size(), std::begin(temp), numericVector.begin(), std::begin(m_numericVector), VML_HA);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator/=(const value_type scalingDivider) noexcept
{
	cblas_dscal(m_numericVector.size(), (1.0 / scalingDivider), std::begin(m_numericVector), 1);
	return *this;
}

inline NumericalAnalysis::LinearAlgebra::NumericVector& NumericalAnalysis::LinearAlgebra::NumericVector::operator/=(const NumericVector& numericVector) noexcept
{
	std::valarray<value_type> temp = std::move(m_numericVector);
	m_numericVector.resize(temp.size());
	vmdDiv(m_numericVector.size(), std::begin(temp), numericVector.begin(), std::begin(m_numericVector), VML_HA);
	return *this;
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access functions

inline double NumericalAnalysis::LinearAlgebra::NumericVector::operator[](size_type index) const noexcept
{
	return m_numericVector[index];
}

inline double& NumericalAnalysis::LinearAlgebra::NumericVector::operator[](size_type index) noexcept
{
	return m_numericVector[index];
}

inline double NumericalAnalysis::LinearAlgebra::NumericVector::at(size_type index) const
{
	if (index < m_numericVector.size())
	{
		return m_numericVector[index];
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Out of range\n";
		errorMessage += "Location: The function of \"at\" in the class of NumericalAnalysis::LinearAlgebra::NumericVector.\n";
		errorMessage += "Message : The index is out of range.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::out_of_range(errorMessage);
	}
}

inline double& NumericalAnalysis::LinearAlgebra::NumericVector::at(size_type index)
{
	if (index < m_numericVector.size())
	{
		return m_numericVector[index];
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Out of range\n";
		errorMessage += "Location: The function of \"at\" in the class of NumericalAnalysis::LinearAlgebra::NumericVector.\n";
		errorMessage += "Message : The index is out of range.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::out_of_range(errorMessage);
	}
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::pointer NumericalAnalysis::LinearAlgebra::NumericVector::begin() noexcept
{
	return std::begin(m_numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::const_pointer NumericalAnalysis::LinearAlgebra::NumericVector::begin() const noexcept
{
	return std::begin(m_numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::const_pointer NumericalAnalysis::LinearAlgebra::NumericVector::cbegin() const noexcept
{
	return std::begin(m_numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::pointer NumericalAnalysis::LinearAlgebra::NumericVector::end() noexcept
{
	return std::end(m_numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::const_pointer NumericalAnalysis::LinearAlgebra::NumericVector::end() const noexcept
{
	return std::end(m_numericVector);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::const_pointer NumericalAnalysis::LinearAlgebra::NumericVector::cend() const noexcept
{
	return std::end(m_numericVector);
}

// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Slice functions

inline NumericalAnalysis::LinearAlgebra::NumericVector NumericalAnalysis::LinearAlgebra::NumericVector::operator[](const std::slice& sl) const noexcept
{
	NumericVector numericVector(sl.size());
	cblas_dcopy(numericVector.size(), (std::begin(m_numericVector) + sl.start()), sl.stride(), std::begin(numericVector), 1);
	return numericVector;
}

// Slice functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline NumericalAnalysis::LinearAlgebra::NumericVector::size_type NumericalAnalysis::LinearAlgebra::NumericVector::size() const noexcept
{
	return m_numericVector.size();
}

inline void NumericalAnalysis::LinearAlgebra::NumericVector::resize(const size_type size, const value_type value) noexcept
{
	m_numericVector.resize(size, value);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVector::clear() noexcept
{
	std::valarray<value_type> va;
	m_numericVector.swap(va);
}

inline bool NumericalAnalysis::LinearAlgebra::NumericVector::empty() const noexcept
{
	return !(static_cast<bool>(m_numericVector.size()));
}

inline void NumericalAnalysis::LinearAlgebra::NumericVector::swap(NumericVector& numericVector) noexcept
{
	std::swap(m_numericVector, numericVector.m_numericVector);
}

inline double NumericalAnalysis::LinearAlgebra::NumericVector::norm() const noexcept
{
	return cblas_dnrm2(m_numericVector.size(), std::begin(m_numericVector), 1);
}

inline double NumericalAnalysis::LinearAlgebra::NumericVector::normSquare() const noexcept
{
	return cblas_ddot(m_numericVector.size(), std::begin(m_numericVector), 1, std::begin(m_numericVector), 1);
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::value_type NumericalAnalysis::LinearAlgebra::NumericVector::max()
{
	size_type index = static_cast<size_type>(cblas_idamax(m_numericVector.size(), std::begin(m_numericVector), 1));

	if (index == 0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The function of \"max\" in the class of NumericalAnalysis::LinearAlgebra::NumericVector.\n";
		errorMessage += "Message : Could not find the maximum value.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		return m_numericVector[index];
	}
}

inline NumericalAnalysis::LinearAlgebra::NumericVector::value_type NumericalAnalysis::LinearAlgebra::NumericVector::min()
{
	size_type index = static_cast<size_type>(cblas_idamin(m_numericVector.size(), std::begin(m_numericVector), 1));

	if (index == 0)
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The function of \"min\" in the class of NumericalAnalysis::LinearAlgebra::NumericVector.\n";
		errorMessage += "Message : Could not find the minimum value.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		return m_numericVector[index];
	}
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTOR_H
