#ifndef NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTORSLICE_H
#define NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTORSLICE_H

#include <valarray>

#include "NumericVector.h"
#include "NumericMatrix.h"


namespace NumericalAnalysis
{
	namespace LinearAlgebra
	{
		class NumericVectorSlice final
		{
		public:
			using size_type = std::size_t;
			using value_type = double;
			using pointer = value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors and destructor

		public:
			NumericVectorSlice(const NumericVectorSlice&) = default;
			~NumericVectorSlice() {}

		// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Arithmetic functions

			void add(const value_type scalingMultiplier, const NumericVector&) noexcept;
			void add(const value_type scalingMultiplier, const NumericVectorSlice&) noexcept;
			void add(const value_type scalingMultiplier, const NumericVector&, const value_type selfScalingMultiplier) noexcept;
			void add(const value_type scalingMultiplier, const NumericVectorSlice&, const value_type selfScalingMultiplier) noexcept;

			template<typename matrix_transform_tag = numeric_matrix_identity_tag>
			void add(const value_type scalingMultiplier, const NumericMatrix&, const NumericVector&, const value_type selfScalingMultiplier = 1.0) noexcept;

			template<typename matrix_transform_tag = numeric_matrix_identity_tag>
			void add(const value_type scalingMultiplier, const NumericMatrix&, const NumericVectorSlice&, const value_type selfScalingMultiplier = 1.0) noexcept;

			template<typename matrix_transform_tag = numeric_matrix_identity_tag>
			void transform(const NumericMatrix&, const value_type scalingMultiplier = 1.0);

		// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			const NumericVectorSlice& operator=(const NumericVectorSlice&) const;
			void operator=(const NumericVector&) const;
			void operator=(const value_type scalar) const;

			void operator+=(const NumericVector&) noexcept;
			void operator+=(const NumericVectorSlice&) noexcept;
			void operator-=(const NumericVector&) noexcept;
			void operator-=(const NumericVectorSlice&) noexcept;
			void operator*=(const value_type scalingMultiplier) noexcept;
			void operator*=(const NumericVector&) noexcept;
			void operator*=(const NumericVectorSlice&) noexcept;
			void operator/=(const value_type scalingDivider);

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access functions

			size_type start() const noexcept;
			size_type size() const noexcept;
			size_type stride() const noexcept;

		// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			double norm() const noexcept;
			double normSquare() const noexcept;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends and private methods

			friend class NumericVector;
			friend class NumericMatrix;

		private:
			NumericVectorSlice(std::valarray<value_type>&, const std::slice&);
			
			pointer begin() noexcept;
			pointer begin() const noexcept;

		// Friends and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::valarray<value_type>& m_numericVectorSlice;
			size_type m_start;
			size_type m_size;
			size_type m_stride;

		private:
			NumericVectorSlice() = delete;
			NumericVectorSlice(NumericVectorSlice&&) noexcept = delete;
			NumericVectorSlice& operator=(NumericVectorSlice&&) noexcept = delete;
		};
	}
}


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic functions

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericVector& numericVector) noexcept
{
	cblas_daxpy(m_size, scalingMultiplier, numericVector.begin(), 1, begin(), m_stride);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericVectorSlice& numericVectorSlice) noexcept
{
	cblas_daxpy(m_size, scalingMultiplier, numericVectorSlice.begin(), numericVectorSlice.m_stride, begin(), m_stride);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericVector& numericVector, const value_type selfScalingMultiplier) noexcept
{
	cblas_daxpby(m_size, scalingMultiplier, numericVector.begin(), 1, selfScalingMultiplier, begin(), m_stride);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericVectorSlice& numericVectorSlice, const value_type selfScalingMultiplier) noexcept
{
	cblas_daxpby(m_size, scalingMultiplier, numericVectorSlice.begin(), numericVectorSlice.m_stride, selfScalingMultiplier, begin(), m_stride);
}

template<typename matrix_transform_tag>
inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericMatrix& numericMatrix, const NumericVector& numericVector, const value_type selfScalingMultiplier) noexcept
{
	constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_tag>();
	cblas_dgemv(CblasColMajor, transposeType, numericMatrix.columnSize(), numericMatrix.rowSize(), scalingMultiplier, numericMatrix.begin(), numericMatrix.columnSize(), numericVector.begin(), 1, selfScalingMultiplier, begin(), m_stride);
}

template<typename matrix_transform_tag>
inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::add(const value_type scalingMultiplier, const NumericMatrix& numericMatrix, const NumericVectorSlice& numericVectorSlice, const value_type selfScalingMultiplier) noexcept
{
	constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_tag>();
	cblas_dgemv(CblasColMajor, transposeType, numericMatrix.columnSize(), numericMatrix.rowSize(), scalingMultiplier, numericMatrix.begin(), numericMatrix.columnSize(), numericVectorSlice.begin(), numericVectorSlice.m_stride, selfScalingMultiplier, begin(), m_stride);
}

template<typename matrix_transform_tag>
inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::transform(const NumericMatrix& numericMatrix, const value_type scalingMultiplier)
{
	constexpr CBLAS_TRANSPOSE transposeType = translateNumericMatrixTransposeTagToInt<matrix_transform_tag>();
	std::valarray<value_type> result(m_numericVectorSlice.size());
	cblas_dcopy(m_size, begin(), m_stride, std::begin(result), 1);
	cblas_dgemv(CblasColMajor, transposeType, numericMatrix.columnSize(), numericMatrix.rowSize(), scalingMultiplier, numericMatrix.begin(), numericMatrix.columnSize(), std::begin(result), 1, 0.0, begin(), m_stride);
}

// Arithmetic functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

inline const NumericalAnalysis::LinearAlgebra::NumericVectorSlice& NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator=(const NumericVectorSlice& numericVectorSlice) const
{
	cblas_dcopy(m_size, numericVectorSlice.begin(), numericVectorSlice.m_stride, begin(), m_stride);
	return *this;
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator=(const NumericVector& numericVector) const
{
	cblas_dcopy(m_size, numericVector.begin(), 1, begin(), m_stride);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator=(const value_type scalar) const
{
	std::slice_array<value_type> sa = m_numericVectorSlice[std::slice(m_start, m_size, m_stride)];
	sa = scalar;
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator+=(const NumericVector& numericVector) noexcept
{
	add(1.0, numericVector);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator+=(const NumericVectorSlice& numericVectorSlice) noexcept
{
	add(1.0, numericVectorSlice);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator-=(const NumericVector& numericVector) noexcept
{
	add(-1.0, numericVector);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator-=(const NumericVectorSlice& numericVectorSlice) noexcept
{
	add(-1.0, numericVectorSlice);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator*=(const value_type scalingMultiplier) noexcept
{
	cblas_dscal(m_size, scalingMultiplier, begin(), m_stride);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator*=(const NumericVector& numericVector) noexcept
{
	NumericVectorSlice temp(m_numericVectorSlice, std::slice(m_start, m_size, m_stride));
	vmdMulI(m_size, numericVector.begin(), 1, temp.begin(), m_stride, begin(), m_stride, VML_HA);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator*=(const NumericVectorSlice& numericVectorSlice) noexcept
{
	NumericVectorSlice temp(m_numericVectorSlice, std::slice(m_start, m_size, m_stride));
	vmdMulI(m_size, numericVectorSlice.begin(), numericVectorSlice.m_stride, temp.begin(), m_stride, begin(), m_stride, VML_HA);
}

inline void NumericalAnalysis::LinearAlgebra::NumericVectorSlice::operator/=(const value_type scalingDivider)
{
	if (std::isnormal(scalingDivider))
	{
		cblas_dscal(m_size, (1.0 / scalingDivider), (std::begin(m_numericVectorSlice) + m_start), m_stride);
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The function of \"operator/=\" in the class of NumericalAnalysis::LinearAlgebra::NumericVectorSlice.\n";
		errorMessage += "Message : scalingDivider is not a normalized value.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access functions

inline NumericalAnalysis::LinearAlgebra::NumericVectorSlice::size_type NumericalAnalysis::LinearAlgebra::NumericVectorSlice::start() const noexcept
{
	return m_start;
}

inline NumericalAnalysis::LinearAlgebra::NumericVectorSlice::size_type NumericalAnalysis::LinearAlgebra::NumericVectorSlice::size() const noexcept
{
	return m_size;
}

inline NumericalAnalysis::LinearAlgebra::NumericVectorSlice::size_type NumericalAnalysis::LinearAlgebra::NumericVectorSlice::stride() const noexcept
{
	return m_stride;
}

// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline double NumericalAnalysis::LinearAlgebra::NumericVectorSlice::norm() const noexcept
{
	return cblas_dnrm2(m_size, begin(), m_stride);
}

inline double NumericalAnalysis::LinearAlgebra::NumericVectorSlice::normSquare() const noexcept
{
	return cblas_ddot(m_size, begin(), m_stride, begin(), m_stride);
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline NumericalAnalysis::LinearAlgebra::NumericVectorSlice::pointer NumericalAnalysis::LinearAlgebra::NumericVectorSlice::begin() noexcept
{
	return (std::begin(m_numericVectorSlice) + m_start);
}

inline NumericalAnalysis::LinearAlgebra::NumericVectorSlice::pointer NumericalAnalysis::LinearAlgebra::NumericVectorSlice::begin() const noexcept
{
	return (std::begin(m_numericVectorSlice) + m_start);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Slice functions


#endif // !NUMERICALANALYSIS_LINEARALGEBRA_NUMERICVECTORSLICE_H
