#ifndef CONTROLCENTER_WINDOW_ORDERFORM_H
#define CONTROLCENTER_WINDOW_ORDERFORM_H

#include <string>
#include <filesystem>

#include "CrystallizationPlan.h"
#include "MonitoringPlan.h"
#include "RequestedFeedstocks.h"


namespace ControlCenter
{
	namespace Window
	{
		class Receptionist;

		class OrderForm
		{
			using CrystallizationPlan = SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan;
			using MonitoringPlan = SolidStatePhysics::Crystallization::Simulators::MonitoringPlan;
			using RequestedFeedstocks = ControlCenter::Administering::RequestedFeedstocks;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			OrderForm();
			virtual ~OrderForm() = default;

			OrderForm(const OrderForm&) = default;
			OrderForm(OrderForm&&) noexcept = default;
			OrderForm& operator=(const OrderForm&) = default;
			OrderForm& operator=(OrderForm&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			std::string orderName() const noexcept;
			std::size_t maxThreading() const noexcept;
			std::size_t mklThreading() const noexcept;
			const std::filesystem::path& outputDirectoryPath() const;
			const std::filesystem::path& workDirectoryPath() const;

			const CrystallizationPlan& crystallizationPlan() const noexcept;
			const MonitoringPlan& monitoringPlan() const noexcept;
			const RequestedFeedstocks& requestedFeedstocks() const;

			bool shouldGenerateMaterials() const noexcept;
			bool shouldReheatMaterials() const noexcept;

			bool isValid() const noexcept;
			void clear() noexcept;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods for friends

		private:
			void resetOrderName(const std::string& orderName = std::string{}) noexcept;
			void resetExecutionMode(const std::string& executionMode = std::string{});
			void resetMaxThreading(const std::size_t maxThreading = 0);
			void resetMklThreading(const std::size_t mklThreading = 0);
			void resetOutputDirectoryPath(const std::filesystem::path& outputDirectoryPath);
			void resetWorkDirectoryPath(const std::filesystem::path& workDirectoryPath);

			void resetCrystallizationPlan(const CrystallizationPlan&) noexcept;
			void resetCrystallizationPlan(CrystallizationPlan&&) noexcept;
			void resetMonitoringPlan(const MonitoringPlan&) noexcept;
			void resetMonitoringPlan(MonitoringPlan&&) noexcept;
			void resetRequestedFeedstocks(const RequestedFeedstocks&);
			void resetRequestedFeedstocks(RequestedFeedstocks&&);

		// Private methods for friends
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::string m_orderName;
			std::string m_executionMode; // GENERATE, REHEAT
			std::size_t m_maxThreading;
			std::size_t m_mklThreading;
			std::filesystem::path m_outputDirectoryPath;
			std::filesystem::path m_workDirectoryPath;

			CrystallizationPlan m_crystallizationPlan;
			MonitoringPlan m_monitoringPlan;
			RequestedFeedstocks m_requestedFeedstocks;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends

			friend class Receptionist;

		// Friends
// **********************************************************************************************************************************************************************************************************************************************************************************************

		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::size_t ControlCenter::Window::OrderForm::maxThreading() const noexcept
{
	return m_maxThreading;
}

inline std::size_t ControlCenter::Window::OrderForm::mklThreading() const noexcept
{
	return m_mklThreading;
}

inline const std::filesystem::path& ControlCenter::Window::OrderForm::outputDirectoryPath() const
{
	if (m_outputDirectoryPath.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"outputDirectoryPath\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : \"m_outputDirectoryPath\" is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}

	else
		return m_outputDirectoryPath;
}

inline const std::filesystem::path& ControlCenter::Window::OrderForm::workDirectoryPath() const
{
	if (m_workDirectoryPath.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"workDirectoryPath\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : \"m_workDirectoryPath\" is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}

	else
		return m_workDirectoryPath;
}

inline const SolidStatePhysics::Crystallization::Simulators::CrystallizationPlan& ControlCenter::Window::OrderForm::crystallizationPlan() const noexcept
{
	return m_crystallizationPlan;
}

inline const SolidStatePhysics::Crystallization::Simulators::MonitoringPlan& ControlCenter::Window::OrderForm::monitoringPlan() const noexcept
{
	return m_monitoringPlan;
}

inline const ControlCenter::Administering::RequestedFeedstocks& ControlCenter::Window::OrderForm::requestedFeedstocks() const
{
	if (m_executionMode == "GENERATE")
	{
		if (m_requestedFeedstocks.isValid())
			return m_requestedFeedstocks;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Logic error\n";
			errorMessage += "Location: The member function of \"requestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
			errorMessage += "Message : \"m_requestedFeedstocks\" is invalid.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::logic_error(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"requestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : The execution mode is not \"GENERATE\".\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline bool ControlCenter::Window::OrderForm::shouldGenerateMaterials() const noexcept
{
	if (m_executionMode == "GENERATE")
		return true;
	else
		return false;
}

inline bool ControlCenter::Window::OrderForm::shouldReheatMaterials() const noexcept
{
	if (m_executionMode == "REHEAT")
		return true;
	else
		return false;
}

inline bool ControlCenter::Window::OrderForm::isValid() const noexcept
{
	{
		if (m_executionMode == "GENERATE" && !(m_requestedFeedstocks.isValid()))
			return false;

		if (m_outputDirectoryPath.empty())
			return false;

		if (m_workDirectoryPath.empty())
			return false;
	}

	return true;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void ControlCenter::Window::OrderForm::resetExecutionMode(const std::string& executionMode)
{
	if (executionMode.empty())
		m_executionMode = "GENERATE";

	else
	{
		if (executionMode == "GENERATE" || executionMode == "Generate" || executionMode == "generate")
			m_executionMode = "GENERATE";

		else if (executionMode == "REHEAT" || executionMode == "Reheat" || executionMode == "reheat")
			m_executionMode = "REHEAT";

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetExecutionMode\" in the class of ControlCenter::Window::OrderForm.\n";
			errorMessage += "Message : \"executionMode\" is invalid.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}
}

inline void ControlCenter::Window::OrderForm::resetOutputDirectoryPath(const std::filesystem::path& outputDirectoryPath)
{
	if (outputDirectoryPath.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"setOutputDirectoryPath\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : outputDirectoryPath is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		m_outputDirectoryPath = outputDirectoryPath;
		m_outputDirectoryPath.make_preferred();
	}
}

inline void ControlCenter::Window::OrderForm::resetWorkDirectoryPath(const std::filesystem::path& workDirectoryPath)
{
	if (workDirectoryPath.empty())
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Invalid argument\n";
		errorMessage += "Location: The member function of \"setWorkDirectoryPath\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : workDirectoryPath is empty.\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::invalid_argument(errorMessage);
	}

	else
	{
		m_workDirectoryPath = workDirectoryPath;
		m_workDirectoryPath.make_preferred();
	}
}

inline void ControlCenter::Window::OrderForm::resetCrystallizationPlan(const CrystallizationPlan& crystallizationPlan) noexcept
{
	m_crystallizationPlan = crystallizationPlan;
}

inline void ControlCenter::Window::OrderForm::resetCrystallizationPlan(CrystallizationPlan&& crystallizationPlan) noexcept
{
	m_crystallizationPlan = std::move(crystallizationPlan);
}

inline void ControlCenter::Window::OrderForm::resetMonitoringPlan(const MonitoringPlan& monitoringPlan) noexcept
{
	m_monitoringPlan = monitoringPlan;
}

inline void ControlCenter::Window::OrderForm::resetMonitoringPlan(MonitoringPlan&& monitoringPlan) noexcept
{
	m_monitoringPlan = std::move(monitoringPlan);
}

inline void ControlCenter::Window::OrderForm::resetRequestedFeedstocks(const RequestedFeedstocks& requestedFeedstocks)
{
	if (m_executionMode == "GENERATE")
	{
		if (requestedFeedstocks.isValid())
			m_requestedFeedstocks = requestedFeedstocks;

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetRequestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
			errorMessage += "Message : \"requestedFeedstocks\" is invalid.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"resetRequestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : The execution mode is not \"GENERATE\".\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

inline void ControlCenter::Window::OrderForm::resetRequestedFeedstocks(RequestedFeedstocks&& requestedFeedstocks)
{
	if (m_executionMode == "GENERATE")
	{
		if (requestedFeedstocks.isValid())
			m_requestedFeedstocks = std::move(requestedFeedstocks);

		else
		{
			std::string errorMessage = "------------------------------------------------------\n\n";
			errorMessage += "Error   : Invalid argument\n";
			errorMessage += "Location: The member function of \"resetRequestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
			errorMessage += "Message : \"requestedFeedstocks\" is invalid.\n\n";
			errorMessage += "------------------------------------------------------\n\n";
			throw std::invalid_argument(errorMessage);
		}
	}

	else
	{
		std::string errorMessage = "------------------------------------------------------\n\n";
		errorMessage += "Error   : Logic error\n";
		errorMessage += "Location: The member function of \"resetRequestedFeedstocks\" in the class of ControlCenter::Window::OrderForm.\n";
		errorMessage += "Message : The execution mode is not \"GENERATE\".\n\n";
		errorMessage += "------------------------------------------------------\n\n";
		throw std::logic_error(errorMessage);
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CONTROLCENTER_WINDOW_ORDERFORM_H
