#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_POLYHEDRALCONSTRAINTER_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_POLYHEDRALCONSTRAINTER_H

#include <string>
#include <stdexcept>
#include <functional>
#include <vector>
#include <set>
#include <utility>

#include "ConstraintOwners.h"
#include "MinimumConstrainter.h"
#include "IntelligentAtom.h"

#include "UnitConversion.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Components
		{
			class PolyhedralConstrainter :public SolidStatePhysics::Crystallization::Calculators::MinimumConstrainter
			{
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			public:
				PolyhedralConstrainter(PolyhedralConstraintOwner& realOwner, const PolyhedralConstraintOwnerAdaptor& attachedOwner);
				PolyhedralConstrainter(PolyhedralConstraintOwner& realOwner, PolyhedralConstraintOwnerAdaptor&& attachedOwner);
				explicit PolyhedralConstrainter(const std::shared_ptr<PolyhedralConstrainter>& twinPtr);

				virtual ~PolyhedralConstrainter();

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static void create(PolyhedralConstraintOwner& realOwner, const PolyhedralConstraintOwnerAdaptor& attachedOwner);
				static void erase(std::shared_ptr<PolyhedralConstrainter>&& polyhedralConstrainterPtr);
				static bool canCreate(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept;

				static void resetVertexSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;
				static void resetEdgeSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;
				static void resetFaceSharingCationicPairs(const std::set<std::pair<std::string, std::string>>&) noexcept;

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void renew() override;

				bool isFeasible(const double maxOverlapErrorRate) const noexcept;
				bool isAnionicConstrainter() const noexcept;
				bool isCationicConstrainter() const noexcept;
				bool isCommonVertexSharing() const noexcept;
				bool isCommonEdgeSharing() const noexcept;
				bool isCommonFaceSharing() const noexcept;
				size_type commonBridgingAtomsSharing() const noexcept;

				void shareCommonBridgingAtom(const AtomicAdaptor& commonBridgingAtom) noexcept;
				void removeExcessCommonBridgingAtoms(PeriodicLiquid& ownerPeriodicLiquid);

				PolyhedralConstraintOwner& realOwner() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void synchronizeCommonBridgingAtoms() noexcept;
				double getDistanceToFace(const double cationicRadius, const double minAnionicRadius, const size_type maxPossibleBridgingAtomsSharing) const noexcept;

				static size_type getMaxPossibleBridgingAtomsSharing(const InteractingAtom&, const InteractingAtom&) noexcept;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::reference_wrapper<PolyhedralConstraintOwner> m_realOwner;
				PolyhedralConstraintOwnerAdaptor m_attachedOwner;
				std::vector<AtomicAdaptor> m_commonBridgingAtoms;
				bool m_isAnionicConstrainter;
				size_type m_maxPossibleBridgingAtomsSharing;

				static std::set<std::pair<std::string, std::string>> m_vertexSharingCationicPairs;
				static std::set<std::pair<std::string, std::string>> m_edgeSharingCationicPairs;
				static std::set<std::pair<std::string, std::string>> m_faceSharingCationicPairs;


			private:
				PolyhedralConstrainter() = delete;
				PolyhedralConstrainter(const PolyhedralConstrainter&) = delete;
				PolyhedralConstrainter(PolyhedralConstrainter&&) noexcept = delete;
				PolyhedralConstrainter& operator=(const PolyhedralConstrainter&) = delete;
				PolyhedralConstrainter& operator=(PolyhedralConstrainter&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::create(PolyhedralConstraintOwner& realOwner, const PolyhedralConstraintOwnerAdaptor& attachedOwner)
{
	InteratomicConnector::createInteratomicComponent(realOwner, attachedOwner);
}

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::erase(std::shared_ptr<PolyhedralConstrainter>&& polyhedralConstrainterPtr)
{
	InteratomicConnector::eraseInteratomicComponent(std::move(polyhedralConstrainterPtr));
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::canCreate(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept
{
	if (ownerAtom.isAnion() && attachedAtom.entity().isAnion())
		return true;

	else
	{
		if (ownerAtom.isCation() && attachedAtom.entity().isCation())
			return true;
		else
			return false;
	}
}


inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetVertexSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& vertexSharingCationicPairs) noexcept
{
	m_vertexSharingCationicPairs = vertexSharingCationicPairs;
}

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetEdgeSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& edgeSharingCationicPairs) noexcept
{
	m_edgeSharingCationicPairs = edgeSharingCationicPairs;
}

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::resetFaceSharingCationicPairs(const std::set<std::pair<std::string, std::string>>& faceSharingCationicPairs) noexcept
{
	m_faceSharingCationicPairs = faceSharingCationicPairs;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::renew()
{
	if (isAnionicConstrainter())
		renewMinimumDistance(ownerAtom().maximumIonicRadius() + attachedAtom().entity().maximumIonicRadius());

	else
	{
		if (1 == m_maxPossibleBridgingAtomsSharing)
			renewMinimumDistance(ownerAtom().metallicRadius() + attachedAtom().entity().metallicRadius());

		else
		{
			double minAnionicRadius = std::numeric_limits<double>::max();
			{
				for (const auto& commonBridgingAtom : m_commonBridgingAtoms)
				{
					if (commonBridgingAtom.entity().minimumIonicRadius() < minAnionicRadius)
						minAnionicRadius = commonBridgingAtom.entity().minimumIonicRadius();
				}
			}

			double ownerDistanceToFace = getDistanceToFace(ownerAtom().maximumIonicRadius(), minAnionicRadius, m_maxPossibleBridgingAtomsSharing);
			double attachedDistanceToFace = getDistanceToFace(attachedAtom().entity().maximumIonicRadius(), minAnionicRadius, m_maxPossibleBridgingAtomsSharing);

			renewMinimumDistance(0.95 * (ownerDistanceToFace + attachedDistanceToFace));
		}
	}
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isFeasible(const double maxOverlapErrorRate) const noexcept
{
	double minBondLength = (1.0 - maxOverlapErrorRate) * minimumDistance();

	if (interatomicDistance() < minBondLength)
		return false;
	else
		return true;
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isAnionicConstrainter() const noexcept
{
	return m_isAnionicConstrainter;
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isCationicConstrainter() const noexcept
{
	return !(m_isAnionicConstrainter);
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isCommonVertexSharing() const noexcept
{
	if (m_commonBridgingAtoms.size() == 1)
		return true;
	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isCommonEdgeSharing() const noexcept
{
	if (m_commonBridgingAtoms.size() == 2)
		return true;
	else
		return false;
}

inline bool SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::isCommonFaceSharing() const noexcept
{
	if (2 < m_commonBridgingAtoms.size())
		return true;
	else
		return false;
}

inline SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::size_type SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::commonBridgingAtomsSharing() const noexcept
{
	return m_commonBridgingAtoms.size();
}

inline void SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::shareCommonBridgingAtom(const AtomicAdaptor& commonBridgingAtom) noexcept
{
	m_commonBridgingAtoms.push_back(commonBridgingAtom);

	AtomicAdaptor twinCommonBridgingAtom = commonBridgingAtom;
	twinCommonBridgingAtom.resetLatticePoint(attachedAtom().getRelativeLatticePointTo(commonBridgingAtom));

	std::shared_ptr<PolyhedralConstrainter> twinPolyhedralConstrainterPtr = std::dynamic_pointer_cast<PolyhedralConstrainter>(twinPtr());
	twinPolyhedralConstrainterPtr->m_commonBridgingAtoms.push_back(twinCommonBridgingAtom);
}

inline SolidStatePhysics::Crystallization::Components::PolyhedralConstraintOwner& SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter::realOwner() const noexcept
{
	return m_realOwner.get();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_COMPONENTS_POLYHEDRALCONSTRAINTER_H
