#ifndef SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_TRANSLATINGADAPTOR_H
#define SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_TRANSLATINGADAPTOR_H

#include <array>
#include <string>
#include <functional>

#include "LinearAlgebra.h"
#include "CrystalLattice.h"


namespace SolidStatePhysics
{
	namespace Crystallization
	{
		namespace Calculators
		{
			template <typename C>
			class TranslatingAdaptor
			{
			protected:
				using NumericVector = NumericalAnalysis::LinearAlgebra::NumericVector;

			public:
				using size_type = std::size_t;
				using CrystalLattice = SolidStatePhysics::Materials::CrystalLattice;
				using LatticePoint = std::array<int, 3>;
				using Component = C;

				struct Hasher
				{
					std::size_t operator()(const TranslatingAdaptor& translatingAdaptor) const noexcept { return translatingAdaptor.hashValue(); }
				};

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

				TranslatingAdaptor(Component& component, CrystalLattice& crystalLattice);
				TranslatingAdaptor(Component& component, CrystalLattice& crystalLattice, const LatticePoint& latticePoint);

				template <typename AC>
				TranslatingAdaptor(Component& component, const TranslatingAdaptor<AC>&);

				virtual ~TranslatingAdaptor() = default;

				TranslatingAdaptor(const TranslatingAdaptor&) = default;
				TranslatingAdaptor(TranslatingAdaptor&&) noexcept;
				TranslatingAdaptor& operator=(const TranslatingAdaptor&) = default;
				TranslatingAdaptor& operator=(TranslatingAdaptor&&) noexcept;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access functions

				Component& entity() const noexcept;
				void resetEntity(Component&) noexcept;

				CrystalLattice& crystalLattice() const noexcept;
				void resetCrystalLattice(CrystalLattice&) noexcept;

				LatticePoint latticePoint() const noexcept;
				void resetLatticePoint(const LatticePoint& latticePoint) noexcept;
				void reverseLatticePoint() noexcept;
				void translateLatticePoint(const LatticePoint& latticePoint) noexcept;

				size_type hashValue() const noexcept;

			// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility
				
				NumericVector translationVector() const noexcept;

				bool isInOriginalCell() const noexcept;
				bool isSameEntity(const Component&) const noexcept;

				template <typename AC>
				bool isSameLatticePoint(const TranslatingAdaptor<AC>&) const noexcept;

				template <typename AC>
				LatticePoint getRelativeLatticePointTo(const TranslatingAdaptor<AC>&) const noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void updateHashValue() noexcept;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			protected:
				std::reference_wrapper<Component> m_component;
				std::reference_wrapper<CrystalLattice> m_crystalLattice;
				std::vector<int> m_latticePoint;
				NumericVector m_latticePointVector;
				size_type m_hashValue;
			};


			template <typename C>
			inline bool operator==(const TranslatingAdaptor<C>& taa, const TranslatingAdaptor<C>& tab)
			{
				if (&(taa.entity()) != &(tab.entity()))
					return false;
				else
					return taa.isSameLatticePoint(tab);
			}

			template <typename C>
			inline bool operator!=(const TranslatingAdaptor<C>& taa, const TranslatingAdaptor<C>& tab)
			{
				return !(taa == tab);
			}

			template <typename C>
			inline TranslatingAdaptor<C> createTranslatingAdaptor(C& component, const SolidStatePhysics::Materials::CrystalLattice& crystalLattice) noexcept
			{
				return TranslatingAdaptor<C>{ component, crystalLattice };
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

template <typename C>
inline SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::TranslatingAdaptor(Component& component, CrystalLattice& crystalLattice)
	: m_component{ std::ref(component) }
	, m_crystalLattice{ std::ref(crystalLattice) }
	, m_latticePoint{}
	, m_latticePointVector{ 0.0,0.0,0.0 }
	, m_hashValue{ 0 }
{
	m_hashValue = std::hash<const Component*>{}(&(m_component.get()));
	m_hashValue ^= std::hash<std::string>{}(std::string{ "000" });
}

template <typename C>
inline SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::TranslatingAdaptor(Component& component, CrystalLattice& crystalLattice, const LatticePoint& latticePoint)
	: m_component{ std::ref(component) }
	, m_crystalLattice{ std::ref(crystalLattice) }
	, m_latticePoint{ latticePoint[0], latticePoint[1], latticePoint[2] }
	, m_latticePointVector{ 0.0,0.0,0.0 }
	, m_hashValue{ 0 }
{
	m_hashValue = std::hash<const Component*>{}(&(m_component.get()));

	if (m_latticePoint[0] != 0 || m_latticePoint[1] != 0 || m_latticePoint[2] != 0)
	{
		m_latticePointVector[0] = static_cast<double>(m_latticePoint[0]);
		m_latticePointVector[1] = static_cast<double>(m_latticePoint[1]);
		m_latticePointVector[2] = static_cast<double>(m_latticePoint[2]);

		std::string latticePointString;
		{
			for (const auto num : m_latticePoint)
				latticePointString += std::to_string(num);
		}

		m_hashValue ^= std::hash<std::string>{}(latticePointString);
	}

	else
	{
		m_latticePoint.clear();
		m_hashValue ^= std::hash<std::string>{}(std::string{ "000" });
	}
}

template <typename C>
template <typename AC>
inline SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::TranslatingAdaptor(Component& component, const TranslatingAdaptor<AC>& anotherComponent)
	: m_component{ std::ref(component) }
	, m_crystalLattice{ std::ref(anotherComponent.crystalLattice()) }
	, m_latticePoint{}
	, m_latticePointVector{ 0.0,0.0,0.0 }
	, m_hashValue{ 0 }
{
	LatticePoint latticePoint = anotherComponent.latticePoint();
	{
		if (latticePoint[0] != 0 || latticePoint[1] != 0 || latticePoint[2] != 0)
		{
			m_latticePoint.resize(3);
			m_latticePoint[0] = latticePoint[0];
			m_latticePoint[1] = latticePoint[1];
			m_latticePoint[2] = latticePoint[2];

			m_latticePointVector[0] = static_cast<double>(m_latticePoint[0]);
			m_latticePointVector[1] = static_cast<double>(m_latticePoint[1]);
			m_latticePointVector[2] = static_cast<double>(m_latticePoint[2]);
		}
	}

	m_hashValue = std::hash<const Component*>{}(&(m_component.get()));
	{
		if (m_latticePoint.empty())
			m_hashValue ^= std::hash<std::string>{}(std::string{ "000" });

		else
		{
			std::string latticePointString;
			{
				for (const auto num : m_latticePoint)
					latticePointString += std::to_string(num);
			}

			m_hashValue ^= std::hash<std::string>{}(latticePointString);
		}
	}
}

template <typename C>
inline SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::TranslatingAdaptor(TranslatingAdaptor&& ta) noexcept
	: m_component{ ta.m_component }
	, m_crystalLattice{ ta.m_crystalLattice }
	, m_latticePoint{ std::move(ta.m_latticePoint) }
	, m_latticePointVector{ std::move(ta.m_latticePointVector) }
	, m_hashValue{ ta.m_hashValue }
{
}

template <typename C>
inline SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>& SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::operator=(TranslatingAdaptor&& ta) noexcept
{
	m_component = ta.m_component;
	m_crystalLattice = ta.m_crystalLattice;
	m_latticePoint = std::move(ta.m_latticePoint);
	m_latticePointVector = std::move(ta.m_latticePointVector);
	m_hashValue = ta.m_hashValue;

	return *this;
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access functions

template <typename C>
inline typename SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::Component& SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::entity() const noexcept
{
	return m_component.get();
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::resetEntity(Component& component) noexcept
{
	m_component = std::ref(component);
}

template <typename C>
inline typename SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::CrystalLattice& SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::crystalLattice() const noexcept
{
	return m_crystalLattice.get();
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::resetCrystalLattice(CrystalLattice& crystalLattice) noexcept
{
	m_crystalLattice = std::ref(crystalLattice);
}

template <typename C>
inline typename SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::LatticePoint SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::latticePoint() const noexcept
{
	if (m_latticePoint.empty())
		return LatticePoint{ 0,0,0 };
	else
		return LatticePoint{ m_latticePoint[0], m_latticePoint[1], m_latticePoint[2] };
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::resetLatticePoint(const LatticePoint& latticePoint) noexcept
{
	m_latticePoint = std::vector<int>{ latticePoint[0], latticePoint[1], latticePoint[2] };

	if (m_latticePoint[0] != 0 || m_latticePoint[1] != 0 || m_latticePoint[2] != 0)
	{
		m_latticePointVector[0] = static_cast<double>(latticePoint[0]);
		m_latticePointVector[1] = static_cast<double>(latticePoint[1]);
		m_latticePointVector[2] = static_cast<double>(latticePoint[2]);
	}

	else
	{
		m_latticePoint.clear();
		m_latticePointVector = NumericVector{ 0.0,0.0,0.0 };
	}

	updateHashValue();
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::reverseLatticePoint() noexcept
{
	if (!(m_latticePoint.empty()))
	{
		m_latticePoint[0] *= (-1);
		m_latticePoint[1] *= (-1);
		m_latticePoint[2] *= (-1);

		m_latticePointVector[0] *= (-1.0);
		m_latticePointVector[1] *= (-1.0);
		m_latticePointVector[2] *= (-1.0);

		updateHashValue();
	}
}

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::translateLatticePoint(const LatticePoint& latticePoint) noexcept
{
	{
		if (m_latticePoint.empty())
			m_latticePoint.resize(3, 0);

		m_latticePoint[0] += latticePoint[0];
		m_latticePoint[1] += latticePoint[1];
		m_latticePoint[2] += latticePoint[2];
	}

	if (m_latticePoint[0] != 0 || m_latticePoint[1] != 0 || m_latticePoint[2] != 0)
	{
		m_latticePointVector[0] = static_cast<double>(m_latticePoint[0]);
		m_latticePointVector[1] = static_cast<double>(m_latticePoint[1]);
		m_latticePointVector[2] = static_cast<double>(m_latticePoint[2]);
	}

	else
	{
		m_latticePoint.clear();
		m_latticePointVector = NumericVector{ 0.0,0.0,0.0 };
	}

	updateHashValue();
}

template <typename C>
inline std::size_t SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::hashValue() const noexcept
{
	return m_hashValue;
}

// Access functions
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

template <typename C>
inline NumericalAnalysis::LinearAlgebra::NumericVector SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::translationVector() const noexcept
{
	if (m_latticePoint.empty())
		return m_latticePointVector;
	else
		return (m_crystalLattice.get().primitiveLatticeVectors() * m_latticePointVector);
}

template <typename C>
inline bool SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::isInOriginalCell() const noexcept
{
	return m_latticePoint.empty();
}

template <typename C>
inline bool SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::isSameEntity(const Component& component) const noexcept
{
	if (&(m_component.get()) == &(m_component))
		return true;
	else
		return false;
}

template <typename C>
template <typename AC>
inline bool SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::isSameLatticePoint(const TranslatingAdaptor<AC>& anotherAdaptor) const noexcept
{
	if (isInOriginalCell())
		return anotherAdaptor.isInOriginalCell();

	else
	{
		LatticePoint anotherLatticePoint = anotherAdaptor.latticePoint();

		if (m_latticePoint[0] == anotherLatticePoint[0] && m_latticePoint[1] == anotherLatticePoint[1] && m_latticePoint[2] == anotherLatticePoint[2])
			return true;
		else
			return false;
	}
}

template <typename C>
template <typename AC>
inline typename SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::LatticePoint SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::getRelativeLatticePointTo(const TranslatingAdaptor<AC>& anotherAdaptor) const noexcept
{
	if (isInOriginalCell())
	{
		if(anotherAdaptor.isInOriginalCell())
			return LatticePoint{ 0,0,0 };
		else
			return anotherAdaptor.latticePoint();
	}

	else
	{
		if (anotherAdaptor.isInOriginalCell())
		{
			LatticePoint relativeLatticePoint = latticePoint();
			relativeLatticePoint[0] *= (-1);
			relativeLatticePoint[1] *= (-1);
			relativeLatticePoint[2] *= (-1);

			return relativeLatticePoint;
		}

		else
		{
			LatticePoint thisLatticePoint = latticePoint();

			LatticePoint resultLatticePoint = anotherAdaptor.latticePoint();
			resultLatticePoint[0] -= thisLatticePoint[0];
			resultLatticePoint[1] -= thisLatticePoint[1];
			resultLatticePoint[2] -= thisLatticePoint[2];

			return resultLatticePoint;
		}
	}
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename C>
inline void SolidStatePhysics::Crystallization::Calculators::TranslatingAdaptor<C>::updateHashValue() noexcept
{
	m_hashValue = std::hash<const Component*>{}(&(m_component.get()));
	{
		if (m_latticePoint.empty())
			m_hashValue ^= std::hash<std::string>{}(std::string{ "000" });

		else
		{
			std::string latticePointString;
			{
				for (const auto num : m_latticePoint)
					latticePointString += std::to_string(num);
			}

			m_hashValue ^= std::hash<std::string>{}(latticePointString);
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_CRYSTALLIZATION_CALCULATORS_TRANSLATINGADAPTOR_H
