#ifndef SOLIDSTATEPHYSICS_UNITSYSTEM_UNITCONVERSION_H
#define SOLIDSTATEPHYSICS_UNITSYSTEM_UNITCONVERSION_H

#include <type_traits>

#include "LinearAlgebra.h"

#include "EnergyUnit.h"
#include "LengthUnit.h"
#include "MassUnit.h"
#include "PressureUnit.h"
#include "TemperatureUnit.h"
#include "TimeUnit.h"
#include "VelocityUnit.h"


namespace SolidStatePhysics
{
	namespace UnitSystem
	{
		template <typename to, typename from>
		struct time_cast
		{
			constexpr double operator()(const double time) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from>
		struct mass_cast
		{
			constexpr double operator()(const double mass) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from = SolidStatePhysics::UnitSystem::LengthUnit::Angstrom>
		struct length_cast
		{
			constexpr double operator()(const double length) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from>
		struct energy_cast
		{
			constexpr double operator()(const double energy) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from>
		struct pressure_cast
		{
			constexpr double operator()(const double pressure) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from>
		struct speed_cast
		{
			constexpr double operator()(const double speed) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};

		template <typename to, typename from>
		struct velocity_cast
		{
			NumericalAnalysis::LinearAlgebra::NumericVector operator()(const NumericalAnalysis::LinearAlgebra::NumericVector& velocity) const noexcept;
			constexpr double scalingMultiplier() const noexcept;
		};
	}
}


namespace SolidStatePhysics
{
	namespace UnitSystem
	{
		template <typename from, typename to = SolidStatePhysics::UnitSystem::TemperatureUnit::Kelvin>
		inline constexpr double temperature_cast(const double temperature) noexcept
		{
			if constexpr (std::is_same<to, from>::value == true)
				return temperature;
			else
				return (temperature + from::zeroTemperatureInSI() - to::zeroTemperatureInSI());
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::time_cast<to, from>::operator()(const double time) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return time;
	else
		return ((time * from::unitTimeInSI()) / to::unitTimeInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::time_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitTimeInSI() / to::unitTimeInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::mass_cast<to, from>::operator()(const double mass) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return mass;
	else
		return ((mass * from::unitMassInSI()) / to::unitMassInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::mass_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitMassInSI() / to::unitMassInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::length_cast<to, from>::operator()(const double length) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return length;
	else
		return ((length * from::unitLengthInSI()) / to::unitLengthInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::length_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitLengthInSI() / to::unitLengthInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::energy_cast<to, from>::operator()(const double energy) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return energy;
	else
		return ((energy * from::unitEnergyInSI()) / to::unitEnergyInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::energy_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitEnergyInSI() / to::unitEnergyInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::pressure_cast<to, from>::operator()(const double pressure) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return pressure;
	else
		return ((pressure * from::unitPressureInSI()) / to::unitPressureInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::pressure_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitPressureInSI() / to::unitPressureInSI());
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::speed_cast<to, from>::operator()(const double speed) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return speed;
	else
		return (speed * (from::unitVelocityInSI() / to::unitVelocityInSI()));
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::speed_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitVelocityInSI() / to::unitVelocityInSI());
}

template <typename to, typename from>
inline NumericalAnalysis::LinearAlgebra::NumericVector SolidStatePhysics::UnitSystem::velocity_cast<to, from>::operator()(const NumericalAnalysis::LinearAlgebra::NumericVector& velocity) const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return velocity;
	else
		return (velocity * (from::unitVelocityInSI() / to::unitVelocityInSI()));
}

template <typename to, typename from>
inline constexpr double SolidStatePhysics::UnitSystem::velocity_cast<to, from>::scalingMultiplier() const noexcept
{
	if constexpr (std::is_same<to, from>::value == true)
		return 1.0;
	else
		return (from::unitVelocityInSI() / to::unitVelocityInSI());
}

// Crystal Information
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !SOLIDSTATEPHYSICS_UNITSYSTEM_UNITCONVERSION_H
