#include "ChemicalBond.h"

#include <limits>
#include <cmath>

using namespace SolidStatePhysics::Crystallization::Components;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and destructor

double ChemicalBond::m_maxBondingLengthExtensionRate{ 2.0 };


ChemicalBond::ChemicalBond(ChemicalBondOwner& realOwner, const ChemicalBondOwnerAdaptor& attachedOwner)
	: MinMaxConstrainter{ realOwner.ownerAtom(), AtomicAdaptor(attachedOwner.entity().ownerAtom(), attachedOwner) }
	, m_realOwner{ std::ref(realOwner) }
	, m_attachedOwner{ attachedOwner }
	, m_bondingEnergy{ 0.0 }
{
	m_bondingEnergy = getBondingEnergy(ownerAtom(), attachedAtom());
}

ChemicalBond::ChemicalBond(ChemicalBondOwner& realOwner, ChemicalBondOwnerAdaptor&& attachedOwner)
	: MinMaxConstrainter{ realOwner.ownerAtom(), AtomicAdaptor(attachedOwner.entity().ownerAtom(), attachedOwner) }
	, m_realOwner{ std::ref(realOwner) }
	, m_attachedOwner{ std::move(attachedOwner) }
	, m_bondingEnergy{ 0.0 }
{
	m_bondingEnergy = getBondingEnergy(ownerAtom(), attachedAtom());
}

ChemicalBond::ChemicalBond(const std::shared_ptr<ChemicalBond>& twinPtr)
	: MinMaxConstrainter{ std::static_pointer_cast<MinMaxConstrainter>(twinPtr) }
	, m_realOwner{ twinPtr->m_attachedOwner.entity() }
	, m_attachedOwner{ twinPtr->m_realOwner, twinPtr->m_attachedOwner }
	, m_bondingEnergy{ twinPtr->m_bondingEnergy }
{
	m_attachedOwner.reverseLatticePoint();
}

ChemicalBond::~ChemicalBond()
{
	if (std::shared_ptr<ChemicalBond> twinChemicalBondPtr = std::dynamic_pointer_cast<ChemicalBond>(twinPtr()))
	{
		twinChemicalBondPtr->m_realOwner.get().disjoin();
		twinChemicalBondPtr->resetTwinPtr();
	}
}

// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

bool ChemicalBond::shouldCreate(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept
{
	if ((ownerAtom.isCation() && attachedAtom.entity().isAnion()) || (ownerAtom.isAnion() && attachedAtom.entity().isCation()))
	{
		NumericVector displacement = attachedAtom.entity().cartesianCoordinate() + attachedAtom.translationVector();
		displacement -= ownerAtom.cartesianCoordinate();
		double maxBondLength = m_maxBondingLengthExtensionRate * (ownerAtom.maximumIonicRadius() + attachedAtom.entity().maximumIonicRadius());

		if (displacement.normSquare() < (maxBondLength * maxBondLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

bool ChemicalBond::shouldCreate(const std::shared_ptr<InteratomicConnector>& interatomicConnectorPtr) noexcept
{
	if ((interatomicConnectorPtr->ownerAtom().isCation() && interatomicConnectorPtr->attachedAtom().entity().isAnion()) || (interatomicConnectorPtr->ownerAtom().isAnion() && interatomicConnectorPtr->attachedAtom().entity().isCation()))
	{
		double maxBondLength = m_maxBondingLengthExtensionRate * (interatomicConnectorPtr->ownerAtom().maximumIonicRadius() + interatomicConnectorPtr->attachedAtom().entity().maximumIonicRadius());

		if (interatomicConnectorPtr->displacement().normSquare() < (maxBondLength * maxBondLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

double ChemicalBond::getBondingEnergy(const double ownerBondingAffinity, const double attachedBondingAffinity) noexcept
{
	return (ownerBondingAffinity * attachedBondingAffinity);
}

double ChemicalBond::getBondingEnergy(const InteractingAtom& ownerAtom, const AtomicAdaptor& attachedAtom) noexcept
{
	return getBondingEnergy(ownerAtom.bondingMotivation().second, attachedAtom.entity().bondingMotivation().second);
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
