#include "GeneratePeriodicSolids.h"

#include "MpiController.h"
#include "DirectoryManipulator.h"
#include "LogWriter.h"
#include "Clock.h"

#include "Crystallizer.h"
#include "IntelligentLiquid.h"

using namespace ControlCenter::Administering;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and destructor

SolidStatePhysics::Crystallization::Simulators::FinalSimulationsReporter GeneratePeriodicSolids::m_finalSimulationsReporter;
std::size_t GeneratePeriodicSolids::m_sampleIndex = 0;
std::mutex GeneratePeriodicSolids::m_generateMutex;
std::mutex GeneratePeriodicSolids::m_reportMutex;
std::mutex GeneratePeriodicSolids::m_directoryMutex;


GeneratePeriodicSolids::GeneratePeriodicSolids()
	: m_threadIndex{ 0 }
	, m_crystallizationPlan{}
{
}

GeneratePeriodicSolids::GeneratePeriodicSolids(const size_type threadIndex)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{}
{
}

GeneratePeriodicSolids::GeneratePeriodicSolids(const size_type threadIndex, const CrystallizationPlan& crystallizationPlan, const MonitoringPlan& monitoringPlan)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{ crystallizationPlan }
	, m_monitoringPlan{ monitoringPlan }
{
}

GeneratePeriodicSolids::GeneratePeriodicSolids(const size_type threadIndex, CrystallizationPlan&& crystallizationPlan, MonitoringPlan&& monitoringPlan)
	: m_threadIndex{ threadIndex }
	, m_crystallizationPlan{ std::move(crystallizationPlan) }
	, m_monitoringPlan{ std::move(monitoringPlan) }
{
}

GeneratePeriodicSolids::~GeneratePeriodicSolids()
{
}

// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

void GeneratePeriodicSolids::reportResults(const StructuralDictionary& structuralDictionary, const std::string& elapsedTime)
{
	std::string finalReportString;
	{
		finalReportString += "-----------------------------------------------------------------------------------------------------------------------\n";
		finalReportString += "-----------------------------------------------------------------------------------------------------------------------\n\n";
		finalReportString += structuralDictionary.chemicalConstituentsDirectoryPath().stem().generic_string();
		finalReportString += " system\n\n";
		finalReportString += "Creation date: ";
		finalReportString += ControlCenter::System::Clock::tellNowTime();
		finalReportString += "\n";
		finalReportString += "Elapsed time : ";
		finalReportString += elapsedTime;
		finalReportString += "\n";
		finalReportString += "-----------------------------------------------------------------------------------------------------------------------\n";
		finalReportString += "-----------------------------------------------------------------------------------------------------------------------\n";

		finalReportString += structuralDictionary.writeFinalReport();
	}

	m_finalSimulationsReporter.createFinalReportFile(finalReportString);
	m_finalSimulationsReporter.createDistributionFiles();
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void GeneratePeriodicSolids::operator()(const std::filesystem::path& chemicalConstituentsDirectoryPath, const ChemicalConstituents& chemicalConstituents, const size_type feedstocks) noexcept
{
	m_feedstocks = feedstocks;
	m_chemicalConstituents = chemicalConstituents;
	m_structuralDictionary = StructuralDictionary{ chemicalConstituentsDirectoryPath, m_crystallizationPlan.maxGeometricalOptimizing() };
	{
		if (m_monitoringPlan.needFailureData())
		{
			if (m_threadIndex == 0)
			{
				std::filesystem::path failureDirectoryPath = chemicalConstituentsDirectoryPath;
				failureDirectoryPath /= "Failure";
				ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(failureDirectoryPath);

				std::filesystem::path exceptionDirectoryPath = chemicalConstituentsDirectoryPath;
				exceptionDirectoryPath /= "Exception";
				ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(exceptionDirectoryPath);
			}
		}
	}

	std::unique_lock<std::mutex> generateLock{ m_generateMutex, std::defer_lock };

	while (shouldMakeSample())
	{
		generateLock.lock();

		if (tryUseFeedstock())
		{
			std::string sampleName = getSampleName();
			generateLock.unlock();

			std::filesystem::path sampleDirectoryPath = chemicalConstituentsDirectoryPath;
			sampleDirectoryPath /= sampleName;
			ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(sampleDirectoryPath);

			operateCrystallizer(sampleDirectoryPath);
		}

		else
		{
			generateLock.unlock();
			break;
		}
	}

	m_feedstocks = 0;
	m_chemicalConstituents.clear();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void GeneratePeriodicSolids::operateCrystallizer(const std::filesystem::path& sampleDirectoryPath)
{
	SolidStatePhysics::Crystallization::Components::IntelligentLiquid intelligentLiquid;
	intelligentLiquid.renewAtomicArrangementRandomly(m_chemicalConstituents);
	intelligentLiquid.renewCrystalLatticeRandomly(1.0, 2.0);
	intelligentLiquid.renewAtomicArrangementRandomly();
	intelligentLiquid.resetPressure(m_crystallizationPlan.pressure());

	try
	{
		SolidStatePhysics::Crystallization::Simulators::Crystallizer crystallizer{ m_crystallizationPlan, m_monitoringPlan };
		crystallizer.initialize(sampleDirectoryPath);
		crystallizer.run(intelligentLiquid);

		if (intelligentLiquid.isDefinitelyCrystallized())
		{
			m_structuralDictionary.addCrystallizedSamplesCount();
			const std::string structuralFingerprint = intelligentLiquid.writeFingerprint();
			{
				if (m_structuralDictionary.contains(structuralFingerprint))
				{
					registerFinalSimulationReport(structuralFingerprint, crystallizer.elapsedTime(), crystallizer.localOptimizationCount(), crystallizer.remodellingCount());

					if (!(m_monitoringPlan.needAllSuccessSampleData()))
					{
						if (std::filesystem::exists(sampleDirectoryPath))
							std::filesystem::remove_all(sampleDirectoryPath);
					}
				}

				else
				{
					m_structuralDictionary.registerCrystalStructure(structuralFingerprint, intelligentLiquid.getAtomicEnvironments());
					registerFinalSimulationReport(structuralFingerprint, crystallizer.elapsedTime(), crystallizer.localOptimizationCount(), crystallizer.remodellingCount());

					const size_type structuralIndex = m_structuralDictionary.getStructuralIndex(structuralFingerprint);
					transmitSampleToStructuralDirectory(sampleDirectoryPath, structuralIndex, intelligentLiquid.getAtomicEnvironments());
				}
			}
		}

		else
		{
			m_structuralDictionary.addFailedSamplesCount();
			transmitSampleToFailureDirectory(sampleDirectoryPath);
		}

		crystallizer.finalize();
	}

	catch (const std::exception& e)
	{
		m_structuralDictionary.addExceptionSamplesCount();
		transmitSampleToExceptionDirectory(sampleDirectoryPath);

		std::string errorMessage;
		{
			errorMessage += "Thread_";
			errorMessage += std::to_string(m_threadIndex);
			errorMessage += " of ";
			errorMessage += "Process_";
			errorMessage += std::to_string(ControlCenter::System::MpiController::thisMpiNumber());
			errorMessage += '\n';
			errorMessage += e.what();
		}

		ControlCenter::System::IO::LogWriter logWriter;
		logWriter.report(errorMessage);
	}
}

void GeneratePeriodicSolids::registerFinalSimulationReport(const std::string& structuralFingerprint, const size_type elapsedTime, const size_type localOptimizationCount, const size_type remodellingCount)
{
	const size_type structuralIndex = m_structuralDictionary.getStructuralIndex(structuralFingerprint);

	if (m_structuralDictionary.needFinalSimulationsReportsPerStructure())
	{
		m_structuralDictionary.getFinalSimulationsReporter(structuralIndex).registerElapsedTime(elapsedTime);
		m_structuralDictionary.getFinalSimulationsReporter(structuralIndex).registerLocalOptimizationHistory(localOptimizationCount);
		m_structuralDictionary.getFinalSimulationsReporter(structuralIndex).registerRemodellingHistory(remodellingCount);
	}

	std::lock_guard<std::mutex> guard{ m_reportMutex };
	m_finalSimulationsReporter.registerElapsedTime(elapsedTime);
	m_finalSimulationsReporter.registerLocalOptimizationHistory(localOptimizationCount);
	m_finalSimulationsReporter.registerRemodellingHistory(remodellingCount);
}

void GeneratePeriodicSolids::transmitSampleToStructuralDirectory(const std::filesystem::path& sampleDirectoryPath, const size_type structuralIndex, const std::map<std::string, size_type>& atomicEnvironments)
{
	std::filesystem::path structuralDirectoryPath = sampleDirectoryPath.parent_path();
	{
		std::string atomicEnvironmentsCountDirectoryName;
		{
			size_type atomicEnvironmentsCount = 0;
			{
				for (const auto& symbolAndNum : atomicEnvironments)
					atomicEnvironmentsCount += symbolAndNum.second;
			}

			atomicEnvironmentsCountDirectoryName += std::to_string(atomicEnvironmentsCount);
			atomicEnvironmentsCountDirectoryName += "_AtomicEnvironments";
		}
		std::string atomicEnvironmentsName;
		{
			for (const auto& symbolAndNum : atomicEnvironments)
			{
				atomicEnvironmentsName += symbolAndNum.first;
				atomicEnvironmentsName += "-";
				atomicEnvironmentsName += std::to_string(symbolAndNum.second);
				atomicEnvironmentsName += "_";
			}

			atomicEnvironmentsName.pop_back();
		}
		std::string structureName{ "Structure_" };
		structureName += std::to_string(structuralIndex);
		

		std::lock_guard<std::mutex> guard{ m_directoryMutex };
		structuralDirectoryPath /= atomicEnvironmentsCountDirectoryName;
		ControlCenter::System::IO::DirectoryManipulator::tryCreateDirectory(structuralDirectoryPath);
		structuralDirectoryPath /= atomicEnvironmentsName;
		ControlCenter::System::IO::DirectoryManipulator::tryCreateDirectory(structuralDirectoryPath);
		structuralDirectoryPath /= structureName;
		ControlCenter::System::IO::DirectoryManipulator::tryCreateDirectory(structuralDirectoryPath);
	}
	structuralDirectoryPath /= getDirectoryName(sampleDirectoryPath);

	ControlCenter::System::IO::DirectoryManipulator::overCreateDirectory(structuralDirectoryPath);
	ControlCenter::System::IO::DirectoryManipulator::overTransmitAllFilesRecursively(sampleDirectoryPath, structuralDirectoryPath);
}

void GeneratePeriodicSolids::transmitSampleToFailureDirectory(const std::filesystem::path& sampleDirectoryPath)
{
	if (m_monitoringPlan.needFailureData())
	{
		std::filesystem::path newSampleDirectoryPath = sampleDirectoryPath.parent_path();
		newSampleDirectoryPath /= "Failure";
		newSampleDirectoryPath /= getDirectoryName(sampleDirectoryPath);
		ControlCenter::System::IO::DirectoryManipulator::overTransmitAllFilesRecursively(sampleDirectoryPath, newSampleDirectoryPath);
	}

	else
	{
		if (std::filesystem::exists(sampleDirectoryPath))
			std::filesystem::remove_all(sampleDirectoryPath);
	}
}

void GeneratePeriodicSolids::transmitSampleToExceptionDirectory(const std::filesystem::path& sampleDirectoryPath)
{
	if (m_monitoringPlan.needFailureData())
	{
		std::filesystem::path newSampleDirectoryPath = sampleDirectoryPath.parent_path();
		newSampleDirectoryPath /= "Exception";
		newSampleDirectoryPath /= getDirectoryName(sampleDirectoryPath);
		ControlCenter::System::IO::DirectoryManipulator::overTransmitAllFilesRecursively(sampleDirectoryPath, newSampleDirectoryPath);
	}

	else
	{
		if (std::filesystem::exists(sampleDirectoryPath))
			std::filesystem::remove_all(sampleDirectoryPath);
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
