#include "InteractingAtom.h"

#include <limits>
#include <algorithm>

#include "AbstractChemicalElement.h"

#include "ChemicalBond.h"
#include "IonicConstrainter.h"
#include "MetallicConstrainter.h"
#include "PolyhedralConstrainter.h"
#include "SecondNeighborConstrainter.h"

using namespace SolidStatePhysics::Crystallization::Components;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and operators

InteractingAtom::InteractingAtom(const size_type atomicNumber)
	: AtomicNucleus{ atomicNumber }
	, m_minimumIonicRadius{ 0.0 }
	, m_maximumIonicRadius{ 0.0 }
	, m_metallicRadius{ 0.0 }
	, m_chemicalBondOwners{}
	, m_ionicConstraintOwners{}
	, m_metallicConstraintOwners{}
	, m_polyhedralConstraintOwners{}
	, m_secondNeighborConstraintOwners{}
{
	SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement abstractChemicalElement;
	abstractChemicalElement.resetElementNumber(atomicNumber);

	m_minimumIonicRadius = abstractChemicalElement.getMinimumIonicRadius();
	m_maximumIonicRadius = abstractChemicalElement.getMaximumIonicRadius();
	m_metallicRadius = abstractChemicalElement.getMetallicRadius();
}

InteractingAtom::InteractingAtom(const size_type atomicNumber, const NumericVector& cartesianCoordinate)
	: AtomicNucleus{ atomicNumber, cartesianCoordinate }
	, m_minimumIonicRadius{ 0.0 }
	, m_maximumIonicRadius{ 0.0 }
	, m_metallicRadius{ 0.0 }
	, m_chemicalBondOwners{}
	, m_ionicConstraintOwners{}
	, m_metallicConstraintOwners{}
	, m_polyhedralConstraintOwners{}
	, m_secondNeighborConstraintOwners{}
{
	SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement abstractChemicalElement;
	abstractChemicalElement.resetElementNumber(atomicNumber);

	m_minimumIonicRadius = abstractChemicalElement.getMinimumIonicRadius();
	m_maximumIonicRadius = abstractChemicalElement.getMaximumIonicRadius();
	m_metallicRadius = abstractChemicalElement.getMetallicRadius();
}

InteractingAtom::InteractingAtom(const InteractingAtom& interactingAtom)
	: AtomicNucleus{ interactingAtom.atomicNumber(), interactingAtom.cartesianCoordinate() }
	, m_minimumIonicRadius{ interactingAtom.m_minimumIonicRadius }
	, m_maximumIonicRadius{ interactingAtom.m_maximumIonicRadius }
	, m_metallicRadius{ interactingAtom.m_metallicRadius }
	, m_chemicalBondOwners{}
	, m_ionicConstraintOwners{}
	, m_metallicConstraintOwners{}
	, m_polyhedralConstraintOwners{}
	, m_secondNeighborConstraintOwners{}
{
}

InteractingAtom::InteractingAtom(InteractingAtom&& interactingAtom) noexcept
	: AtomicNucleus{ std::move(interactingAtom) }
	, m_minimumIonicRadius{ interactingAtom.m_minimumIonicRadius }
	, m_maximumIonicRadius{ interactingAtom.m_maximumIonicRadius }
	, m_metallicRadius{ interactingAtom.m_metallicRadius }
	, m_chemicalBondOwners{}
	, m_ionicConstraintOwners{}
	, m_metallicConstraintOwners{}
	, m_polyhedralConstraintOwners{}
	, m_secondNeighborConstraintOwners{}
{
	interactingAtom.initialize();
}

InteractingAtom& InteractingAtom::operator=(const InteractingAtom& interactingAtom)
{
	AtomicNucleus::operator=(interactingAtom);
	InteractingAtom::initialize();

	return *this;
}

InteractingAtom& InteractingAtom::operator=(InteractingAtom&& interactingAtom) noexcept
{
	AtomicNucleus::operator=(std::move(interactingAtom));
	InteractingAtom::initialize();

	interactingAtom.initialize();

	return *this;
}

// Constructors and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

InteractingAtom::size_type InteractingAtom::countMaxCommonBridgingAtomsSharing() const noexcept
{
	size_type maxCommonBridgingAtomsSharing = 0;
	{
		for (const auto& polyhedralConstraintOwner : m_polyhedralConstraintOwners)
		{
			if (polyhedralConstraintOwner.second.isActive())
			{
				size_type commonBridgingAtomsSharing = polyhedralConstraintOwner.second.owingPolyhedralConstrainter().commonBridgingAtomsSharing();

				if (maxCommonBridgingAtomsSharing < commonBridgingAtomsSharing)
					maxCommonBridgingAtomsSharing = commonBridgingAtomsSharing;
			}
		}
	}

	return maxCommonBridgingAtomsSharing;
}

double InteractingAtom::getMaxEquilibriumBondLength() const noexcept
{
	double maxEquilibriumBondingDistance = 0.0;
	{
		for (const auto& chemicalBondOwner : m_chemicalBondOwners)
		{
			if (chemicalBondOwner.second.isActive())
			{
				double equilibriumBondDistance = chemicalBondOwner.second.owingChemicalBond().maximumDistance();

				if (maxEquilibriumBondingDistance < equilibriumBondDistance)
					maxEquilibriumBondingDistance = equilibriumBondDistance;
			}
		}
	}

	return maxEquilibriumBondingDistance;
}

double InteractingAtom::getMinEquilibriumBondLength() const noexcept
{
	double minEquilibriumBondingDistance = std::numeric_limits<double>::max();
	{
		for (const auto& chemicalBondOwner : m_chemicalBondOwners)
		{
			if (chemicalBondOwner.second.isActive())
			{
				double equilibriumBondDistance = chemicalBondOwner.second.owingChemicalBond().minimumDistance();

				if (equilibriumBondDistance < minEquilibriumBondingDistance)
					minEquilibriumBondingDistance = equilibriumBondDistance;
			}
		}
	}

	return minEquilibriumBondingDistance;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void InteractingAtom::initialize() noexcept
{
	AtomicNucleus::initialize();

	clearOwingChemicalBonds();
	clearOwingIonicConstrainters();
	clearOwingMetallicConstrainters();
	clearOwingPolyhedralConstrainters();
	clearOwingSecondNeighborConstrainters();
}

void InteractingAtom::normalizeFractionalCoordinate(const CrystalLattice& crystalLattice)
{
	clearOwingChemicalBonds();
	clearOwingIonicConstrainters();
	clearOwingMetallicConstrainters();
	clearOwingPolyhedralConstrainters();
	clearOwingSecondNeighborConstrainters();

	NumericVector fractionalCoordinate = (crystalLattice.inversePrimitiveLatticeVectors() * cartesianCoordinate());
	int floorValueA = static_cast<int>(std::floor(fractionalCoordinate[0]));
	int floorValueB = static_cast<int>(std::floor(fractionalCoordinate[1]));
	int floorValueC = static_cast<int>(std::floor(fractionalCoordinate[2]));

	if (floorValueA != 0 || floorValueB != 0 || floorValueC != 0)
	{
		fractionalCoordinate[0] -= static_cast<double>(floorValueA);
		fractionalCoordinate[1] -= static_cast<double>(floorValueB);
		fractionalCoordinate[2] -= static_cast<double>(floorValueC);

		resetCartesianCoordinate(crystalLattice.primitiveLatticeVectors() * fractionalCoordinate);
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

bool InteractingAtom::configureChemicalBonds(PeriodicLiquid& ownerPeriodicLiquid)
{
	if (bondingMotivation().first <= getCoordinationNumber())
		return false;

	else
	{
		const size_type additionalCoordinationNumber = bondingMotivation().first - getCoordinationNumber();
		std::map<double, AtomicAdaptor> promisingAtomicAdaptors;
		{
			for (auto iter = m_ionicConstraintOwners.begin(); iter != m_ionicConstraintOwners.end(); ++iter)
			{
				if (iter->second.isActive())
				{
					if (ChemicalBond::shouldCreate(*this, iter->first) && !(hasChemicalBond(iter->first)))
						promisingAtomicAdaptors.emplace(iter->second.owingIonicConstrainter().interatomicDistance(), iter->first);
				}
			}
		}

		return tryConfigureChemicalBonds(promisingAtomicAdaptors, additionalCoordinationNumber, ownerPeriodicLiquid);
	}
}

void InteractingAtom::configurePolyhedralConstrainters(PeriodicLiquid& ownerPeriodicLiquid)
{
	eraseInactiveConstraintOwners(m_chemicalBondOwners);

	for (auto iter = m_chemicalBondOwners.begin(); iter != m_chemicalBondOwners.end(); ++iter)
	{
		for (auto anotherIter = std::next(iter); anotherIter != m_chemicalBondOwners.end(); ++anotherIter)
		{
			AtomicAdaptor anotherAtomicAdaptor{ anotherIter->first };
			anotherAtomicAdaptor.resetLatticePoint(iter->first.getRelativeLatticePointTo(anotherIter->first));

			if (!(iter->first.entity().hasChemicalBond(anotherAtomicAdaptor)))
			{
				AtomicAdaptor thisAtomicAdaptor{ *this, iter->first };
				thisAtomicAdaptor.reverseLatticePoint();
				iter->first.entity().createPolyhedralConstrainter(anotherAtomicAdaptor, ownerPeriodicLiquid, thisAtomicAdaptor);

				if (iter->first.entity().hasIonicConstrainter(anotherAtomicAdaptor))
					iter->first.entity().eraseIonicConstrainter(anotherAtomicAdaptor);

				if (iter->first.entity().hasMetallicConstrainter(anotherAtomicAdaptor))
					iter->first.entity().eraseMetallicConstrainter(anotherAtomicAdaptor);
			}
		}
	}
}

void InteractingAtom::createIonicConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid)
{
	createConstrainter(attachedAtom, m_ionicConstraintOwners, attachedAtom.entity().m_ionicConstraintOwners, ownerPeriodicLiquid);
}

void InteractingAtom::createMetallicConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid)
{
	createConstrainter(attachedAtom, m_metallicConstraintOwners, attachedAtom.entity().m_metallicConstraintOwners, ownerPeriodicLiquid);
}

void InteractingAtom::createSecondNeighborConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid)
{
	createConstrainter(attachedAtom, m_secondNeighborConstraintOwners, attachedAtom.entity().m_secondNeighborConstraintOwners, ownerPeriodicLiquid);
}

void InteractingAtom::eraseInfeasibleChemicalBonds(const double maxBondingExtensionRate, PeriodicLiquid& ownerPeriodicLiquid)
{
	for (auto iter = m_chemicalBondOwners.begin(); iter != m_chemicalBondOwners.end();)
	{
		if (iter->second.isActive())
		{
			if (iter->second.owingChemicalBond().isFeasible(maxBondingExtensionRate))
				++iter;

			else
			{
				iter->second.disjoin();

				createSecondNeighborConstrainter(iter->first, ownerPeriodicLiquid);
				iter = m_chemicalBondOwners.erase(iter);
			}
		}

		else
			iter = m_chemicalBondOwners.erase(iter);
	}
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

void InteractingAtom::createChemicalBond(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid)
{
	createConstrainter(attachedAtom, m_chemicalBondOwners, attachedAtom.entity().m_chemicalBondOwners, ownerPeriodicLiquid);
}

void InteractingAtom::createPolyhedralConstrainter(const AtomicAdaptor& attachedAtom, PeriodicLiquid& ownerPeriodicLiquid, const AtomicAdaptor& bridgingAtom)
{
	if (!(hasPolyhedralConstrainter(attachedAtom)))
		createConstrainter(attachedAtom, m_polyhedralConstraintOwners, attachedAtom.entity().m_polyhedralConstraintOwners, ownerPeriodicLiquid);

	auto iter = m_polyhedralConstraintOwners.find(attachedAtom);
	iter->second.owingPolyhedralConstrainter().shareCommonBridgingAtom(bridgingAtom);
}

std::string InteractingAtom::getChemicalBondsKey() const noexcept
{
	std::vector<size_type> bondingAtomicNumbers;
	{
		for (const auto& chemicalBondOwner : m_chemicalBondOwners)
		{
			if (chemicalBondOwner.second.isActive())
				bondingAtomicNumbers.push_back(chemicalBondOwner.first.entity().atomicNumber());
		}

		std::sort(bondingAtomicNumbers.begin(), bondingAtomicNumbers.end());
	}

	std::string chemicalBondsKey;
	{
		for (const auto num : bondingAtomicNumbers)
		{
			chemicalBondsKey += SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol(num);
			chemicalBondsKey += "-";
		}


		if (!(chemicalBondsKey.empty()))
			chemicalBondsKey.pop_back();
	}

	return chemicalBondsKey;
}

std::string InteractingAtom::getCommonBridgingAtomsSharingsKey() const noexcept
{
	std::vector<std::pair<size_type, size_type>> commonBridgingAtomSharings;
	{
		for (const auto& polyhedralConstraintOwner : m_polyhedralConstraintOwners)
		{
			if (polyhedralConstraintOwner.second.isActive())
			{
				const size_type attachedAtomicNumber = polyhedralConstraintOwner.first.entity().atomicNumber();
				const PolyhedralConstrainter& polyhedralConstrainter = polyhedralConstraintOwner.second.owingPolyhedralConstrainter();

				commonBridgingAtomSharings.push_back(std::make_pair(attachedAtomicNumber, polyhedralConstrainter.commonBridgingAtomsSharing()));
			}
		}

		std::sort(commonBridgingAtomSharings.begin(), commonBridgingAtomSharings.end());
	}

	std::string bridgingAtomSharingsKey;
	{
		for (const auto numAndCount : commonBridgingAtomSharings)
		{
			bridgingAtomSharingsKey += SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol(numAndCount.first);
			bridgingAtomSharingsKey += std::to_string(numAndCount.second);
			bridgingAtomSharingsKey += "-";
		}


		if (!(bridgingAtomSharingsKey.empty()))
			bridgingAtomSharingsKey.pop_back();
	}

	return bridgingAtomSharingsKey;
}

std::string InteractingAtom::getIonicConstraintsKey(const double maxBondingLengthErrorRate) const noexcept
{
	std::vector<size_type> bondingAtomicNumbers;
	{
		for (const auto& ionicConstraintOwner : m_ionicConstraintOwners)
		{
			if (ionicConstraintOwner.second.isActive())
			{
				const size_type attachedAtomicNumber = ionicConstraintOwner.first.entity().atomicNumber();
				const IonicConstrainter& ionicConstrainter = ionicConstraintOwner.second.owingIonicConstrainter();

				double maxBondingLength = maxBondingLengthErrorRate * ionicConstrainter.minimumDistance();

				if(ionicConstrainter.interatomicDistance() < maxBondingLength)
					bondingAtomicNumbers.push_back(attachedAtomicNumber);
			}
		}

		std::sort(bondingAtomicNumbers.begin(), bondingAtomicNumbers.end());
	}

	std::string ionicConstraintersKey;
	{
		for (const auto num : bondingAtomicNumbers)
		{
			ionicConstraintersKey += SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol(num);
			ionicConstraintersKey += "-";
		}


		if (!(ionicConstraintersKey.empty()))
			ionicConstraintersKey.pop_back();
	}

	return ionicConstraintersKey;
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

bool InteractingAtom::tryConfigureChemicalBonds(const std::map<double, AtomicAdaptor>& promisingAtomicAdaptors, const size_type additionalCoordinationNumber, PeriodicLiquid& ownerPeriodicLiquid)
{
	if (additionalCoordinationNumber <= promisingAtomicAdaptors.size())
	{
		size_type chemicalBondsCreating = 0;
		{
			for (auto iter = promisingAtomicAdaptors.begin(); iter != promisingAtomicAdaptors.end(); ++iter)
			{
				++chemicalBondsCreating;
				createChemicalBond(iter->second, ownerPeriodicLiquid);
				eraseIonicConstrainter(iter->second);

				if (chemicalBondsCreating == additionalCoordinationNumber)
					break;
			}
		}

		return true;
	}

	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
