#include "PeriodicLiquid.h"

#include <cmath>

using namespace SolidStatePhysics::Crystallization::Components;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

PeriodicLiquid::PeriodicLiquid()
	: CrystalStructure{}
	, m_chemicalBondsPtrs{}
	, m_ionicConstraintersPtrs{}
	, m_metallicConstraintersPtrs{}
	, m_polyhedralConstraintersPtrs{}
	, m_secondNeighborConstraintersPtrs{}
{
}

PeriodicLiquid::PeriodicLiquid(const CrystalLattice& crystalLattice, const AtomicArrangement& atomicArrangement, const std::string& name)
	: CrystalStructure{ crystalLattice, atomicArrangement, name }
	, m_chemicalBondsPtrs{}
	, m_ionicConstraintersPtrs{}
	, m_metallicConstraintersPtrs{}
	, m_polyhedralConstraintersPtrs{}
	, m_secondNeighborConstraintersPtrs{}
{
}

PeriodicLiquid::PeriodicLiquid(const CrystalLattice& crystalLattice, AtomicArrangement&& atomicArrangement, const std::string& name)
	: CrystalStructure{ crystalLattice, std::move(atomicArrangement), name }
	, m_chemicalBondsPtrs{}
	, m_ionicConstraintersPtrs{}
	, m_metallicConstraintersPtrs{}
	, m_polyhedralConstraintersPtrs{}
	, m_secondNeighborConstraintersPtrs{}
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

void PeriodicLiquid::renewCrystalLatticeRandomly(const double minLength, const double maxLength)
{
	CrystalStructure::renewCrystalLatticeRandomly(minLength, maxLength);
	NumericMatrix primitiveLatticeVectors = crystalLattice().primitiveLatticeVectors();
	double latticeVolume = crystalLattice().getPrimitiveCellVolume();

	double sphericalVolumeSum = 0.0;
	{
		constexpr double volumeCoef = 4.0 * NumericalAnalysis::MathematicalConstants::getPi() / 3.0;

		for (const auto& atom : atomicArrangement())
		{
			double radius = atom.metallicRadius();
			sphericalVolumeSum += volumeCoef * radius * radius * radius;
		}
	}

	double scalingMultiplier = 0.0;
	{
		double scalingMultiplierCubic = sphericalVolumeSum / (latticeVolume * 0.7);
		scalingMultiplier = std::cbrt(scalingMultiplierCubic);
	}

	primitiveLatticeVectors *= scalingMultiplier;
	crystalLattice().resetPrimitiveLatticeVectors(primitiveLatticeVectors);
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

std::list<SolidStatePhysics::Crystallization::Components::PeriodicLiquid::AtomicAdaptor::LatticePoint> PeriodicLiquid::getVisibleLatticePoints(const IntelligentAtom& originAtom, const IntelligentAtom& targetAtom) const noexcept
{
	double neighborsDetectionZoneRadius = 0.0;
	{
		if (originAtom.isCation() && targetAtom.isCation())
			neighborsDetectionZoneRadius = MetallicConstrainter::formationRangeCoef() * (originAtom.metallicRadius() + targetAtom.metallicRadius());
		else
			neighborsDetectionZoneRadius = IonicConstrainter::formationRangeCoef() * (originAtom.maximumIonicRadius() + targetAtom.maximumIonicRadius());
	}
	double neighborsDetectionZoneRadiusSquare = (neighborsDetectionZoneRadius * neighborsDetectionZoneRadius);

	std::list<AtomicAdaptor::LatticePoint> latticePoints;
	NumericMatrix inverseMatrix = crystalLattice().inversePrimitiveLatticeVectors();

	const NumericVector fractionalCoordinate = inverseMatrix * originAtom.cartesianCoordinate();
	double lengthA = (neighborsDetectionZoneRadius * inverseMatrix.rowVectorSlice(0).norm());
	double lengthB = (neighborsDetectionZoneRadius * inverseMatrix.rowVectorSlice(1).norm());
	double lengthC = (neighborsDetectionZoneRadius * inverseMatrix.rowVectorSlice(2).norm());

	std::vector<int> indicesA;
	{
		int min = static_cast<int>(std::floor(fractionalCoordinate[0] - lengthA));
		int max = static_cast<int>(std::floor(fractionalCoordinate[0] + lengthA));

		for (int index = min; index <= max; ++index)
			indicesA.push_back(index);
	}

	std::vector<int> indicesB;
	{
		int min = static_cast<int>(std::floor(fractionalCoordinate[1] - lengthB));
		int max = static_cast<int>(std::floor(fractionalCoordinate[1] + lengthB));

		for (int index = min; index <= max; ++index)
			indicesB.push_back(index);
	}

	std::vector<int> indicesC;
	{
		int min = static_cast<int>(std::floor(fractionalCoordinate[2] - lengthC));
		int max = static_cast<int>(std::floor(fractionalCoordinate[2] + lengthC));

		for (int index = min; index <= max; ++index)
			indicesC.push_back(index);
	}

	for (const auto numA : indicesA)
	{
		for (const auto numB : indicesB)
		{
			for (const auto numC : indicesC)
				latticePoints.push_back(AtomicAdaptor::LatticePoint{ numA, numB, numC });
		}
	}

	return latticePoints;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
