#include "PolyhedralConstrainter.h"

#include "AbstractChemicalElement.h"

#include <algorithm>
#include <cmath>

using namespace SolidStatePhysics::Crystallization::Components;

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and destructor

std::set<std::pair<std::string, std::string>> PolyhedralConstrainter::m_vertexSharingCationicPairs;
std::set<std::pair<std::string, std::string>> PolyhedralConstrainter::m_edgeSharingCationicPairs;
std::set<std::pair<std::string, std::string>> PolyhedralConstrainter::m_faceSharingCationicPairs;


PolyhedralConstrainter::PolyhedralConstrainter(PolyhedralConstraintOwner& realOwner, const PolyhedralConstraintOwnerAdaptor& attachedOwner)
	: MinimumConstrainter{ realOwner.ownerAtom(), AtomicAdaptor(attachedOwner.entity().ownerAtom(), attachedOwner) }
	, m_realOwner{ std::ref(realOwner) }
	, m_attachedOwner{ attachedOwner }
	, m_commonBridgingAtoms{}
	, m_isAnionicConstrainter{}
	, m_maxPossibleBridgingAtomsSharing{ 0 }
{
	{
		if (ownerAtom().isAnion() && attachedAtom().entity().isAnion())
			m_isAnionicConstrainter = true;

		else
		{
			if (ownerAtom().isCation() && attachedAtom().entity().isCation())
				m_isAnionicConstrainter = false;

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Logic error\n";
				errorMessage += "Location: The constructor of the class of SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter.\n";
				errorMessage += "Message : It is neither anionic nor cationic constrainter.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::logic_error(errorMessage);
			}
		}
	}

	m_maxPossibleBridgingAtomsSharing = getMaxPossibleBridgingAtomsSharing(m_realOwner.get().ownerAtom(), m_attachedOwner.entity().ownerAtom());
}

PolyhedralConstrainter::PolyhedralConstrainter(PolyhedralConstraintOwner& realOwner, PolyhedralConstraintOwnerAdaptor&& attachedOwner)
	: MinimumConstrainter{ realOwner.ownerAtom(), AtomicAdaptor(attachedOwner.entity().ownerAtom(), attachedOwner) }
	, m_realOwner{ std::ref(realOwner) }
	, m_attachedOwner{ std::move(attachedOwner) }
	, m_commonBridgingAtoms{}
	, m_isAnionicConstrainter{}
	, m_maxPossibleBridgingAtomsSharing{ 0 }
{
	{
		if (ownerAtom().isAnion() && attachedAtom().entity().isAnion())
			m_isAnionicConstrainter = true;

		else
		{
			if (ownerAtom().isCation() && attachedAtom().entity().isCation())
				m_isAnionicConstrainter = false;

			else
			{
				std::string errorMessage = "------------------------------------------------------\n\n";
				errorMessage += "Error   : Logic error\n";
				errorMessage += "Location: The constructor of the class of SolidStatePhysics::Crystallization::Components::PolyhedralConstrainter.\n";
				errorMessage += "Message : It is neither anionic nor cationic constrainter.\n\n";
				errorMessage += "------------------------------------------------------\n\n";
				throw std::logic_error(errorMessage);
			}
		}
	}

	m_maxPossibleBridgingAtomsSharing = getMaxPossibleBridgingAtomsSharing(m_realOwner.get().ownerAtom(), m_attachedOwner.entity().ownerAtom());
}

PolyhedralConstrainter::PolyhedralConstrainter(const std::shared_ptr<PolyhedralConstrainter>& twinPtr)
	: MinimumConstrainter{ std::static_pointer_cast<MinimumConstrainter>(twinPtr) }
	, m_realOwner{ std::ref(twinPtr->m_attachedOwner.entity()) }
	, m_attachedOwner{ twinPtr->m_realOwner, twinPtr->m_attachedOwner }
	, m_commonBridgingAtoms{}
	, m_isAnionicConstrainter{ twinPtr->m_isAnionicConstrainter }
	, m_maxPossibleBridgingAtomsSharing{ twinPtr->m_maxPossibleBridgingAtomsSharing }
{
	for (const auto& commonBridgingAtom : twinPtr->m_commonBridgingAtoms)
	{
		AtomicAdaptor twinAtomicAdaptor{ commonBridgingAtom };
		twinAtomicAdaptor.resetLatticePoint(m_attachedOwner.getRelativeLatticePointTo(commonBridgingAtom));
		m_commonBridgingAtoms.push_back(std::move(twinAtomicAdaptor));
	}

	m_attachedOwner.reverseLatticePoint();
}

PolyhedralConstrainter::~PolyhedralConstrainter()
{
	if (std::shared_ptr<PolyhedralConstrainter> twinPolyhedralConstrainterPtr = std::dynamic_pointer_cast<PolyhedralConstrainter>(twinPtr()))
	{
		twinPolyhedralConstrainterPtr->m_realOwner.get().disjoin();
		twinPolyhedralConstrainterPtr->resetTwinPtr();
	}
}

// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void PolyhedralConstrainter::removeExcessCommonBridgingAtoms(PeriodicLiquid& ownerPeriodicLiquid)
{
	std::shared_ptr<PolyhedralConstrainter> twinPolyhedralConstrainterPtr = std::dynamic_pointer_cast<PolyhedralConstrainter>(twinPtr());


	if (ownerAtom().bondingMotivation().second < attachedAtom().entity().bondingMotivation().second)
	{
		while (m_maxPossibleBridgingAtomsSharing < m_commonBridgingAtoms.size())
		{
			std::sort(m_commonBridgingAtoms.begin(), m_commonBridgingAtoms.end(), [](const AtomicAdaptor& aaa, const AtomicAdaptor& aab) { return (aaa.entity().bondingMotivation().second > aab.entity().bondingMotivation().second); });
			ownerAtom().eraseChemicalBond(m_commonBridgingAtoms.back());

			if (!(ownerAtom().hasIonicConstrainter(m_commonBridgingAtoms.back())))
				ownerAtom().createIonicConstrainter(m_commonBridgingAtoms.back(), ownerPeriodicLiquid);

			m_commonBridgingAtoms.pop_back();
		}

		synchronizeCommonBridgingAtoms();
	}

	else
	{
		std::vector<AtomicAdaptor>& twinCommonBridgingAtoms = twinPolyhedralConstrainterPtr->m_commonBridgingAtoms;

		while (m_maxPossibleBridgingAtomsSharing < twinCommonBridgingAtoms.size())
		{
			std::sort(twinCommonBridgingAtoms.begin(), twinCommonBridgingAtoms.end(), [](const AtomicAdaptor& aaa, const AtomicAdaptor& aab) { return (aaa.entity().bondingMotivation().second > aab.entity().bondingMotivation().second); });
			attachedAtom().entity().eraseChemicalBond(twinCommonBridgingAtoms.back());

			if (!(attachedAtom().entity().hasIonicConstrainter(twinCommonBridgingAtoms.back())))
				attachedAtom().entity().createIonicConstrainter(twinCommonBridgingAtoms.back(), ownerPeriodicLiquid);

			twinCommonBridgingAtoms.pop_back();
		}

		twinPolyhedralConstrainterPtr->synchronizeCommonBridgingAtoms();
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void PolyhedralConstrainter::synchronizeCommonBridgingAtoms() noexcept
{
	std::shared_ptr<PolyhedralConstrainter> twinPolyhedralConstrainterPtr = std::dynamic_pointer_cast<PolyhedralConstrainter>(twinPtr());
	twinPolyhedralConstrainterPtr->m_commonBridgingAtoms.clear();

	for (const auto& commonBridgingAtom : m_commonBridgingAtoms)
	{
		AtomicAdaptor twinAtomicAdaptor{ commonBridgingAtom };
		twinAtomicAdaptor.resetLatticePoint(m_attachedOwner.getRelativeLatticePointTo(commonBridgingAtom));
		twinPolyhedralConstrainterPtr->m_commonBridgingAtoms.push_back(std::move(twinAtomicAdaptor));
	}
}

double PolyhedralConstrainter::getDistanceToFace(const double cationicRadius, const double minAnionicRadius, const size_type maxPossibleBridgingAtomsSharing) const noexcept
{
	double minBondingDistance = cationicRadius + minAnionicRadius;

	if (maxPossibleBridgingAtomsSharing == 2)
	{
		double distanceToFaceSquare = (minBondingDistance * minBondingDistance) - (minAnionicRadius * minAnionicRadius);
		return std::sqrt(distanceToFaceSquare);
	}

	else if (maxPossibleBridgingAtomsSharing == 3)
	{
		constexpr double coefficient = (4.0 / 3.0);
		double distanceToFaceSquare = (minBondingDistance * minBondingDistance) - (coefficient * minAnionicRadius * minAnionicRadius);
		return std::sqrt(distanceToFaceSquare);
	}

	else if (maxPossibleBridgingAtomsSharing == 4)
	{
		double distanceToFaceSquare = (minBondingDistance * minBondingDistance) - (2.0 * minAnionicRadius * minAnionicRadius);
		return std::sqrt(distanceToFaceSquare);
	}

	else
		return cationicRadius;
}

PolyhedralConstrainter::size_type PolyhedralConstrainter::getMaxPossibleBridgingAtomsSharing(const InteractingAtom& ownerAtom, const InteractingAtom& attachedAtomEntity) noexcept
{
	std::string ownerElementSymbol = SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol(ownerAtom.atomicNumber());
	std::string attachedElementSymbol = SolidStatePhysics::Crystallization::Calculators::AbstractChemicalElement::getElementSymbol(attachedAtomEntity.atomicNumber());
	std::pair<std::string, std::string> elementSymbolsPair = std::make_pair(ownerElementSymbol, attachedElementSymbol);
	std::pair<std::string, std::string> elementSymbolsPairReverse = std::make_pair(attachedElementSymbol, ownerElementSymbol);


	if (m_vertexSharingCationicPairs.find(elementSymbolsPair) != m_vertexSharingCationicPairs.end())
		return 1;
	else if (m_vertexSharingCationicPairs.find(elementSymbolsPairReverse) != m_vertexSharingCationicPairs.end())
		return 1;
	else if (m_edgeSharingCationicPairs.find(elementSymbolsPair) != m_edgeSharingCationicPairs.end())
		return 2;
	else if (m_edgeSharingCationicPairs.find(elementSymbolsPairReverse) != m_edgeSharingCationicPairs.end())
		return 2;
	else if (m_faceSharingCationicPairs.find(elementSymbolsPair) != m_faceSharingCationicPairs.end())
		return 3;
	else if (m_faceSharingCationicPairs.find(elementSymbolsPairReverse) != m_faceSharingCationicPairs.end())
		return 3;
	else
		return std::min(ownerAtom.maxCommonBridgingAtomsSharing(), attachedAtomEntity.maxCommonBridgingAtomsSharing());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
